@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2015 The Qt Company Ltd.
+** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the ActiveQt framework of the Qt Toolkit.
 **
@@ -17,8 +17,8 @@
 **     notice, this list of conditions and the following disclaimer in
 **     the documentation and/or other materials provided with the
 **     distribution.
-**   * Neither the name of Digia Plc and its Subsidiary(-ies) nor the names
-**     of its contributors may be used to endorse or promote products derived
+**   * Neither the name of The Qt Company Ltd nor the names of its
+**     contributors may be used to endorse or promote products derived
 **     from this software without specific prior written permission.
 **
 **
@@ -46,11 +46,14 @@
 #include <QtCore/qfactoryinterface.h>
 #include <QtCore/qmetaobject.h>
 #include <QtCore/qstringlist.h>
+#include <QtCore/qsettings.h>
 
 struct IUnknown;
 struct IDispatch;
 
 QT_BEGIN_NAMESPACE
+QT_WARNING_PUSH
+QT_WARNING_DISABLE_GCC("-Woverloaded-virtual") // gcc complains about QObject::metaObject() being hidden.
 
 class QWidget;
 class QSettings;
@@ -60,35 +63,35 @@ class QAxFactory : public QObject
 public:
     QAxFactory(const QUuid &libId, const QUuid &appId);
     virtual ~QAxFactory();
-    
+
     virtual QStringList featureList() const = 0;
 
     virtual QObject *createObject(const QString &key) = 0;
     virtual const QMetaObject *metaObject(const QString &key) const = 0;
     virtual bool createObjectWrapper(QObject *object, IDispatch **wrapper);
-    
+
     virtual QUuid classID(const QString &key) const;
     virtual QUuid interfaceID(const QString &key) const;
     virtual QUuid eventsID(const QString &key) const;
-    
+
     virtual QUuid typeLibID() const;
     virtual QUuid appID() const;
-    
+
     virtual void registerClass(const QString &key, QSettings *) const;
     virtual void unregisterClass(const QString &key, QSettings *) const;
-    
+
     virtual bool validateLicenseKey(const QString &key, const QString &licenseKey) const;
-    
+
     virtual QString exposeToSuperClass(const QString &key) const;
     virtual bool stayTopLevel(const QString &key) const;
     virtual bool hasStockEvents(const QString &key) const;
     virtual bool isService() const;
-    
+
     enum ServerType {
         SingleInstance,
         MultipleInstances
     };
-    
+
     static bool isServer();
     static QString serverDirPath();
     static QString serverFilePath();
@@ -96,7 +99,7 @@ public:
     static bool stopServer();
 
     static bool registerActiveObject(QObject *object);
-    
+
 private:
     QUuid typelib;
     QUuid app;
@@ -120,12 +123,12 @@ inline bool QAxFactory::stopServer()
     return qax_stopServer();
 }
 
-#define QAXFACTORY_EXPORT(IMPL, TYPELIB, APPID)	\
+#define QAXFACTORY_EXPORT(IMPL, TYPELIB, APPID) \
     QT_BEGIN_NAMESPACE \
-    QAxFactory *qax_instantiate()		\
-    {							\
-        IMPL *impl = new IMPL(QUuid(TYPELIB), QUuid(APPID));	\
-        return impl;					\
+    QAxFactory *qax_instantiate()               \
+    {                                                   \
+        IMPL *impl = new IMPL(QUuid(TYPELIB), QUuid(APPID));    \
+        return impl;                                    \
     } \
     QT_END_NAMESPACE
 
@@ -136,37 +139,37 @@ inline bool QAxFactory::stopServer()
     public: \
         QAxDefaultFactory(const QUuid &app, const QUuid &lib) \
         : QAxFactory(app, lib), className(QLatin1String(#Class)) {} \
-        QStringList featureList() const \
+        QStringList featureList() const override \
         { \
             QStringList list; \
             list << className; \
             return list; \
         } \
-        const QMetaObject *metaObject(const QString &key) const \
+        const QMetaObject *metaObject(const QString &key) const override \
         { \
             if (key == className) \
             return &Class::staticMetaObject; \
-            return 0; \
+            return Q_NULLPTR; \
         } \
-        QObject *createObject(const QString &key) \
+        QObject *createObject(const QString &key) override \
         { \
             if (key == className) \
-                return new Class(0); \
-            return 0; \
+                return new Class(Q_NULLPTR); \
+            return Q_NULLPTR; \
         } \
-        QUuid classID(const QString &key) const \
+        QUuid classID(const QString &key) const override \
         { \
             if (key == className) \
                 return QUuid(IIDClass); \
             return QUuid(); \
         } \
-        QUuid interfaceID(const QString &key) const \
+        QUuid interfaceID(const QString &key) const override \
         { \
             if (key == className) \
                 return QUuid(IIDInterface); \
             return QUuid(); \
         } \
-        QUuid eventsID(const QString &key) const \
+        QUuid eventsID(const QString &key) const override \
         { \
             if (key == className) \
                 return QUuid(IIDEvents); \
@@ -185,17 +188,47 @@ public:
     QAxClass(const QString &libId, const QString &appId)
     : QAxFactory(libId, appId)
     {}
-    
-    const QMetaObject *metaObject(const QString &) const { return &T::staticMetaObject; }
-    QStringList featureList() const { return QStringList(QString(T::staticMetaObject.className())); }
-    QObject *createObject(const QString &key)
+
+    const QMetaObject *metaObject(const QString &) const Q_DECL_OVERRIDE { return &T::staticMetaObject; }
+    QStringList featureList() const Q_DECL_OVERRIDE { return QStringList(QLatin1String(T::staticMetaObject.className())); }
+    QObject *createObject(const QString &key) Q_DECL_OVERRIDE
     {
         const QMetaObject &mo = T::staticMetaObject;
         if (key != QLatin1String(mo.className()))
-            return 0;
+            return Q_NULLPTR;
         if (!qstrcmp(mo.classInfo(mo.indexOfClassInfo("Creatable")).value(), "no"))
-            return 0;
-        return new T(0);
+            return Q_NULLPTR;
+        return new T(Q_NULLPTR);
+    }
+
+    void registerClass(const QString &key, QSettings *settings) const Q_DECL_OVERRIDE
+    {
+        const QStringList categories = getImplementedCategories();
+
+        for (QStringList::const_iterator it = categories.begin(), end = categories.end(); it != end; ++it) {
+            settings->setValue(QLatin1String("/CLSID/") + classID(key).toString()
+                             + QLatin1String("/Implemented Categories/") + *it + QLatin1String("/."),
+                               QString());
+        }
+    }
+
+    void unregisterClass(const QString &key, QSettings *settings) const Q_DECL_OVERRIDE
+    {
+        const QStringList categories = getImplementedCategories();
+
+        for (QStringList::const_iterator it = categories.begin(), end = categories.end(); it != end; ++it) {
+            settings->remove(QLatin1String("/CLSID/") + classID(key).toString()
+                           + QLatin1String("/Implemented Categories/") + *it + QLatin1String("/."));
+        }
+    }
+
+private:
+    /*! Retrieve list of comma-separated "Implemented Categories" Q_CLASSINFO UUIDs from T. */
+    static QStringList getImplementedCategories()
+    {
+        const QMetaObject &mo = T::staticMetaObject;
+        QString catids = mo.classInfo(mo.indexOfClassInfo("Implemented Categories")).value();
+        return catids.split(QLatin1Char(','));
     }
 };
 
@@ -210,7 +243,7 @@ public:
         QAxFactoryList() \
         : QAxFactory(IDTypeLib, IDApp) \
         { \
-            QAxFactory *factory = 0; \
+            QAxFactory *factory = Q_NULLPTR; \
             QStringList keys; \
             QStringList::Iterator it; \
 
@@ -218,76 +251,77 @@ public:
             factory = new QAxClass<Class>(typeLibID().toString(), appID().toString()); \
             qRegisterMetaType<Class*>(#Class"*"); \
             keys = factory->featureList(); \
-            for (it = keys.begin(); it != keys.end(); ++it) { \
-                factoryKeys += *it; \
-                factories.insert(*it, factory); \
-                creatable.insert(*it, true); \
+            Q_FOREACH (const QString &key, keys) { \
+                factoryKeys += key; \
+                factories.insert(key, factory); \
+                creatable.insert(key, true); \
             }\
 
 #define QAXTYPE(Class) \
             factory = new QAxClass<Class>(typeLibID().toString(), appID().toString()); \
             qRegisterMetaType<Class*>(#Class"*"); \
             keys = factory->featureList(); \
-            for (it = keys.begin(); it != keys.end(); ++it) { \
-                factoryKeys += *it; \
-                factories.insert(*it, factory); \
-                creatable.insert(*it, false); \
+            Q_FOREACH (const QString &key, keys) { \
+                factoryKeys += key; \
+                factories.insert(key, factory); \
+                creatable.insert(key, false); \
             }\
 
 #define QAXFACTORY_END() \
         } \
-        ~QAxFactoryList() { qDeleteAll(factories); } \
-        QStringList featureList() const {  return factoryKeys; } \
-        const QMetaObject *metaObject(const QString&key) const { \
+        ~QAxFactoryList() Q_DECL_OVERRIDE { qDeleteAll(factories); } \
+        QStringList featureList() const Q_DECL_OVERRIDE {  return factoryKeys; } \
+        const QMetaObject *metaObject(const QString&key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories[key]; \
-            return f ? f->metaObject(key) : 0; \
+            return f ? f->metaObject(key) : Q_NULLPTR; \
         } \
-        QObject *createObject(const QString &key) { \
+        QObject *createObject(const QString &key) Q_DECL_OVERRIDE { \
             if (!creatable.value(key)) \
-                return 0; \
+                return Q_NULLPTR; \
             QAxFactory *f = factories[key]; \
-            return f ? f->createObject(key) : 0; \
+            return f ? f->createObject(key) : Q_NULLPTR; \
         } \
-        QUuid classID(const QString &key) { \
+        QUuid classID(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->classID(key) : QUuid(); \
         } \
-        QUuid interfaceID(const QString &key) { \
+        QUuid interfaceID(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->interfaceID(key) : QUuid(); \
         } \
-        QUuid eventsID(const QString &key) { \
+        QUuid eventsID(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->eventsID(key) : QUuid(); \
         } \
-        void registerClass(const QString &key, QSettings *s) const { \
+        void registerClass(const QString &key, QSettings *s) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             if (f) f->registerClass(key, s); \
         } \
-        void unregisterClass(const QString &key, QSettings *s) const { \
+        void unregisterClass(const QString &key, QSettings *s) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             if (f) f->unregisterClass(key, s); \
         } \
-        QString exposeToSuperClass(const QString &key) const { \
+        QString exposeToSuperClass(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->exposeToSuperClass(key) : QString(); \
         } \
-        bool stayTopLevel(const QString &key) const { \
+        bool stayTopLevel(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->stayTopLevel(key) : false; \
         } \
-        bool hasStockEvents(const QString &key) const { \
+        bool hasStockEvents(const QString &key) const Q_DECL_OVERRIDE { \
             QAxFactory *f = factories.value(key); \
             return f ? f->hasStockEvents(key) : false; \
         } \
     }; \
-    QAxFactory *qax_instantiate()		\
-    {							\
-        QAxFactoryList *impl = new QAxFactoryList();	\
-        return impl;					\
+    QAxFactory *qax_instantiate()               \
+    {                                                   \
+        QAxFactoryList *impl = new QAxFactoryList();    \
+        return impl;                                    \
     } \
     QT_END_NAMESPACE
 
+QT_WARNING_POP
 QT_END_NAMESPACE
 
 #ifndef Q_COM_METATYPE_DECLARED
