@@ -1,8 +1,8 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Copyright (C) 2013 Intel Corporation
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2016 Intel Corporation.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtCore module of the Qt Toolkit.
 **
@@ -11,30 +11,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -176,6 +174,11 @@ QT_BEGIN_NAMESPACE
     Prevents the library from being unloaded from the address space if close()
     is called. The library's static variables are not reinitialized if open()
     is called at a later time.
+    \value DeepBindHint
+    Instructs the linker to prefer definitions in the loaded library
+    over exported definitions in the loading application when resolving
+    external symbols in the loaded library. This option is only supported
+    on Linux.
 
     \sa loadHints
 */
@@ -256,7 +259,8 @@ static bool findPatternUnloaded(const QString &library, QLibraryPrivate *lib)
     */
     bool hasMetaData = false;
     long pos = 0;
-    const char pattern[] = "QTMETADATA  ";
+    char pattern[] = "qTMETADATA  ";
+    pattern[0] = 'Q'; // Ensure the pattern "QTMETADATA" is not found in this library should QPluginLoader ever encounter it.
     const ulong plen = qstrlen(pattern);
 #if defined (Q_OF_ELF) && defined(Q_CC_GNU)
     int r = QElfParser().parse(filedata, fdlen, library, lib, &pos, &fdlen);
@@ -304,7 +308,7 @@ static bool findPatternUnloaded(const QString &library, QLibraryPrivate *lib)
     if (pos >= 0) {
         if (hasMetaData) {
             const char *data = filedata + pos;
-            QJsonDocument doc = QLibraryPrivate::fromRawMetaData(data);
+            QJsonDocument doc = qJsonFromRawLibraryMetaData(data);
             lib->metaData = doc.object();
             if (qt_debug_component())
                 qWarning("Found metadata in lib %s, metadata=\n%s\n",
@@ -339,12 +343,10 @@ static void installCoverageTool(QLibraryPrivate *libPrivate)
 
     if (qt_debug_component()) {
         if (ret >= 0) {
-            qDebug("%s: coverage data for %s registered",
-                     Q_FUNC_INFO,
+            qDebug("coverage data for %s registered",
                      qPrintable(libPrivate->fileName));
         } else {
-            qWarning("%s: could not register %s: error %d; coverage data may be incomplete",
-                     Q_FUNC_INFO,
+            qWarning("could not register %s: error %d; coverage data may be incomplete",
                      qPrintable(libPrivate->fileName),
                      ret);
         }
@@ -358,7 +360,7 @@ class QLibraryStore
 {
 public:
     inline ~QLibraryStore();
-    static inline QLibraryPrivate *findOrCreate(const QString &fileName, const QString &version);
+    static inline QLibraryPrivate *findOrCreate(const QString &fileName, const QString &version, QLibrary::LoadHints loadHints);
     static inline void releaseLibrary(QLibraryPrivate *lib);
 
     static inline void cleanup();
@@ -373,6 +375,7 @@ private:
 
 static QBasicMutex qt_library_mutex;
 static QLibraryStore *qt_library_data = 0;
+static bool qt_library_data_once;
 
 QLibraryStore::~QLibraryStore()
 {
@@ -409,7 +412,7 @@ inline void QLibraryStore::cleanup()
 
     if (qt_debug_component()) {
         // dump all objects that remain
-        foreach (QLibraryPrivate *lib, data->libraryMap) {
+        for (QLibraryPrivate *lib : qAsConst(data->libraryMap)) {
             if (lib)
                 qDebug() << "On QtCore unload," << lib->fileName << "was leaked, with"
                          << lib->libraryRefCount.load() << "users";
@@ -428,23 +431,33 @@ Q_DESTRUCTOR_FUNCTION(qlibraryCleanup)
 // must be called with a locked mutex
 QLibraryStore *QLibraryStore::instance()
 {
-    if (Q_UNLIKELY(!qt_library_data))
+    if (Q_UNLIKELY(!qt_library_data_once && !qt_library_data)) {
+        // only create once per process lifetime
         qt_library_data = new QLibraryStore;
+        qt_library_data_once = true;
+    }
     return qt_library_data;
 }
 
-inline QLibraryPrivate *QLibraryStore::findOrCreate(const QString &fileName, const QString &version)
+inline QLibraryPrivate *QLibraryStore::findOrCreate(const QString &fileName, const QString &version,
+                                                    QLibrary::LoadHints loadHints)
 {
     QMutexLocker locker(&qt_library_mutex);
     QLibraryStore *data = instance();
 
     // check if this library is already loaded
-    QLibraryPrivate *lib = data->libraryMap.value(fileName);
+    QLibraryPrivate *lib = 0;
+    if (Q_LIKELY(data)) {
+        lib = data->libraryMap.value(fileName);
+        if (lib)
+            lib->mergeLoadHints(loadHints);
+    }
     if (!lib)
-        lib = new QLibraryPrivate(fileName, version);
+        lib = new QLibraryPrivate(fileName, version, loadHints);
 
     // track this library
-    data->libraryMap.insert(fileName, lib);
+    if (Q_LIKELY(data) && !fileName.isEmpty())
+        data->libraryMap.insert(fileName, lib);
 
     lib->libraryRefCount.ref();
     return lib;
@@ -463,27 +476,42 @@ inline void QLibraryStore::releaseLibrary(QLibraryPrivate *lib)
     // no one else is using
     Q_ASSERT(lib->libraryUnloadCount.load() == 0);
 
-    QLibraryPrivate *that = data->libraryMap.take(lib->fileName);
-    Q_ASSERT(lib == that);
-    Q_UNUSED(that);
+    if (Q_LIKELY(data) && !lib->fileName.isEmpty()) {
+        QLibraryPrivate *that = data->libraryMap.take(lib->fileName);
+        Q_ASSERT(lib == that);
+        Q_UNUSED(that);
+    }
     delete lib;
 }
 
-QLibraryPrivate::QLibraryPrivate(const QString &canonicalFileName, const QString &version)
+QLibraryPrivate::QLibraryPrivate(const QString &canonicalFileName, const QString &version, QLibrary::LoadHints loadHints)
     : pHnd(0), fileName(canonicalFileName), fullVersion(version), instance(0),
-      loadHints(0),
       libraryRefCount(0), libraryUnloadCount(0), pluginState(MightBeAPlugin)
-{ }
-
-QLibraryPrivate *QLibraryPrivate::findOrCreate(const QString &fileName, const QString &version)
 {
-    return QLibraryStore::findOrCreate(fileName, version);
+    loadHintsInt.store(loadHints);
+    if (canonicalFileName.isEmpty())
+        errorString = QLibrary::tr("The shared library was not found.");
+}
+
+QLibraryPrivate *QLibraryPrivate::findOrCreate(const QString &fileName, const QString &version,
+                                               QLibrary::LoadHints loadHints)
+{
+    return QLibraryStore::findOrCreate(fileName, version, loadHints);
 }
 
 QLibraryPrivate::~QLibraryPrivate()
 {
 }
 
+void QLibraryPrivate::mergeLoadHints(QLibrary::LoadHints lh)
+{
+    // if the library is already loaded, we can't change the load hints
+    if (pHnd)
+        return;
+
+    loadHintsInt.store(lh);
+}
+
 QFunctionPointer QLibraryPrivate::resolve(const char *symbol)
 {
     if (!pHnd)
@@ -491,6 +519,12 @@ QFunctionPointer QLibraryPrivate::resolve(const char *symbol)
     return resolve_sys(symbol);
 }
 
+void QLibraryPrivate::setLoadHints(QLibrary::LoadHints lh)
+{
+    // this locks a global mutex
+    QMutexLocker lock(&qt_library_mutex);
+    mergeLoadHints(lh);
+}
 
 bool QLibraryPrivate::load()
 {
@@ -569,14 +603,14 @@ bool QLibraryPrivate::loadPlugin()
     \row \li Unix/Linux  \li \c .so
     \row \li AIX  \li \c .a
     \row \li HP-UX       \li \c .sl, \c .so (HP-UXi)
-    \row \li Mac OS X    \li \c .dylib, \c .bundle, \c .so
+    \row \li OS X and iOS   \li \c .dylib, \c .bundle, \c .so
     \endtable
 
     Trailing versioning numbers on Unix are ignored.
  */
 bool QLibrary::isLibrary(const QString &fileName)
 {
-#if defined(Q_OS_WIN32) || defined(Q_OS_WINCE)
+#if defined(Q_OS_WIN)
     return fileName.endsWith(QLatin1String(".dll"), Qt::CaseInsensitive);
 #else
     QString completeSuffix = QFileInfo(fileName).completeSuffix();
@@ -647,7 +681,7 @@ static bool qt_get_metadata(QtPluginQueryVerificationDataFunction pfn, QLibraryP
     if (!szData)
         return false;
 
-    QJsonDocument doc = QLibraryPrivate::fromRawMetaData(szData);
+    QJsonDocument doc = qJsonFromRawLibraryMetaData(szData);
     if (doc.isNull())
         return false;
     priv->metaData = doc.object();
@@ -806,7 +840,7 @@ QLibrary::QLibrary(QObject *parent)
     We recommend omitting the file's suffix in \a fileName, since
     QLibrary will automatically look for the file with the appropriate
     suffix in accordance with the platform, e.g. ".so" on Unix,
-    ".dylib" on Mac OS X, and ".dll" on Windows. (See \l{fileName}.)
+    ".dylib" on OS X and iOS, and ".dll" on Windows. (See \l{fileName}.)
  */
 QLibrary::QLibrary(const QString& fileName, QObject *parent)
     :QObject(parent), d(0), did_load(false)
@@ -823,7 +857,7 @@ QLibrary::QLibrary(const QString& fileName, QObject *parent)
     We recommend omitting the file's suffix in \a fileName, since
     QLibrary will automatically look for the file with the appropriate
     suffix in accordance with the platform, e.g. ".so" on Unix,
-    ".dylib" on Mac OS X, and ".dll" on Windows. (See \l{fileName}.)
+    ".dylib" on OS X and iOS, and ".dll" on Windows. (See \l{fileName}.)
 */
 QLibrary::QLibrary(const QString& fileName, int verNum, QObject *parent)
     :QObject(parent), d(0), did_load(false)
@@ -839,7 +873,7 @@ QLibrary::QLibrary(const QString& fileName, int verNum, QObject *parent)
     We recommend omitting the file's suffix in \a fileName, since
     QLibrary will automatically look for the file with the appropriate
     suffix in accordance with the platform, e.g. ".so" on Unix,
-    ".dylib" on Mac OS X, and ".dll" on Windows. (See \l{fileName}.)
+    ".dylib" on OS X and iOS, and ".dll" on Windows. (See \l{fileName}.)
  */
 QLibrary::QLibrary(const QString& fileName, const QString &version, QObject *parent)
     :QObject(parent), d(0), did_load(false)
@@ -871,7 +905,7 @@ QLibrary::~QLibrary()
     suffix (see isLibrary()).
 
     When loading the library, QLibrary searches in all system-specific
-    library locations (e.g. \c LD_LIBRARY_PATH on Unix), unless the
+    library locations (for example, \c LD_LIBRARY_PATH on Unix), unless the
     file name has an absolute path. After loading the library
     successfully, fileName() returns the fully-qualified file name of
     the library, including the full path to the library if one was given
@@ -887,13 +921,12 @@ void QLibrary::setFileName(const QString &fileName)
 {
     QLibrary::LoadHints lh;
     if (d) {
-        lh = d->loadHints;
+        lh = d->loadHints();
         d->release();
         d = 0;
         did_load = false;
     }
-    d = QLibraryPrivate::findOrCreate(fileName);
-    d->loadHints = lh;
+    d = QLibraryPrivate::findOrCreate(fileName, QString(), lh);
 }
 
 QString QLibrary::fileName() const
@@ -916,13 +949,12 @@ void QLibrary::setFileNameAndVersion(const QString &fileName, int verNum)
 {
     QLibrary::LoadHints lh;
     if (d) {
-        lh = d->loadHints;
+        lh = d->loadHints();
         d->release();
         d = 0;
         did_load = false;
     }
-    d = QLibraryPrivate::findOrCreate(fileName, verNum >= 0 ? QString::number(verNum) : QString());
-    d->loadHints = lh;
+    d = QLibraryPrivate::findOrCreate(fileName, verNum >= 0 ? QString::number(verNum) : QString(), lh);
 }
 
 /*!
@@ -938,13 +970,12 @@ void QLibrary::setFileNameAndVersion(const QString &fileName, const QString &ver
 {
     QLibrary::LoadHints lh;
     if (d) {
-        lh = d->loadHints;
+        lh = d->loadHints();
         d->release();
         d = 0;
         did_load = false;
     }
-    d = QLibraryPrivate::findOrCreate(fileName, version);
-    d->loadHints = lh;
+    d = QLibraryPrivate::findOrCreate(fileName, version, lh);
 }
 
 /*!
@@ -1051,7 +1082,7 @@ QString QLibrary::errorString() const
 
     You can give some hints on how the symbols are resolved. Usually,
     the symbols are not resolved at load time, but resolved lazily,
-    (that is, when resolve() is called). If you set the loadHint to
+    (that is, when resolve() is called). If you set the loadHints to
     ResolveAllSymbolsHint, then all symbols will be resolved at load time
     if the platform supports it.
 
@@ -1065,8 +1096,6 @@ QString QLibrary::errorString() const
     to the library \c shr_64.o in the archive file named \c libGL.a. This
     is only supported on the AIX platform.
 
-    Setting PreventUnloadHint will only apply on Unix platforms.
-
     The interpretation of the load hints is platform dependent, and if
     you use it you are probably making some assumptions on which platform
     you are compiling for, so use them only if you understand the consequences
@@ -1075,6 +1104,12 @@ QString QLibrary::errorString() const
     By default, none of these flags are set, so libraries will be loaded with
     lazy symbol resolution, and will not export external symbols for resolution
     in other dynamically-loaded libraries.
+
+    \note Setting this property after the library has been loaded has no effect
+    and loadHints() will not reflect those changes.
+
+    \note This property is shared among all QLibrary instances that refer to
+    the same library.
 */
 void QLibrary::setLoadHints(LoadHints hints)
 {
@@ -1082,21 +1117,18 @@ void QLibrary::setLoadHints(LoadHints hints)
         d = QLibraryPrivate::findOrCreate(QString());   // ugly, but we need a d-ptr
         d->errorString.clear();
     }
-    d->loadHints = hints;
+    d->setLoadHints(hints);
 }
 
 QLibrary::LoadHints QLibrary::loadHints() const
 {
-    return d ? d->loadHints : (QLibrary::LoadHints)0;
+    return d ? d->loadHints() : QLibrary::LoadHints();
 }
 
 /* Internal, for debugging */
 bool qt_debug_component()
 {
-    static int debug_env = -1;
-    if (debug_env == -1)
-       debug_env = QT_PREPEND_NAMESPACE(qgetenv)("QT_DEBUG_PLUGINS").toInt();
-
+    static int debug_env = QT_PREPEND_NAMESPACE(qEnvironmentVariableIntValue)("QT_DEBUG_PLUGINS");
     return debug_env != 0;
 }
 
