@@ -1,9 +1,9 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Copyright (C) 2012 Intel Corporation
+** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2016 Intel Corporation.
 ** Copyright (C) 2012 Olivier Goffart <ogoffart@woboq.com>
-** Contact: http://www.qt-project.org/legal
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtCore module of the Qt Toolkit.
 **
@@ -12,30 +12,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -50,6 +48,7 @@
 #include "qelapsedtimer.h"
 #include "qthread.h"
 #include "qmutex_p.h"
+#include "qtypetraits.h"
 
 #ifndef QT_LINUX_FUTEX
 #include "private/qfreelist_p.h"
@@ -59,7 +58,7 @@ QT_BEGIN_NAMESPACE
 
 static inline bool isRecursive(QMutexData *d)
 {
-    register quintptr u = quintptr(d);
+    quintptr u = quintptr(d);
     if (Q_LIKELY(u <= 0x3))
         return false;
 #ifdef QT_LINUX_FUTEX
@@ -75,8 +74,14 @@ class QRecursiveMutexPrivate : public QMutexData
 public:
     QRecursiveMutexPrivate()
         : QMutexData(QMutex::Recursive), owner(0), count(0) {}
-    Qt::HANDLE owner;
+
+    // written to by the thread that first owns 'mutex';
+    // read during attempts to acquire ownership of 'mutex' from any other thread:
+    QAtomicPointer<QtPrivate::remove_pointer<Qt::HANDLE>::type> owner;
+
+    // only ever accessed from the thread that owns 'mutex':
     uint count;
+
     QMutex mutex;
 
     bool lock(int timeout) QT_MUTEX_LOCK_NOEXCEPT;
@@ -140,6 +145,11 @@ public:
     lock calls unlock(). A non-blocking alternative to lock() is
     tryLock().
 
+    QMutex is optimized to be fast in the non-contended case. A non-recursive
+    QMutex will not allocate memory if there is no contention on that mutex.
+    It is constructed and destroyed with almost no overhead,
+    which means it is fine to have many mutexes as part of other classes.
+
     \sa QMutexLocker, QReadWriteLock, QSemaphore, QWaitCondition
 */
 
@@ -166,6 +176,8 @@ public:
     a thread may only lock a mutex once. The default is
     QMutex::NonRecursive.
 
+    Recursive mutexes are slower and take more memory than non-recursive ones.
+
     \sa lock(), unlock()
 */
 QMutex::QMutex(RecursionMode mode)
@@ -209,9 +221,9 @@ QMutex::~QMutex()
 */
 void QMutex::lock() QT_MUTEX_LOCK_NOEXCEPT
 {
-    if (fastTryLock())
+    QMutexData *current;
+    if (fastTryLock(current))
         return;
-    QMutexData *current = d_ptr.loadAcquire();
     if (QT_PREPEND_NAMESPACE(isRecursive)(current))
         static_cast<QRecursiveMutexPrivate *>(current)->lock(-1);
     else
@@ -220,8 +232,8 @@ void QMutex::lock() QT_MUTEX_LOCK_NOEXCEPT
 
 /*! \fn bool QMutex::tryLock(int timeout)
 
-    Attempts to lock the mutex. This function returns true if the lock
-    was obtained; otherwise it returns false. If another thread has
+    Attempts to lock the mutex. This function returns \c true if the lock
+    was obtained; otherwise it returns \c false. If another thread has
     locked the mutex, this function will wait for at most \a timeout
     milliseconds for the mutex to become available.
 
@@ -243,9 +255,9 @@ void QMutex::lock() QT_MUTEX_LOCK_NOEXCEPT
 */
 bool QMutex::tryLock(int timeout) QT_MUTEX_LOCK_NOEXCEPT
 {
-    if (fastTryLock())
+    QMutexData *current;
+    if (fastTryLock(current))
         return true;
-    QMutexData *current = d_ptr.loadAcquire();
     if (QT_PREPEND_NAMESPACE(isRecursive)(current))
         return static_cast<QRecursiveMutexPrivate *>(current)->lock(timeout);
     else
@@ -261,27 +273,30 @@ bool QMutex::tryLock(int timeout) QT_MUTEX_LOCK_NOEXCEPT
 */
 void QMutex::unlock() Q_DECL_NOTHROW
 {
-    if (fastTryUnlock())
+    QMutexData *current;
+    if (fastTryUnlock(current))
         return;
-    QMutexData *current = d_ptr.loadAcquire();
     if (QT_PREPEND_NAMESPACE(isRecursive)(current))
         static_cast<QRecursiveMutexPrivate *>(current)->unlock();
     else
         unlockInternal();
 }
 
-/*!
-    \fn void QMutex::isRecursive()
-    \since 5.0
-
-    Returns true if the mutex is recursive
-
-*/
-bool QBasicMutex::isRecursive()
+bool QBasicMutex::isRecursive() Q_DECL_NOTHROW
 {
     return QT_PREPEND_NAMESPACE(isRecursive)(d_ptr.loadAcquire());
 }
 
+/*!
+    \overload
+    \since 5.7
+
+    Returns \c true if the mutex is recursive
+*/
+bool QBasicMutex::isRecursive() const Q_DECL_NOTHROW
+{
+    return QT_PREPEND_NAMESPACE(isRecursive)(d_ptr.loadAcquire());
+}
 
 /*!
     \class QMutexLocker
@@ -373,7 +388,7 @@ bool QBasicMutex::isRecursive()
 */
 
 /*!
-    \fn QMutex *QMutexLocker::mutex()
+    \fn QMutex *QMutexLocker::mutex() const
 
     Returns the mutex on which the QMutexLocker is operating.
 
@@ -564,7 +579,35 @@ const int FreeListConstants::Sizes[FreeListConstants::BlockCount] = {
 };
 
 typedef QFreeList<QMutexPrivate, FreeListConstants> FreeList;
-Q_GLOBAL_STATIC(FreeList, freelist);
+// We cannot use Q_GLOBAL_STATIC because it uses QMutex
+#if defined(Q_COMPILER_THREADSAFE_STATICS)
+FreeList *freelist()
+{
+    static FreeList list;
+    return &list;
+}
+#else
+static QBasicAtomicPointer<FreeList> freeListPtr;
+
+FreeList *freelist()
+{
+    FreeList *local = freeListPtr.loadAcquire();
+    if (!local) {
+        local = new FreeList;
+        if (!freeListPtr.testAndSetRelease(0, local)) {
+            delete local;
+            local = freeListPtr.loadAcquire();
+        }
+    }
+    return local;
+}
+
+static void qFreeListDeleter()
+{
+    delete freeListPtr.load();
+}
+Q_DESTRUCTOR_FUNCTION(qFreeListDeleter)
+#endif
 }
 
 QMutexPrivate *QMutexPrivate::allocate()
@@ -611,7 +654,7 @@ void QMutexPrivate::derefWaiters(int value) Q_DECL_NOTHROW
 inline bool QRecursiveMutexPrivate::lock(int timeout) QT_MUTEX_LOCK_NOEXCEPT
 {
     Qt::HANDLE self = QThread::currentThreadId();
-    if (owner == self) {
+    if (owner.load() == self) {
         ++count;
         Q_ASSERT_X(count != 0, "QMutex::lock", "Overflow in recursion counter");
         return true;
@@ -624,7 +667,7 @@ inline bool QRecursiveMutexPrivate::lock(int timeout) QT_MUTEX_LOCK_NOEXCEPT
     }
 
     if (success)
-        owner = self;
+        owner.store(self);
     return success;
 }
 
@@ -636,7 +679,7 @@ inline void QRecursiveMutexPrivate::unlock() Q_DECL_NOTHROW
     if (count > 0) {
         count--;
     } else {
-        owner = 0;
+        owner.store(0);
         mutex.QBasicMutex::unlock();
     }
 }
