@@ -2,7 +2,7 @@
 **
 ** Copyright (C) 2013 Laszlo Papp <lpapp@kde.org>
 ** Copyright (C) 2013 David Faure <faure@kde.org>
-** Contact: http://www.qt-project.org/legal
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtCore module of the Qt Toolkit.
 **
@@ -11,30 +11,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -45,6 +43,10 @@
 #include <qcoreapplication.h>
 #include <qhash.h>
 #include <qvector.h>
+#include <qdebug.h>
+#if defined(Q_OS_WIN) && !defined(QT_BOOTSTRAPPED) && !defined(Q_OS_WINCE) && !defined(Q_OS_WINRT)
+#  include <qt_windows.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 
@@ -57,6 +59,7 @@ class QCommandLineParserPrivate
 public:
     inline QCommandLineParserPrivate()
         : singleDashWordOptionMode(QCommandLineParser::ParseAsCompactedShortOptions),
+          optionsAfterPositionalArgumentsMode(QCommandLineParser::ParseAsOptions),
           builtinVersionOption(false),
           builtinHelpOption(false),
           needsParsing(true)
@@ -107,6 +110,9 @@ public:
     //! The parsing mode for "-abc"
     QCommandLineParser::SingleDashWordOptionMode singleDashWordOptionMode;
 
+    //! How to parse "arg -option"
+    QCommandLineParser::OptionsAfterPositionalArgumentsMode optionsAfterPositionalArgumentsMode;
+
     //! Whether addVersionOption was called
     bool builtinVersionOption;
 
@@ -116,11 +122,12 @@ public:
     //! True if parse() needs to be called
     bool needsParsing;
 };
+Q_DECLARE_TYPEINFO(QCommandLineParserPrivate::PositionalArgumentDefinition, Q_MOVABLE_TYPE);
 
 QStringList QCommandLineParserPrivate::aliases(const QString &optionName) const
 {
-    const NameHash_t::const_iterator it = nameHash.find(optionName);
-    if (it == nameHash.end()) {
+    const NameHash_t::const_iterator it = nameHash.constFind(optionName);
+    if (it == nameHash.cend()) {
         qWarning("QCommandLineParser: option not defined: \"%s\"", qPrintable(optionName));
         return QStringList();
     }
@@ -179,6 +186,10 @@ QStringList QCommandLineParserPrivate::aliases(const QString &optionName) const
     Example:
     \snippet code/src_corelib_tools_qcommandlineparser_main.cpp 0
 
+    If your compiler supports the C++11 standard, the three addOption() calls in
+    the above example can be simplified:
+    \snippet code/src_corelib_tools_qcommandlineparser_main.cpp cxx11
+
     Known limitation: the parsing of Qt options inside QCoreApplication and subclasses
     happens before QCommandLineParser exists, so it can't take it into account. This
     means any option value that looks like a builtin Qt option, will be treated by
@@ -187,6 +198,65 @@ QStringList QCommandLineParserPrivate::aliases(const QString &optionName) const
     QCoreApplication::arguments() before QCommandLineParser defines the \c{profile}
     option and parses the command line.
 
+    \section2 How to Use QCommandLineParser in Complex Applications
+
+    In practice, additional error checking needs to be performed on the positional
+    arguments and option values. For example, ranges of numbers should be checked.
+
+    It is then advisable to introduce a function to do the command line parsing
+    which takes a struct or class receiving the option values returning an
+    enumeration representing the result. The dnslookup example of the QtNetwork
+    module illustrates this:
+
+    \snippet dnslookup.h 0
+
+    \snippet dnslookup.cpp 0
+
+    In the main function, help should be printed to the standard output if the help option
+    was passed and the application should return the exit code 0.
+
+    If an error was detected, the error message should be printed to the standard
+    error output and the application should return an exit code other than 0.
+
+    \snippet dnslookup.cpp 1
+
+    A special case to consider here are GUI applications on Windows and mobile
+    platforms. These applications may not use the standard output or error channels
+    since the output is either discarded or not accessible.
+
+    On Windows, QCommandLineParser uses message boxes to display usage information
+    and errors if no console window can be obtained.
+
+    For other platforms, it is recommended to display help texts and error messages
+    using a QMessageBox. To preserve the formatting of the help text, rich text
+    with \c <pre> elements should be used:
+
+    \code
+
+    switch (parseCommandLine(parser, &query, &errorMessage)) {
+    case CommandLineOk:
+        break;
+    case CommandLineError:
+        QMessageBox::warning(0, QGuiApplication::applicationDisplayName(),
+                             "<html><head/><body><h2>" + errorMessage + "</h2><pre>"
+                             + parser.helpText() + "</pre></body></html>");
+        return 1;
+    case CommandLineVersionRequested:
+        QMessageBox::information(0, QGuiApplication::applicationDisplayName(),
+                                 QGuiApplication::applicationDisplayName() + ' '
+                                 + QCoreApplication::applicationVersion());
+        return 0;
+    case CommandLineHelpRequested:
+        QMessageBox::warning(0, QGuiApplication::applicationDisplayName(),
+                             "<html><head/><body><pre>"
+                             + parser.helpText() + "</pre></body></html>");
+        return 0;
+    }
+    \endcode
+
+    However, this does not apply to the dnslookup example, because it is a
+    console application.
+
     \sa QCommandLineOption, QCoreApplication
 */
 
@@ -239,6 +309,41 @@ void QCommandLineParser::setSingleDashWordOptionMode(QCommandLineParser::SingleD
     d->singleDashWordOptionMode = singleDashWordOptionMode;
 }
 
+/*!
+    \enum QCommandLineParser::OptionsAfterPositionalArgumentsMode
+
+    This enum describes the way the parser interprets options that
+    occur after positional arguments.
+
+    \value ParseAsOptions \c{application argument --opt -t} is interpreted as setting
+    the options \c{opt} and \c{t}, just like \c{application --opt -t argument} would do.
+    This is the default parsing mode. In order to specify that \c{--opt} and \c{-t}
+    are positional arguments instead, the user can use \c{--}, as in
+    \c{application argument -- --opt -t}.
+
+    \value ParseAsPositionalArguments \c{application argument --opt} is interpreted as
+    having two positional arguments, \c{argument} and \c{--opt}.
+    This mode is useful for executables that aim to launch other executables
+    (e.g. wrappers, debugging tools, etc.) or that support internal commands
+    followed by options for the command. \c{argument} is the name of the command,
+    and all options occurring after it can be collected and parsed by another
+    command line parser, possibly in another executable.
+
+    \sa setOptionsAfterPositionalArgumentsMode()
+
+    \since 5.6
+*/
+
+/*!
+    Sets the parsing mode to \a parsingMode.
+    This must be called before process() or parse().
+    \since 5.6
+*/
+void QCommandLineParser::setOptionsAfterPositionalArgumentsMode(QCommandLineParser::OptionsAfterPositionalArgumentsMode parsingMode)
+{
+    d->optionsAfterPositionalArgumentsMode = parsingMode;
+}
+
 /*!
     Adds the option \a option to look for while parsing.
 
@@ -249,10 +354,10 @@ void QCommandLineParser::setSingleDashWordOptionMode(QCommandLineParser::SingleD
  */
 bool QCommandLineParser::addOption(const QCommandLineOption &option)
 {
-    QStringList optionNames = option.names();
+    const QStringList optionNames = option.names();
 
     if (!optionNames.isEmpty()) {
-        foreach (const QString &name, optionNames) {
+        for (const QString &name : optionNames) {
             if (d->nameHash.contains(name))
                 return false;
         }
@@ -260,7 +365,7 @@ bool QCommandLineParser::addOption(const QCommandLineOption &option)
         d->commandLineOptionList.append(option);
 
         const int offset = d->commandLineOptionList.size() - 1;
-        foreach (const QString &name, optionNames)
+        for (const QString &name : optionNames)
             d->nameHash.insert(name, offset);
 
         return true;
@@ -269,6 +374,26 @@ bool QCommandLineParser::addOption(const QCommandLineOption &option)
     return false;
 }
 
+/*!
+    \since 5.4
+
+    Adds the options to look for while parsing. The options are specified by
+    the parameter \a options.
+
+    Returns \c true if adding all of the options was successful; otherwise
+    returns \c false.
+
+    See the documentation for addOption() for when this function may fail.
+*/
+bool QCommandLineParser::addOptions(const QList<QCommandLineOption> &options)
+{
+    // should be optimized (but it's no worse than what was possible before)
+    bool result = true;
+    for (QList<QCommandLineOption>::const_iterator it = options.begin(), end = options.end(); it != end; ++it)
+        result &= addOption(*it);
+    return result;
+}
+
 /*!
     Adds the \c{-v} / \c{--version} option, which displays the version string of the application.
 
@@ -280,9 +405,9 @@ bool QCommandLineParser::addOption(const QCommandLineOption &option)
 */
 QCommandLineOption QCommandLineParser::addVersionOption()
 {
-    d->builtinVersionOption = true;
     QCommandLineOption opt(QStringList() << QStringLiteral("v") << QStringLiteral("version"), tr("Displays version information."));
     addOption(opt);
+    d->builtinVersionOption = true;
     return opt;
 }
 
@@ -300,7 +425,6 @@ QCommandLineOption QCommandLineParser::addVersionOption()
 */
 QCommandLineOption QCommandLineParser::addHelpOption()
 {
-    d->builtinHelpOption = true;
     QCommandLineOption opt(QStringList()
 #ifdef Q_OS_WIN
                 << QStringLiteral("?")
@@ -308,6 +432,7 @@ QCommandLineOption QCommandLineParser::addHelpOption()
                 << QStringLiteral("h")
                 << QStringLiteral("help"), tr("Displays this help."));
     addOption(opt);
+    d->builtinHelpOption = true;
     return opt;
 }
 
@@ -403,6 +528,47 @@ QString QCommandLineParser::errorText() const
     return QString();
 }
 
+enum MessageType { UsageMessage, ErrorMessage };
+
+#if defined(Q_OS_WIN) && !defined(QT_BOOTSTRAPPED) && !defined(Q_OS_WINCE) && !defined(Q_OS_WINRT)
+// Return whether to use a message box. Use handles if a console can be obtained
+// or we are run with redirected handles (for example, by QProcess).
+static inline bool displayMessageBox()
+{
+    if (GetConsoleWindow())
+        return false;
+    STARTUPINFO startupInfo;
+    startupInfo.cb = sizeof(STARTUPINFO);
+    GetStartupInfo(&startupInfo);
+    return !(startupInfo.dwFlags & STARTF_USESTDHANDLES);
+}
+#endif // Q_OS_WIN && !QT_BOOTSTRAPPED && !Q_OS_WIN && !Q_OS_WINRT
+
+static void showParserMessage(const QString &message, MessageType type)
+{
+#if defined(Q_OS_WINRT)
+    if (type == UsageMessage)
+        qInfo(qPrintable(message));
+    else
+        qCritical(qPrintable(message));
+    return;
+#elif defined(Q_OS_WIN) && !defined(QT_BOOTSTRAPPED) && !defined(Q_OS_WINCE)
+    if (displayMessageBox()) {
+        const UINT flags = MB_OK | MB_TOPMOST | MB_SETFOREGROUND
+            | (type == UsageMessage ? MB_ICONINFORMATION : MB_ICONERROR);
+        QString title;
+        if (QCoreApplication::instance())
+            title = QCoreApplication::instance()->property("applicationDisplayName").toString();
+        if (title.isEmpty())
+            title = QCoreApplication::applicationName();
+        MessageBoxW(0, reinterpret_cast<const wchar_t *>(message.utf16()),
+                    reinterpret_cast<const wchar_t *>(title.utf16()), flags);
+        return;
+    }
+#endif // Q_OS_WIN && !QT_BOOTSTRAPPED && !Q_OS_WINCE
+    fputs(qPrintable(message), type == UsageMessage ? stdout : stderr);
+}
+
 /*!
     Processes the command line \a arguments.
 
@@ -419,14 +585,12 @@ QString QCommandLineParser::errorText() const
 void QCommandLineParser::process(const QStringList &arguments)
 {
     if (!d->parse(arguments)) {
-        fprintf(stderr, "%s\n", qPrintable(errorText()));
+        showParserMessage(errorText() + QLatin1Char('\n'), ErrorMessage);
         ::exit(EXIT_FAILURE);
     }
 
-    if (d->builtinVersionOption && isSet(QStringLiteral("version"))) {
-        printf("%s %s\n", qPrintable(QCoreApplication::applicationName()), qPrintable(QCoreApplication::applicationVersion()));
-        ::exit(EXIT_SUCCESS);
-    }
+    if (d->builtinVersionOption && isSet(QStringLiteral("version")))
+        showVersion();
 
     if (d->builtinHelpOption && isSet(QStringLiteral("help")))
         showHelp(EXIT_SUCCESS);
@@ -528,7 +692,7 @@ bool QCommandLineParserPrivate::parse(const QStringList &args)
     const QLatin1Char dashChar('-');
     const QLatin1Char assignChar('=');
 
-    bool doubleDashFound = false;
+    bool forcePositional = false;
     errorText.clear();
     positionalArgumentList.clear();
     optionNames.clear();
@@ -546,7 +710,7 @@ bool QCommandLineParserPrivate::parse(const QStringList &args)
     for (; argumentIterator != args.end() ; ++argumentIterator) {
         QString argument = *argumentIterator;
 
-        if (doubleDashFound) {
+        if (forcePositional) {
             positionalArgumentList.append(argument);
         } else if (argument.startsWith(doubleDashString)) {
             if (argument.length() > 2) {
@@ -558,7 +722,7 @@ bool QCommandLineParserPrivate::parse(const QStringList &args)
                     error = true;
                 }
             } else {
-                doubleDashFound = true;
+                forcePositional = true;
             }
         } else if (argument.startsWith(dashChar)) {
             if (argument.size() == 1) { // single dash ("stdin")
@@ -610,6 +774,8 @@ bool QCommandLineParserPrivate::parse(const QStringList &args)
             }
         } else {
             positionalArgumentList.append(argument);
+            if (optionsAfterPositionalArgumentsMode == QCommandLineParser::ParseAsPositionalArguments)
+                forcePositional = true;
         }
         if (argumentIterator == args.end())
             break;
@@ -637,7 +803,7 @@ bool QCommandLineParser::isSet(const QString &name) const
     if (d->optionNames.contains(name))
         return true;
     const QStringList aliases = d->aliases(name);
-    foreach (const QString &optionName, d->optionNames) {
+    for (const QString &optionName : qAsConst(d->optionNames)) {
         if (aliases.contains(optionName))
             return true;
     }
@@ -694,8 +860,8 @@ QString QCommandLineParser::value(const QString &optionName) const
 QStringList QCommandLineParser::values(const QString &optionName) const
 {
     d->checkParsed("values");
-    const NameHash_t::const_iterator it = d->nameHash.find(optionName);
-    if (it != d->nameHash.end()) {
+    const NameHash_t::const_iterator it = d->nameHash.constFind(optionName);
+    if (it != d->nameHash.cend()) {
         const int optionOffset = *it;
         QStringList values = d->optionValuesHash.value(optionOffset);
         if (values.isEmpty())
@@ -720,7 +886,9 @@ QStringList QCommandLineParser::values(const QString &optionName) const
 */
 bool QCommandLineParser::isSet(const QCommandLineOption &option) const
 {
-    return isSet(option.names().first());
+    // option.names() might be empty if the constructor failed
+    const auto names = option.names();
+    return !names.isEmpty() && isSet(names.first());
 }
 
 /*!
@@ -738,7 +906,7 @@ bool QCommandLineParser::isSet(const QCommandLineOption &option) const
 */
 QString QCommandLineParser::value(const QCommandLineOption &option) const
 {
-    return value(option.names().first());
+    return value(option.names().constFirst());
 }
 
 /*!
@@ -756,7 +924,7 @@ QString QCommandLineParser::value(const QCommandLineOption &option) const
 */
 QStringList QCommandLineParser::values(const QCommandLineOption &option) const
 {
-    return values(option.names().first());
+    return values(option.names().constFirst());
 }
 
 /*!
@@ -814,6 +982,24 @@ QStringList QCommandLineParser::unknownOptionNames() const
     return d->unknownOptionNames;
 }
 
+/*!
+    Displays the version information from QCoreApplication::applicationVersion(),
+    and exits the application.
+    This is automatically triggered by the --version option, but can also
+    be used to display the version when not using process().
+    The exit code is set to EXIT_SUCCESS (0).
+
+    \sa addVersionOption()
+    \since 5.4
+*/
+Q_NORETURN void QCommandLineParser::showVersion()
+{
+    showParserMessage(QCoreApplication::applicationName() + QLatin1Char(' ')
+                      + QCoreApplication::applicationVersion() + QLatin1Char('\n'),
+                      UsageMessage);
+    ::exit(EXIT_SUCCESS);
+}
+
 /*!
     Displays the help information, and exits the application.
     This is automatically triggered by the --help option, but can also
@@ -827,7 +1013,7 @@ QStringList QCommandLineParser::unknownOptionNames() const
 */
 Q_NORETURN void QCommandLineParser::showHelp(int exitCode)
 {
-    fprintf(stdout, "%s", qPrintable(d->helpText()));
+    showParserMessage(d->helpText(), UsageMessage);
     ::exit(exitCode);
 }
 
@@ -844,12 +1030,51 @@ QString QCommandLineParser::helpText() const
 static QString wrapText(const QString &names, int longestOptionNameString, const QString &description)
 {
     const QLatin1Char nl('\n');
-    QString text = QStringLiteral("  ") + names.leftJustified(longestOptionNameString) + QLatin1Char(' ');
-    const int leftColumnWidth = text.length();
-    const int rightColumnWidth = 79 - leftColumnWidth;
-    text += description.left(rightColumnWidth) + nl;
-    for (int n = rightColumnWidth; n < description.length(); n += rightColumnWidth)
-        text += QStringLiteral(" ").repeated(leftColumnWidth) + description.mid(n, rightColumnWidth) + nl;
+    QString text = QLatin1String("  ") + names.leftJustified(longestOptionNameString) + QLatin1Char(' ');
+    const int indent = text.length();
+    int lineStart = 0;
+    int lastBreakable = -1;
+    const int max = 79 - indent;
+    int x = 0;
+    const int len = description.length();
+
+    for (int i = 0; i < len; ++i) {
+        ++x;
+        const QChar c = description.at(i);
+        if (c.isSpace())
+            lastBreakable = i;
+
+        int breakAt = -1;
+        int nextLineStart = -1;
+        if (x > max && lastBreakable != -1) {
+            // time to break and we know where
+            breakAt = lastBreakable;
+            nextLineStart = lastBreakable + 1;
+        } else if ((x > max - 1 && lastBreakable == -1) || i == len - 1) {
+            // time to break but found nowhere [-> break here], or end of last line
+            breakAt = i + 1;
+            nextLineStart = breakAt;
+        } else if (c == nl) {
+            // forced break
+            breakAt = i;
+            nextLineStart = i + 1;
+        }
+
+        if (breakAt != -1) {
+            const int numChars = breakAt - lineStart;
+            //qDebug() << "breakAt=" << description.at(breakAt) << "breakAtSpace=" << breakAtSpace << lineStart << "to" << breakAt << description.mid(lineStart, numChars);
+            if (lineStart > 0)
+                text += QString(indent, QLatin1Char(' '));
+            text += description.midRef(lineStart, numChars) + nl;
+            x = 0;
+            lastBreakable = -1;
+            lineStart = nextLineStart;
+            if (lineStart < len && description.at(lineStart).isSpace())
+                ++lineStart; // don't start a line with a space
+            i = lineStart;
+        }
+    }
+
     return text;
 }
 
@@ -857,16 +1082,12 @@ QString QCommandLineParserPrivate::helpText() const
 {
     const QLatin1Char nl('\n');
     QString text;
-    const QString exeName = QCoreApplication::instance()->arguments().first();
-    QString usage = exeName;
-    if (!commandLineOptionList.isEmpty()) {
-        usage += QLatin1Char(' ');
-        usage += QCommandLineParser::tr("[options]");
-    }
-    foreach (const PositionalArgumentDefinition &arg, positionalArgumentDefinitions) {
-        usage += QLatin1Char(' ');
-        usage += arg.syntax;
-    }
+    QString usage;
+    usage += QCoreApplication::instance()->arguments().constFirst(); // executable name
+    if (!commandLineOptionList.isEmpty())
+        usage += QLatin1Char(' ') + QCommandLineParser::tr("[options]");
+    for (const PositionalArgumentDefinition &arg : positionalArgumentDefinitions)
+        usage += QLatin1Char(' ') + arg.syntax;
     text += QCommandLineParser::tr("Usage: %1").arg(usage) + nl;
     if (!description.isEmpty())
        text += description + nl;
@@ -874,33 +1095,39 @@ QString QCommandLineParserPrivate::helpText() const
     if (!commandLineOptionList.isEmpty())
         text += QCommandLineParser::tr("Options:") + nl;
     QStringList optionNameList;
+    optionNameList.reserve(commandLineOptionList.size());
     int longestOptionNameString = 0;
-    foreach (const QCommandLineOption &option, commandLineOptionList) {
-        QStringList optionNames;
-        foreach (const QString &optionName, option.names()) {
-            if (optionName.length() == 1)
-                optionNames.append(QLatin1Char('-') + optionName);
-            else
-                optionNames.append(QStringLiteral("--") + optionName);
+    for (const QCommandLineOption &option : commandLineOptionList) {
+        if (option.isHidden())
+            continue;
+        const QStringList optionNames = option.names();
+        QString optionNamesString;
+        for (const QString &optionName : optionNames) {
+            const int numDashes = optionName.length() == 1 ? 1 : 2;
+            optionNamesString += QLatin1String("--", numDashes) + optionName + QLatin1String(", ");
         }
-        QString optionNamesString = optionNames.join(QStringLiteral(", "));
-        if (!option.valueName().isEmpty())
-            optionNamesString += QStringLiteral(" <") + option.valueName() + QLatin1Char('>');
+        if (!optionNames.isEmpty())
+            optionNamesString.chop(2); // remove trailing ", "
+        const auto valueName = option.valueName();
+        if (!valueName.isEmpty())
+            optionNamesString += QLatin1String(" <") + valueName + QLatin1Char('>');
         optionNameList.append(optionNamesString);
         longestOptionNameString = qMax(longestOptionNameString, optionNamesString.length());
     }
     ++longestOptionNameString;
-    for (int i = 0; i < commandLineOptionList.count(); ++i) {
-        const QCommandLineOption &option = commandLineOptionList.at(i);
-        text += wrapText(optionNameList.at(i), longestOptionNameString, option.description());
+    auto optionNameIterator = optionNameList.cbegin();
+    for (const QCommandLineOption &option : commandLineOptionList) {
+        if (option.isHidden())
+            continue;
+        text += wrapText(*optionNameIterator, longestOptionNameString, option.description());
+        ++optionNameIterator;
     }
     if (!positionalArgumentDefinitions.isEmpty()) {
         if (!commandLineOptionList.isEmpty())
             text += nl;
         text += QCommandLineParser::tr("Arguments:") + nl;
-        foreach (const PositionalArgumentDefinition &arg, positionalArgumentDefinitions) {
+        for (const PositionalArgumentDefinition &arg : positionalArgumentDefinitions)
             text += wrapText(arg.name, longestOptionNameString, arg.description);
-        }
     }
     return text;
 }
