@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtGui module of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -62,6 +60,8 @@
 #include <limits.h>
 
 #include <qpa/qplatformscreen.h>
+#include <qpa/qplatformintegration.h>
+#include <qpa/qplatformfontdatabase.h>
 #include <QtGui/private/qguiapplication_p.h>
 
 #include <QtCore/QMutexLocker>
@@ -118,9 +118,6 @@ bool QFontDef::exactMatch(const QFontDef &other) const
     QFontDatabase::parseFontName(family, this_foundry, this_family);
     QFontDatabase::parseFontName(other.family, other_foundry, other_family);
 
-    this_family = QFontDatabase::resolveFontFamilyAlias(this_family);
-    other_family = QFontDatabase::resolveFontFamilyAlias(other_family);
-
     return (styleHint     == other.styleHint
             && styleStrategy == other.styleStrategy
             && weight        == other.weight
@@ -137,7 +134,7 @@ extern bool qt_is_gui_used;
 
 Q_GUI_EXPORT int qt_defaultDpiX()
 {
-    if (qApp->testAttribute(Qt::AA_Use96Dpi))
+    if (QCoreApplication::instance()->testAttribute(Qt::AA_Use96Dpi))
         return 96;
 
     if (!qt_is_gui_used)
@@ -152,7 +149,7 @@ Q_GUI_EXPORT int qt_defaultDpiX()
 
 Q_GUI_EXPORT int qt_defaultDpiY()
 {
-    if (qApp->testAttribute(Qt::AA_Use96Dpi))
+    if (QCoreApplication::instance()->testAttribute(Qt::AA_Use96Dpi))
         return 96;
 
     if (!qt_is_gui_used)
@@ -172,14 +169,14 @@ Q_GUI_EXPORT int qt_defaultDpi()
 
 QFontPrivate::QFontPrivate()
     : engineData(0), dpi(qt_defaultDpi()), screen(0),
-      rawMode(false), underline(false), overline(false), strikeOut(false), kerning(true),
+      underline(false), overline(false), strikeOut(false), kerning(true),
       capital(0), letterSpacingIsAbsolute(false), scFont(0)
 {
 }
 
 QFontPrivate::QFontPrivate(const QFontPrivate &other)
     : request(other.request), engineData(0), dpi(other.dpi), screen(other.screen),
-      rawMode(other.rawMode), underline(other.underline), overline(other.overline),
+      underline(other.underline), overline(other.overline),
       strikeOut(other.strikeOut), kerning(other.kerning),
       capital(other.capital), letterSpacingIsAbsolute(other.letterSpacingIsAbsolute),
       letterSpacing(other.letterSpacing), wordSpacing(other.wordSpacing),
@@ -208,7 +205,7 @@ QFontEngine *QFontPrivate::engineForScript(int script) const
     QMutexLocker locker(qt_fontdatabase_mutex());
     if (script <= QChar::Script_Latin)
         script = QChar::Script_Common;
-    if (engineData && engineData->fontCache != QFontCache::instance()) {
+    if (engineData && engineData->fontCacheId != QFontCache::instance()->id()) {
         // throw out engineData that came from a different thread
         if (!engineData->ref.deref())
             delete engineData;
@@ -317,7 +314,7 @@ void QFontPrivate::resolve(uint mask, const QFontPrivate *other)
 
 
 QFontEngineData::QFontEngineData()
-    : ref(0), fontCache(QFontCache::instance())
+    : ref(0), fontCacheId(QFontCache::instance()->id())
 {
     memset(engines, 0, QChar::ScriptCount * sizeof(QFontEngine *));
 }
@@ -412,9 +409,6 @@ QFontEngineData::~QFontEngineData()
     setPointSize() has a similar effect and provides device
     independence.
 
-    In X11 you can set a font using its system
-    specific name with setRawName().
-
     Loading fonts can be expensive, especially on X11. QFont contains
     extensive optimizations to make the copying of QFont objects fast,
     and to cache the results of the slow window system functions it
@@ -544,77 +538,6 @@ QFontEngineData::~QFontEngineData()
     \since 5.2
 */
 
-/*!
-    \fn QString QFont::rawName() const
-
-    Returns the name of the font within the underlying window system.
-
-    On X11, this function will return an empty string.
-
-    Using the return value of this function is usually \e not \e
-    portable.
-
-    \sa setRawName()
-*/
-
-/*!
-    \fn void QFont::setRawName(const QString &name)
-
-    Sets a font by its system specific name.
-
-    A font set with setRawName() is still a full-featured QFont. It can
-    be queried (for example with italic()) or modified (for example with
-    setItalic()) and is therefore also suitable for rendering rich text.
-
-    If Qt's internal font database cannot resolve the raw name, the
-    font becomes a raw font with \a name as its family.
-
-    \sa rawName(), setRawMode(), setFamily()
-*/
-
-/*!
-    \fn QString QFont::lastResortFamily() const
-
-    Returns the "last resort" font family name.
-
-    The current implementation tries a wide variety of common fonts,
-    returning the first one it finds. Is is possible that no family is
-    found in which case an empty string is returned.
-
-    \sa lastResortFont()
-*/
-
-/*!
-    \fn QString QFont::defaultFamily() const
-
-    Returns the family name that corresponds to the current style
-    hint.
-
-    \sa StyleHint, styleHint(), setStyleHint()
-*/
-
-/*!
-    \fn QString QFont::lastResortFont() const
-
-    Returns a "last resort" font name for the font matching algorithm.
-    This is used if the last resort family is not available. It will
-    always return a name, if necessary returning something like
-    "fixed" or "system".
-
-    The current implementation tries a wide variety of common fonts,
-    returning the first one it finds. The implementation may change
-    at any time, but this function will always return a string
-    containing something.
-
-    It is theoretically possible that there really isn't a
-    lastResortFont() in which case Qt will abort with an error
-    message. We have not been able to identify a case where this
-    happens. Please \l{bughowto.html}{report it as a bug} if
-    it does, preferably with a list of the fonts you have installed.
-
-    \sa lastResortFamily(), rawName()
-*/
-
 /*!
   Constructs a font from \a font for use on the paint device \a pd.
 */
@@ -629,7 +552,7 @@ QFont::QFont(const QFont &font, QPaintDevice *pd)
         d->dpi = dpi;
         d->screen = screen;
     } else {
-        d = font.d.data();
+        d = font.d;
     }
 }
 
@@ -734,7 +657,7 @@ QFont::QFont(const QString &family, int pointSize, int weight, bool italic)
     Constructs a font that is a copy of \a font.
 */
 QFont::QFont(const QFont &font)
-    : d(font.d.data()), resolve_mask(font.resolve_mask)
+    : d(font.d), resolve_mask(font.resolve_mask)
 {
 }
 
@@ -750,7 +673,7 @@ QFont::~QFont()
 */
 QFont &QFont::operator=(const QFont &font)
 {
-    d = font.d.data();
+    d = font.d;
     resolve_mask = font.resolve_mask;
     return *this;
 }
@@ -805,7 +728,7 @@ void QFont::setFamily(const QString &family)
     Returns the requested font style name, it will be used to match the
     font with irregular styles (that can't be normalized in other style
     properties). It depends on system font support, thus only works for
-    Mac OS X and X11 so far. On Windows irregular styles will be added
+    OS X and X11 so far. On Windows irregular styles will be added
     as separate font families so there is no need for this.
 
     \sa setFamily(), setStyle()
@@ -900,7 +823,7 @@ int QFont::pointSize() const
     \li Vertical hinting (light)
     \li Full hinting
     \row
-    \li Cocoa on Mac OS X
+    \li Cocoa on OS X
     \li No hinting
     \li No hinting
     \li No hinting
@@ -1091,8 +1014,8 @@ void QFont::setStyle(Style style)
 }
 
 /*!
-    Returns the weight of the font which is one of the enumerated
-    values from \l{QFont::Weight}.
+    Returns the weight of the font, using the same scale as the
+    \l{QFont::Weight} enumeration.
 
     \sa setWeight(), Weight, QFontInfo
 */
@@ -1105,21 +1028,25 @@ int QFont::weight() const
     \enum QFont::Weight
 
     Qt uses a weighting scale from 0 to 99 similar to, but not the
-    same as, the scales used in Windows or CSS. A weight of 0 is
-    ultralight, whilst 99 will be an extremely black.
+    same as, the scales used in Windows or CSS. A weight of 0 will be
+    thin, whilst 99 will be extremely black.
 
     This enum contains the predefined font weights:
 
+    \value Thin 0
+    \value ExtraLight 12
     \value Light 25
     \value Normal 50
+    \value Medium 57
     \value DemiBold 63
     \value Bold 75
+    \value ExtraBold 81
     \value Black 87
 */
 
 /*!
-    Sets the weight the font to \a weight, which should be a value
-    from the \l QFont::Weight enumeration.
+    Sets the weight of the font to \a weight, using the scale defined by
+    \l QFont::Weight enumeration.
 
     \sa weight(), QFontInfo
 */
@@ -1140,7 +1067,7 @@ void QFont::setWeight(int weight)
     \fn bool QFont::bold() const
 
     Returns \c true if weight() is a value greater than
-   \l{Weight}{QFont::Normal}; otherwise returns \c false.
+   \l{Weight}{QFont::Medium}; otherwise returns \c false.
 
     \sa weight(), setBold(), QFontInfo::bold()
 */
@@ -1282,7 +1209,7 @@ bool QFont::kerning() const
     When kerning is enabled, glyph metrics do not add up anymore,
     even for Latin text. In other words, the assumption that
     width('a') + width('b') is equal to width("ab") is not
-    neccesairly true.
+    necessarily true.
 
     \sa kerning(), QFontMetrics
 */
@@ -1374,6 +1301,7 @@ QFont::StyleHint QFont::styleHint() const
     \value PreferOutline prefers outline fonts (as opposed to bitmap fonts).
     \value ForceOutline forces the use of outline fonts.
     \value NoAntialias don't antialias the fonts.
+    \value NoSubpixelAntialias avoid subpixel antialiasing on the fonts if possible.
     \value PreferAntialias antialias if possible.
     \value OpenGLCompatible forces the use of OpenGL compatible
            fonts.
@@ -1528,7 +1456,8 @@ qreal QFont::letterSpacing() const
 
     Letter spacing changes the default spacing between individual
     letters in the font.  The spacing between the letters can be
-    made smaller as well as larger.
+    made smaller as well as larger either in percentage of the
+    character width or in pixels, depending on the selected spacing type.
 
     \sa letterSpacing(), letterSpacingType(), setWordSpacing()
 */
@@ -1642,8 +1571,11 @@ QFont::Capitalization QFont::capitalization() const
     return static_cast<QFont::Capitalization> (d->capital);
 }
 
-
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
+    \fn void QFont::setRawMode(bool enable)
+    \deprecated
+
     If \a enable is true, turns raw mode on; otherwise turns raw mode
     off. This function only has an effect under X11.
 
@@ -1655,20 +1587,14 @@ QFont::Capitalization QFont::capitalization() const
     return the values set in the QFont for all parameters, including
     the family name).
 
-    \warning Do not use raw mode unless you really, really need it! In
-    most (if not all) cases, setRawName() is a much better choice.
+    \warning Enabling raw mode has no effect since Qt 5.0.
 
-    \sa rawMode(), setRawName()
+    \sa rawMode()
 */
-void QFont::setRawMode(bool enable)
+void QFont::setRawMode(bool)
 {
-    if ((bool) d->rawMode == enable) return;
-
-    // might change behavior, thus destroy engine data
-    detach();
-
-    d->rawMode = enable;
 }
+#endif
 
 /*!
     Returns \c true if a window system font exactly matching the settings
@@ -1680,9 +1606,7 @@ bool QFont::exactMatch() const
 {
     QFontEngine *engine = d->engineForScript(QChar::Script_Common);
     Q_ASSERT(engine != 0);
-    return (d->rawMode
-            ? engine->type() != QFontEngine::Box
-            : d->request.exactMatch(engine->fontDef));
+    return d->request.exactMatch(engine->fontDef);
 }
 
 /*!
@@ -1690,8 +1614,7 @@ bool QFont::exactMatch() const
     false.
 
     Two QFonts are considered equal if their font attributes are
-    equal. If rawMode() is enabled for both fonts, only the family
-    fields are compared.
+    equal.
 
     \sa operator!=(), isCopyOf()
 */
@@ -1727,8 +1650,8 @@ bool QFont::operator<(const QFont &f) const
 {
     if (f.d == d) return false;
     // the < operator for fontdefs ignores point sizes.
-    QFontDef &r1 = f.d->request;
-    QFontDef &r2 = d->request;
+    const QFontDef &r1 = f.d->request;
+    const QFontDef &r2 = d->request;
     if (r1.pointSize != r2.pointSize) return r1.pointSize < r2.pointSize;
     if (r1.pixelSize != r2.pixelSize) return r1.pixelSize < r2.pixelSize;
     if (r1.weight != r2.weight) return r1.weight < r2.weight;
@@ -1754,8 +1677,7 @@ bool QFont::operator<(const QFont &f) const
     returns \c false.
 
     Two QFonts are considered to be different if their font attributes
-    are different. If rawMode() is enabled for both fonts, only the
-    family fields are compared.
+    are different.
 
     \sa operator==()
 */
@@ -1784,16 +1706,20 @@ bool QFont::isCopyOf(const QFont & f) const
     return d == f.d;
 }
 
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
+    \deprecated
+
     Returns \c true if raw mode is used for font name matching; otherwise
     returns \c false.
 
-    \sa setRawMode(), rawName()
+    \sa setRawMode()
 */
 bool QFont::rawMode() const
 {
-    return d->rawMode;
+    return false;
 }
+#endif
 
 /*!
     Returns a new QFont that has attributes copied from \a other that
@@ -1801,10 +1727,8 @@ bool QFont::rawMode() const
 */
 QFont QFont::resolve(const QFont &other) const
 {
-    if (*this == other
-        && (resolve_mask == other.resolve_mask || resolve_mask == 0)
-        && d->dpi == other.d->dpi) {
-        QFont o = other;
+    if (resolve_mask == 0 || (resolve_mask == other.resolve_mask && *this == other)) {
+        QFont o(other);
         o.resolve_mask = resolve_mask;
         return o;
     }
@@ -1904,27 +1828,13 @@ void QFont::insertSubstitutions(const QString &familyName,
     QFontSubst *fontSubst = globalFontSubst();
     Q_ASSERT(fontSubst != 0);
     QStringList &list = (*fontSubst)[familyName.toLower()];
-    foreach (const QString &substituteName, substituteNames) {
+    for (const QString &substituteName : substituteNames) {
         const QString lowerSubstituteName = substituteName.toLower();
         if (!list.contains(lowerSubstituteName))
             list.append(lowerSubstituteName);
     }
 }
 
-/*! \fn void QFont::initialize()
-  \internal
-
-  Internal function that initializes the font system.  The font cache
-  and font dict do not alloc the keys. The key is a QString which is
-  shared between QFontPrivate and QXFontName.
-*/
-
-/*! \fn void QFont::cleanup()
-  \internal
-
-  Internal function that cleans up the font system.
-*/
-
 /*!
     Removes all the substitutions for \a familyName.
 
@@ -1953,20 +1863,15 @@ void QFont::removeSubstitutions(const QString &familyName)
 */
 QStringList QFont::substitutions()
 {
-    typedef QFontSubst::const_iterator QFontSubstConstIterator;
-
     QFontSubst *fontSubst = globalFontSubst();
     Q_ASSERT(fontSubst != 0);
-    QStringList ret;
-    const QFontSubstConstIterator cend = fontSubst->constEnd();
-    for (QFontSubstConstIterator it = fontSubst->constBegin(); it != cend; ++it)
-        ret.append(it.key());
+    QStringList ret = fontSubst->keys();
 
     ret.sort();
     return ret;
 }
 
-
+#ifndef QT_NO_DATASTREAM
 /*  \internal
     Internal function. Converts boolean font settings to an unsigned
     8-bit number. Used for serialization etc.
@@ -1987,8 +1892,6 @@ static quint8 get_font_bits(int version, const QFontPrivate *f)
         bits |= 0x08;
     // if (f.hintSetByUser)
     // bits |= 0x10;
-    if (f->rawMode)
-        bits |= 0x20;
     if (version >= QDataStream::Qt_4_0) {
         if (f->kerning)
             bits |= 0x10;
@@ -2009,8 +1912,6 @@ static quint8 get_extended_font_bits(const QFontPrivate *f)
     return bits;
 }
 
-#ifndef QT_NO_DATASTREAM
-
 /*  \internal
     Internal function. Sets boolean font settings from an unsigned
     8-bit number. Used for serialization etc.
@@ -2024,7 +1925,6 @@ static void set_font_bits(int version, quint8 bits, QFontPrivate *f)
     f->strikeOut             = (bits & 0x04) != 0;
     f->request.fixedPitch    = (bits & 0x08) != 0;
     // f->hintSetByUser      = (bits & 0x10) != 0;
-    f->rawMode               = (bits & 0x20) != 0;
     if (version >= QDataStream::Qt_4_0)
         f->kerning               = (bits & 0x10) != 0;
     if ((bits & 0x80) != 0)
@@ -2039,6 +1939,44 @@ static void set_extended_font_bits(quint8 bits, QFontPrivate *f)
 }
 #endif
 
+#if QT_DEPRECATED_SINCE(5, 3)
+/*!
+    \fn QString QFont::rawName() const
+    \deprecated
+
+    Returns the name of the font within the underlying window system.
+
+    On X11, this function will return an empty string.
+
+    Using the return value of this function is usually \e not \e
+    portable.
+
+    \sa setRawName()
+*/
+QString QFont::rawName() const
+{
+    return QLatin1String("unknown");
+}
+
+/*!
+    \fn void QFont::setRawName(const QString &name)
+    \deprecated
+
+    Sets a font by its system specific name.
+
+    A font set with setRawName() is still a full-featured QFont. It can
+    be queried (for example with italic()) or modified (for example with
+    setItalic()) and is therefore also suitable for rendering rich text.
+
+    If Qt's internal font database cannot resolve the raw name, the
+    font becomes a raw font with \a name as its family.
+
+    \sa rawName(), setFamily()
+*/
+void QFont::setRawName(const QString &)
+{
+}
+#endif
 
 /*!
     Returns the font's key, a textual representation of a font. It is
@@ -2070,7 +2008,19 @@ QString QFont::toString() const
         QString::number((int) underline()) + comma +
         QString::number((int) strikeOut()) + comma +
         QString::number((int)fixedPitch()) + comma +
-        QString::number((int)   rawMode());
+        QString::number((int)   false);
+}
+
+/*!
+    Returns the hash value for \a font. If specified, \a seed is used
+    to initialize the hash.
+
+    \relates QFont
+    \since 5.3
+*/
+uint qHash(const QFont &font, uint seed) Q_DECL_NOTHROW
+{
+    return qHash(QFontPrivate::get(font)->request, seed);
 }
 
 
@@ -2102,7 +2052,6 @@ bool QFont::fromString(const QString &descrip)
         setUnderline(l[5].toInt());
         setStrikeOut(l[6].toInt());
         setFixedPitch(l[7].toInt());
-        setRawMode(l[8].toInt());
     } else if (count == 10) {
         if (l[2].toInt() > 0)
             setPixelSize(l[2].toInt());
@@ -2112,7 +2061,6 @@ bool QFont::fromString(const QString &descrip)
         setUnderline(l[6].toInt());
         setStrikeOut(l[7].toInt());
         setFixedPitch(l[8].toInt());
-        setRawMode(l[9].toInt());
     }
     if (count >= 9 && !d->request.fixedPitch) // assume 'false' fixedPitch equals default
         d->request.ignorePitch = true;
@@ -2120,17 +2068,98 @@ bool QFont::fromString(const QString &descrip)
     return true;
 }
 
+/*! \fn void QFont::initialize()
+  \internal
+
+  Internal function that initializes the font system.  The font cache
+  and font dict do not alloc the keys. The key is a QString which is
+  shared between QFontPrivate and QXFontName.
+*/
+void QFont::initialize()
+{
+}
+
+/*! \fn void QFont::cleanup()
+  \internal
+
+  Internal function that cleans up the font system.
+*/
+void QFont::cleanup()
+{
+    QFontCache::cleanup();
+}
+
 /*! \internal
 
   Internal function that dumps font cache statistics.
 */
 void QFont::cacheStatistics()
 {
-
-
 }
 
+/*!
+    \fn QString QFont::lastResortFamily() const
 
+    Returns the "last resort" font family name.
+
+    The current implementation tries a wide variety of common fonts,
+    returning the first one it finds. Is is possible that no family is
+    found in which case an empty string is returned.
+
+    \sa lastResortFont()
+*/
+QString QFont::lastResortFamily() const
+{
+    return QString::fromLatin1("helvetica");
+}
+
+extern QStringList qt_fallbacksForFamily(const QString &family, QFont::Style style,
+                                         QFont::StyleHint styleHint, QChar::Script script);
+
+/*!
+    \fn QString QFont::defaultFamily() const
+
+    Returns the family name that corresponds to the current style
+    hint.
+
+    \sa StyleHint, styleHint(), setStyleHint()
+*/
+QString QFont::defaultFamily() const
+{
+    const QStringList fallbacks = qt_fallbacksForFamily(QString(), QFont::StyleNormal
+                                      , QFont::StyleHint(d->request.styleHint), QChar::Script_Common);
+    if (!fallbacks.isEmpty())
+        return fallbacks.first();
+    return QString();
+}
+
+/*!
+    \fn QString QFont::lastResortFont() const
+
+    Returns a "last resort" font name for the font matching algorithm.
+    This is used if the last resort family is not available. It will
+    always return a name, if necessary returning something like
+    "fixed" or "system".
+
+    The current implementation tries a wide variety of common fonts,
+    returning the first one it finds. The implementation may change
+    at any time, but this function will always return a string
+    containing something.
+
+    It is theoretically possible that there really isn't a
+    lastResortFont() in which case Qt will abort with an error
+    message. We have not been able to identify a case where this
+    happens. Please \l{bughowto.html}{report it as a bug} if
+    it does, preferably with a list of the fonts you have installed.
+
+    \sa lastResortFamily()
+*/
+QString QFont::lastResortFont() const
+{
+    qFatal("QFont::lastResortFont: Cannot find any reasonable font");
+    // Shut compiler up
+    return QString();
+}
 
 /*****************************************************************************
   QFont stream functions
@@ -2151,6 +2180,8 @@ QDataStream &operator<<(QDataStream &s, const QFont &font)
         s << font.d->request.family.toLatin1();
     } else {
         s << font.d->request.family;
+        if (s.version() >= QDataStream::Qt_5_4)
+            s << font.d->request.styleName;
     }
 
     if (s.version() >= QDataStream::Qt_4_0) {
@@ -2171,8 +2202,14 @@ QDataStream &operator<<(QDataStream &s, const QFont &font)
     }
 
     s << (quint8) font.d->request.styleHint;
-    if (s.version() >= QDataStream::Qt_3_1)
-        s << (quint8) font.d->request.styleStrategy;
+    if (s.version() >= QDataStream::Qt_3_1) {
+        // Continue writing 8 bits for versions < 5.4 so that we don't write too much,
+        // even though we need 16 to store styleStrategy, so there is some data loss.
+        if (s.version() >= QDataStream::Qt_5_4)
+            s << (quint16) font.d->request.styleStrategy;
+        else
+            s << (quint8) font.d->request.styleStrategy;
+    }
     s << (quint8) 0
       << (quint8) font.d->request.weight
       << get_font_bits(s.version(), font.d.data());
@@ -2184,6 +2221,10 @@ QDataStream &operator<<(QDataStream &s, const QFont &font)
         s << font.d->letterSpacing.value();
         s << font.d->wordSpacing.value();
     }
+    if (s.version() >= QDataStream::Qt_5_4)
+        s << (quint8)font.d->request.hintingPreference;
+    if (s.version() >= QDataStream::Qt_5_6)
+        s << (quint8)font.d->capital;
     return s;
 }
 
@@ -2201,7 +2242,8 @@ QDataStream &operator>>(QDataStream &s, QFont &font)
     font.d = new QFontPrivate;
     font.resolve_mask = QFont::AllPropertiesResolved;
 
-    quint8 styleHint, styleStrategy = QFont::PreferDefault, charSet, weight, bits;
+    quint8 styleHint, charSet, weight, bits;
+    quint16 styleStrategy = QFont::PreferDefault;
 
     if (s.version() == 1) {
         QByteArray fam;
@@ -2209,6 +2251,8 @@ QDataStream &operator>>(QDataStream &s, QFont &font)
         font.d->request.family = QString::fromLatin1(fam);
     } else {
         s >> font.d->request.family;
+        if (s.version() >= QDataStream::Qt_5_4)
+            s >> font.d->request.styleName;
     }
 
     if (s.version() >= QDataStream::Qt_4_0) {
@@ -2228,8 +2272,15 @@ QDataStream &operator>>(QDataStream &s, QFont &font)
         font.d->request.pixelSize = pixelSize;
     }
     s >> styleHint;
-    if (s.version() >= QDataStream::Qt_3_1)
-        s >> styleStrategy;
+    if (s.version() >= QDataStream::Qt_3_1) {
+        if (s.version() >= QDataStream::Qt_5_4) {
+            s >> styleStrategy;
+        } else {
+            quint8 tempStyleStrategy;
+            s >> tempStyleStrategy;
+            styleStrategy = tempStyleStrategy;
+        }
+    }
 
     s >> charSet;
     s >> weight;
@@ -2259,7 +2310,16 @@ QDataStream &operator>>(QDataStream &s, QFont &font)
         s >> value;
         font.d->wordSpacing.setValue(value);
     }
-
+    if (s.version() >= QDataStream::Qt_5_4) {
+        quint8 value;
+        s >> value;
+        font.d->request.hintingPreference = QFont::HintingPreference(value);
+    }
+    if (s.version() >= QDataStream::Qt_5_6) {
+        quint8 value;
+        s >> value;
+        font.d->capital = QFont::Capitalization(value);
+    }
     return s;
 }
 
@@ -2331,7 +2391,7 @@ QDataStream &operator>>(QDataStream &s, QFont &font)
     that is not screen-compatible.
 */
 QFontInfo::QFontInfo(const QFont &font)
-    : d(font.d.data())
+    : d(font.d)
 {
 }
 
@@ -2339,7 +2399,7 @@ QFontInfo::QFontInfo(const QFont &font)
     Constructs a copy of \a fi.
 */
 QFontInfo::QFontInfo(const QFontInfo &fi)
-    : d(fi.d.data())
+    : d(fi.d)
 {
 }
 
@@ -2355,7 +2415,7 @@ QFontInfo::~QFontInfo()
 */
 QFontInfo &QFontInfo::operator=(const QFontInfo &fi)
 {
-    d = fi.d.data();
+    d = fi.d;
     return *this;
 }
 
@@ -2383,7 +2443,7 @@ QString QFontInfo::family() const
     \since 4.8
 
     Returns the style name of the matched window system font on
-    system that supports it.
+    systems that support it.
 
     \sa QFont::styleName()
 */
@@ -2533,8 +2593,10 @@ bool QFontInfo::fixedPitch() const
         QChar ch[2] = { QLatin1Char('i'), QLatin1Char('m') };
         QGlyphLayoutArray<2> g;
         int l = 2;
-        engine->stringToCMap(ch, 2, &g, &l, 0);
-        engine->fontDef.fixedPitch = g.advances_x[0] == g.advances_x[1];
+        if (!engine->stringToCMap(ch, 2, &g, &l, 0))
+            Q_UNREACHABLE();
+        Q_ASSERT(l == 2);
+        engine->fontDef.fixedPitch = g.advances[0] == g.advances[1];
         engine->fontDef.fixedPitchComputed = true;
     }
 #endif
@@ -2555,7 +2617,10 @@ QFont::StyleHint QFontInfo::styleHint() const
     return (QFont::StyleHint) engine->fontDef.styleHint;
 }
 
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
+    \deprecated
+
     Returns \c true if the font is a raw mode font; otherwise returns
     false.
 
@@ -2567,8 +2632,9 @@ QFont::StyleHint QFontInfo::styleHint() const
 */
 bool QFontInfo::rawMode() const
 {
-    return d->rawMode;
+    return false;
 }
+#endif
 
 /*!
     Returns \c true if the matched window system font is exactly the same
@@ -2580,9 +2646,7 @@ bool QFontInfo::exactMatch() const
 {
     QFontEngine *engine = d->engineForScript(QChar::Script_Common);
     Q_ASSERT(engine != 0);
-    return (d->rawMode
-            ? engine->type() != QFontEngine::Box
-            : d->request.exactMatch(engine->fontDef));
+    return d->request.exactMatch(engine->fontDef);
 }
 
 
@@ -2638,9 +2702,12 @@ void QFontCache::cleanup()
 }
 #endif // QT_NO_THREAD
 
+QBasicAtomicInt font_cache_id = Q_BASIC_ATOMIC_INITIALIZER(1);
+
 QFontCache::QFontCache()
     : QObject(), total_cost(0), max_cost(min_cost),
-      current_timestamp(0), fast(false), timer_id(-1)
+      current_timestamp(0), fast(false), timer_id(-1),
+      m_id(font_cache_id.fetchAndAddRelaxed(1))
 {
 }
 
@@ -2732,6 +2799,10 @@ void QFontCache::insertEngineData(const QFontDef &def, QFontEngineData *engineDa
     Q_ASSERT(!engineDataCache.contains(def));
 
     engineData->ref.ref();
+    // Decrease now rather than waiting
+    if (total_cost > min_cost * 2)
+        decreaseCache();
+
     engineDataCache.insert(def, engineData);
     increaseCost(sizeof(QFontEngineData));
 }
@@ -2741,6 +2812,10 @@ QFontEngine *QFontCache::findEngine(const Key &key)
     EngineCache::Iterator it = engineCache.find(key),
                          end = engineCache.end();
     if (it == end) return 0;
+
+    Q_ASSERT(it.value().data != Q_NULLPTR);
+    Q_ASSERT(key.multi == (it.value().data->type() == QFontEngine::Multi));
+
     // found... update the hitcount and timestamp
     updateHitCountAndTimeStamp(it.value());
 
@@ -2753,14 +2828,17 @@ void QFontCache::updateHitCountAndTimeStamp(Engine &value)
     value.timestamp = ++current_timestamp;
 
     FC_DEBUG("QFontCache: found font engine\n"
-             "  %p: timestamp %4u hits %3u ref %2d/%2d, type '%s'",
+             "  %p: timestamp %4u hits %3u ref %2d/%2d, type %d",
              value.data, value.timestamp, value.hits,
              value.data->ref.load(), engineCacheCount.value(value.data),
-             value.data->name());
+             value.data->type());
 }
 
 void QFontCache::insertEngine(const Key &key, QFontEngine *engine, bool insertMulti)
 {
+    Q_ASSERT(engine != Q_NULLPTR);
+    Q_ASSERT(key.multi == (engine->type() == QFontEngine::Multi));
+
 #ifdef QFONTCACHE_DEBUG
     FC_DEBUG("QFontCache: inserting new engine %p, refcount %d", engine, engine->ref.load());
     if (!insertMulti && engineCache.contains(key)) {
@@ -2769,8 +2847,10 @@ void QFontCache::insertEngine(const Key &key, QFontEngine *engine, bool insertMu
                  key.def.pixelSize, key.def.weight, key.def.style, key.def.fixedPitch);
     }
 #endif
-
     engine->ref.ref();
+    // Decrease now rather than waiting
+    if (total_cost > min_cost * 2)
+        decreaseCache();
 
     Engine data(engine);
     data.timestamp = ++current_timestamp;
@@ -2831,7 +2911,11 @@ void QFontCache::timerEvent(QTimerEvent *)
 
         return;
     }
+    decreaseCache();
+}
 
+void QFontCache::decreaseCache()
+{
     // go through the cache and count up everything in use
     uint in_use_cost = 0;
 
@@ -2952,10 +3036,10 @@ void QFontCache::timerEvent(QTimerEvent *)
 
         it = jt;
         if (it != end) {
-            FC_DEBUG("    %p: timestamp %4u hits %2u ref %2d/%2d, type '%s'",
+            FC_DEBUG("    %p: timestamp %4u hits %2u ref %2d/%2d, type %d",
                      it.value().data, it.value().timestamp, it.value().hits,
                      it.value().data->ref.load(), engineCacheCount.value(it.value().data),
-                     it.value().data->name());
+                     it.value().data->type());
 
             QFontEngine *fontEngine = it.value().data;
             // get rid of all occurrences
