@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtGui module of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -43,22 +41,21 @@
 
 #include <qplatformdefs.h>
 
+#include <QFile>
 #include <QSocketNotifier>
 #include <QStringList>
-#include <qpa/qwindowsysteminterface.h>
 #include <QCoreApplication>
+#include <QLoggingCategory>
+#include <qpa/qwindowsysteminterface.h>
 #include <private/qcore_unix_p.h>
 
 #include <linux/input.h>
 
-//#define QT_QPA_KEYMAP_DEBUG
-
-#ifdef QT_QPA_KEYMAP_DEBUG
-#include <qdebug.h>
-#endif
-
 QT_BEGIN_NAMESPACE
 
+Q_LOGGING_CATEGORY(qLcEvdevKey, "qt.qpa.input")
+Q_LOGGING_CATEGORY(qLcEvdevKeyMap, "qt.qpa.input.keymap")
+
 // simple builtin US keymap
 #include "qevdevkeyboard_defaultmap_p.h"
 
@@ -68,9 +65,7 @@ QEvdevKeyboardHandler::QEvdevKeyboardHandler(const QString &device, int fd, bool
       m_no_zap(disableZap), m_do_compose(enableCompose),
       m_keymap(0), m_keymap_size(0), m_keycompose(0), m_keycompose_size(0)
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Create keyboard handler with for device" << device;
-#endif
+    qCDebug(qLcEvdevKey) << "Create keyboard handler with for device" << device;
 
     setObjectName(QLatin1String("LinuxInput Keyboard Handler"));
 
@@ -93,23 +88,23 @@ QEvdevKeyboardHandler::~QEvdevKeyboardHandler()
         qt_safe_close(m_fd);
 }
 
-QEvdevKeyboardHandler *QEvdevKeyboardHandler::create(const QString &device, const QString &specification)
+QEvdevKeyboardHandler *QEvdevKeyboardHandler::create(const QString &device,
+                                                     const QString &specification,
+                                                     const QString &defaultKeymapFile)
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Try to create keyboard handler for" << device << specification;
-#endif
+    qCDebug(qLcEvdevKey) << "Try to create keyboard handler for" << device << specification;
 
-    QString keymapFile;
+    QString keymapFile = defaultKeymapFile;
     int repeatDelay = 400;
     int repeatRate = 80;
     bool disableZap = false;
     bool enableCompose = false;
     int grab = 0;
 
-    QStringList args = specification.split(QLatin1Char(':'));
-    foreach (const QString &arg, args) {
+    const auto args = specification.splitRef(QLatin1Char(':'));
+    for (const QStringRef &arg : args) {
         if (arg.startsWith(QLatin1String("keymap=")))
-            keymapFile = arg.mid(7);
+            keymapFile = arg.mid(7).toString();
         else if (arg == QLatin1String("disable-zap"))
             disableZap = true;
         else if (arg == QLatin1String("enable-compose"))
@@ -122,9 +117,7 @@ QEvdevKeyboardHandler *QEvdevKeyboardHandler::create(const QString &device, cons
             grab = arg.mid(5).toInt();
     }
 
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Opening keyboard at" << device;
-#endif
+    qCDebug(qLcEvdevKey) << "Opening keyboard at" << device;
 
     int fd;
     fd = qt_safe_open(device.toLocal8Bit().constData(), O_RDONLY | O_NDELAY, 0);
@@ -144,9 +137,7 @@ QEvdevKeyboardHandler *QEvdevKeyboardHandler::create(const QString &device, cons
 
 void QEvdevKeyboardHandler::switchLed(int led, bool state)
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "switchLed" << led << state;
-#endif
+    qCDebug(qLcEvdevKey) << "switchLed" << led << state;
 
     struct ::input_event led_ie;
     ::gettimeofday(&led_ie.time, 0);
@@ -159,10 +150,6 @@ void QEvdevKeyboardHandler::switchLed(int led, bool state)
 
 void QEvdevKeyboardHandler::readKeycode()
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Read new keycode on" << m_device;
-#endif
-
     struct ::input_event buffer[32];
     int n = 0;
 
@@ -170,11 +157,11 @@ void QEvdevKeyboardHandler::readKeycode()
         int result = qt_safe_read(m_fd, reinterpret_cast<char *>(buffer) + n, sizeof(buffer) - n);
 
         if (result == 0) {
-            qWarning("Got EOF from the input device.");
+            qWarning("evdevkeyboard: Got EOF from the input device");
             return;
         } else if (result < 0) {
             if (errno != EINTR && errno != EAGAIN) {
-                qWarning("Could not read from input device: %s", strerror(errno));
+                qErrnoWarning(errno, "evdevkeyboard: Could not read from input device");
                 return;
             }
         } else {
@@ -224,7 +211,7 @@ void QEvdevKeyboardHandler::processKeyEvent(int nativecode, int unicode, int qtc
 {
     QWindowSystemInterface::handleExtendedKeyEvent(0, (isPress ? QEvent::KeyPress : QEvent::KeyRelease),
                                                    qtcode, modifiers, nativecode + 8, 0, int(modifiers),
-                                                   QString(unicode), autoRepeat);
+                                                   (unicode != 0xffff ) ? QString(unicode) : QString(), autoRepeat);
 }
 
 QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint16 keycode, bool pressed, bool autorepeat)
@@ -235,6 +222,8 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
     const QEvdevKeyboardMap::Mapping *map_plain = 0;
     const QEvdevKeyboardMap::Mapping *map_withmod = 0;
 
+    quint8 modifiers = m_modifiers;
+
     // get a specific and plain mapping for the keycode and the current modifiers
     for (int i = 0; i < m_keymap_size && !(map_plain && map_withmod); ++i) {
         const QEvdevKeyboardMap::Mapping *m = m_keymap + i;
@@ -250,21 +239,20 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
         }
     }
 
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning("Processing key event: keycode=%3d, modifiers=%02x pressed=%d, autorepeat=%d  |  plain=%d, withmod=%d, size=%d", \
-             keycode, m_modifiers, pressed ? 1 : 0, autorepeat ? 1 : 0, \
-             map_plain ? map_plain - m_keymap : -1, \
-             map_withmod ? map_withmod - m_keymap : -1, \
-             m_keymap_size);
-#endif
+    if (m_locks[0] /*CapsLock*/ && map_withmod && (map_withmod->flags & QEvdevKeyboardMap::IsLetter))
+        modifiers ^= QEvdevKeyboardMap::ModShift;
+
+    qCDebug(qLcEvdevKeyMap, "Processing key event: keycode=%3d, modifiers=%02x pressed=%d, autorepeat=%d  |  plain=%d, withmod=%d, size=%d",
+            keycode, modifiers, pressed ? 1 : 0, autorepeat ? 1 : 0,
+            int(map_plain ? map_plain - m_keymap : -1),
+            int(map_withmod ? map_withmod - m_keymap : -1),
+            m_keymap_size);
 
     const QEvdevKeyboardMap::Mapping *it = map_withmod ? map_withmod : map_plain;
 
     if (!it) {
-#ifdef QT_QPA_KEYMAP_DEBUG
         // we couldn't even find a plain mapping
-        qWarning("Could not find a suitable mapping for keycode: %3d, modifiers: %02x", keycode, m_modifiers);
-#endif
+        qCDebug(qLcEvdevKeyMap, "Could not find a suitable mapping for keycode: %3d, modifiers: %02x", keycode, modifiers);
         return result;
     }
 
@@ -285,7 +273,7 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
             lock ^= 1;
 
             switch (qtcode) {
-            case Qt::Key_CapsLock  : result = lock ? CapsLockOn : CapsLockOff; m_modifiers ^= QEvdevKeyboardMap::ModShift; break;
+            case Qt::Key_CapsLock  : result = lock ? CapsLockOn : CapsLockOff; break;
             case Qt::Key_NumLock   : result = lock ? NumLockOn : NumLockOff; break;
             case Qt::Key_ScrollLock: result = lock ? ScrollLockOn : ScrollLockOff; break;
             default                : break;
@@ -347,7 +335,7 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
         // so just report the plain mapping with additional modifiers.
         if ((it == map_plain && it != map_withmod) ||
             (map_withmod && !(map_withmod->qtcode & modmask))) {
-            qtcode |= QEvdevKeyboardHandler::toQtModifiers(m_modifiers);
+            qtcode |= QEvdevKeyboardHandler::toQtModifiers(modifiers);
         }
 
         if (m_composing == 2 && first_press && !(it->flags & QEvdevKeyboardMap::IsModifier)) {
@@ -398,12 +386,63 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
         }
 
         if (!skip) {
-#ifdef QT_QPA_KEYMAP_DEBUG
-            qWarning("Processing: uni=%04x, qt=%08x, qtmod=%08x", unicode, qtcode & ~modmask, (qtcode & modmask));
-#endif
+            // Up until now qtcode contained both the key and modifiers. Split it.
+            Qt::KeyboardModifiers qtmods = Qt::KeyboardModifiers(qtcode & modmask);
+            qtcode &= ~modmask;
 
-            // send the result to the server
-            processKeyEvent(keycode, unicode, qtcode & ~modmask, Qt::KeyboardModifiers(qtcode & modmask), pressed, autorepeat);
+            qCDebug(qLcEvdevKeyMap, "Processing: uni=%04x, qt=%08x, qtmod=%08x", unicode, qtcode, int(qtmods));
+
+            // If NumLockOff and keypad key pressed remap event sent
+            if (!m_locks[1] && (qtmods & Qt::KeypadModifier) &&
+                 keycode >= 71 &&
+                 keycode <= 83 &&
+                 keycode != 74 &&
+                 keycode != 78) {
+
+                unicode = 0xffff;
+                switch (keycode) {
+                case 71: //7 --> Home
+                    qtcode = Qt::Key_Home;
+                    break;
+                case 72: //8 --> Up
+                    qtcode = Qt::Key_Up;
+                    break;
+                case 73: //9 --> PgUp
+                    qtcode = Qt::Key_PageUp;
+                    break;
+                case 75: //4 --> Left
+                    qtcode = Qt::Key_Left;
+                    break;
+                case 76: //5 --> Clear
+                    qtcode = Qt::Key_Clear;
+                    break;
+                case 77: //6 --> right
+                    qtcode = Qt::Key_Right;
+                    break;
+                case 79: //1 --> End
+                    qtcode = Qt::Key_End;
+                    break;
+                case 80: //2 --> Down
+                    qtcode = Qt::Key_Down;
+                    break;
+                case 81: //3 --> PgDn
+                    qtcode = Qt::Key_PageDown;
+                    break;
+                case 82: //0 --> Ins
+                    qtcode = Qt::Key_Insert;
+                    break;
+                case 83: //, --> Del
+                    qtcode = Qt::Key_Delete;
+                    break;
+                }
+            }
+
+            // Map SHIFT + Tab to SHIFT + Backtab, QShortcutMap knows about this translation
+            if (qtcode == Qt::Key_Tab && (qtmods & Qt::ShiftModifier) == Qt::ShiftModifier)
+                qtcode = Qt::Key_Backtab;
+
+            // Generate the QPA event.
+            processKeyEvent(keycode, unicode, qtcode, qtmods, pressed, autorepeat);
         }
     }
     return result;
@@ -411,9 +450,7 @@ QEvdevKeyboardHandler::KeycodeAction QEvdevKeyboardHandler::processKeycode(quint
 
 void QEvdevKeyboardHandler::unloadKeymap()
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Unload current keymap and restore built-in";
-#endif
+    qCDebug(qLcEvdevKey) << "Unload current keymap and restore built-in";
 
     if (m_keymap && m_keymap != s_keymap_default)
         delete [] m_keymap;
@@ -430,13 +467,32 @@ void QEvdevKeyboardHandler::unloadKeymap()
     memset(m_locks, 0, sizeof(m_locks));
     m_composing = 0;
     m_dead_unicode = 0xffff;
+
+    //Set locks according to keyboard leds
+    quint16 ledbits[1];
+    memset(ledbits, 0, sizeof(ledbits));
+    if (::ioctl(m_fd, EVIOCGLED(sizeof(ledbits)), ledbits) < 0) {
+        qWarning("evdevkeyboard: Failed to query led states");
+        switchLed(LED_NUML,false);
+        switchLed(LED_CAPSL, false);
+        switchLed(LED_SCROLLL,false);
+    } else {
+        //Capslock
+        if ((ledbits[0]&0x02) > 0)
+            m_locks[0] = 1;
+        //Numlock
+        if ((ledbits[0]&0x01) > 0)
+            m_locks[1] = 1;
+        //Scrollock
+        if ((ledbits[0]&0x04) > 0)
+            m_locks[2] = 1;
+        qCDebug(qLcEvdevKey, "numlock=%d , capslock=%d, scrolllock=%d", m_locks[1], m_locks[0], m_locks[2]);
+    }
 }
 
 bool QEvdevKeyboardHandler::loadKeymap(const QString &file)
 {
-#ifdef QT_QPA_KEYMAP_DEBUG
-    qWarning() << "Load keymap" << file;
-#endif
+    qCDebug(qLcEvdevKey) << "Loading keymap" << file;
 
     QFile f(file);
 
@@ -460,7 +516,7 @@ bool QEvdevKeyboardHandler::loadKeymap(const QString &file)
     ds >> qmap_magic >> qmap_version >> qmap_keymap_size >> qmap_keycompose_size;
 
     if (ds.status() != QDataStream::Ok || qmap_magic != QEvdevKeyboardMap::FileMagic || qmap_version != 1 || qmap_keymap_size == 0) {
-        qWarning("'%s' is ot a valid.qmap keymap file.", qPrintable(file));
+        qWarning("'%s' is not a valid .qmap keymap file", qPrintable(file));
         return false;
     }
 
