@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the plugins of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -55,7 +53,9 @@
 #include "qcocoamimetypes.h"
 #include "qcocoaaccessibility.h"
 
+#include <qpa/qplatforminputcontextfactory_p.h>
 #include <qpa/qplatformaccessibility.h>
+#include <qpa/qplatforminputcontextfactory_p.h>
 #include <QtCore/qcoreapplication.h>
 
 #include <IOKit/graphics/IOGraphicsLib.h>
@@ -137,23 +137,26 @@ void QCocoaScreen::updateGeometry()
         m_name = QString::fromUtf8([[localizedNames objectForKey:[[localizedNames allKeys] objectAtIndex:0]] UTF8String]);
     [deviceInfo release];
 
-    QWindowSystemInterface::handleScreenGeometryChange(screen(), geometry());
+    QWindowSystemInterface::handleScreenGeometryChange(screen(), geometry(), availableGeometry());
     QWindowSystemInterface::handleScreenLogicalDotsPerInchChange(screen(), m_logicalDpi.first, m_logicalDpi.second);
     QWindowSystemInterface::handleScreenRefreshRateChange(screen(), m_refreshRate);
-    QWindowSystemInterface::handleScreenAvailableGeometryChange(screen(), availableGeometry());
 }
 
 qreal QCocoaScreen::devicePixelRatio() const
 {
-#if MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_7
-    if (QSysInfo::MacintoshVersion >= QSysInfo::MV_10_7) {
-        NSScreen * screen = osScreen();
-        return qreal(screen ? [screen backingScaleFactor] : 1.0);
-    } else
-#endif
-    {
-        return 1.0;
+    QMacAutoReleasePool pool;
+    NSScreen * screen = osScreen();
+    return qreal(screen ? [screen backingScaleFactor] : 1.0);
+}
+
+QPlatformScreen::SubpixelAntialiasingType QCocoaScreen::subpixelAntialiasingTypeHint() const
+{
+    QPlatformScreen::SubpixelAntialiasingType type = QPlatformScreen::subpixelAntialiasingTypeHint();
+    if (type == QPlatformScreen::Subpixel_None) {
+        // Every OSX machine has RGB pixels unless a peculiar or rotated non-Apple screen is attached
+        type = QPlatformScreen::Subpixel_RGB;
     }
+    return type;
 }
 
 QWindow *QCocoaScreen::topLevelAt(const QPoint &point) const
@@ -258,11 +261,25 @@ QPixmap QCocoaScreen::grabWindow(WId window, int x, int y, int width, int height
     return windowPixmap;
 }
 
+static QCocoaIntegration::Options parseOptions(const QStringList &paramList)
+{
+    QCocoaIntegration::Options options;
+    foreach (const QString &param, paramList) {
+#ifndef QT_NO_FREETYPE
+        if (param == QLatin1String("fontengine=freetype"))
+            options |= QCocoaIntegration::UseFreeTypeFontEngine;
+        else
+#endif
+            qWarning() << "Unknown option" << param;
+    }
+    return options;
+}
+
 QCocoaIntegration *QCocoaIntegration::mInstance = 0;
 
-QCocoaIntegration::QCocoaIntegration()
-    : mFontDb(new QCoreTextFontDatabase())
-    , mInputContext(new QCocoaInputContext)
+QCocoaIntegration::QCocoaIntegration(const QStringList &paramList)
+    : mOptions(parseOptions(paramList))
+    , mFontDb(new QCoreTextFontDatabase(mOptions.testFlag(UseFreeTypeFontEngine)))
 #ifndef QT_NO_ACCESSIBILITY
     , mAccessibility(new QCocoaAccessibility)
 #endif
@@ -276,8 +293,12 @@ QCocoaIntegration::QCocoaIntegration()
         qWarning("Creating multiple Cocoa platform integrations is not supported");
     mInstance = this;
 
+    QString icStr = QPlatformInputContextFactory::requested();
+    icStr.isNull() ? mInputContext.reset(new QCocoaInputContext)
+                   : mInputContext.reset(QPlatformInputContextFactory::create(icStr));
+
     initResources();
-    QCocoaAutoReleasePool pool;
+    QMacAutoReleasePool pool;
 
     qApp->setAttribute(Qt::AA_DontUseNativeMenuBar, false);
 
@@ -302,7 +323,7 @@ QCocoaIntegration::QCocoaIntegration()
     // ### For AA_MacPluginApplication we don't want to load the menu nib.
     // Qt 4 also does not set the application delegate, so that behavior
     // is matched here.
-    if (!QCoreApplication::testAttribute(Qt::AA_MacPluginApplication)) {
+    if (!QCoreApplication::testAttribute(Qt::AA_PluginApplication)) {
 
         // Set app delegate, link to the current delegate (if any)
         QCocoaApplicationDelegate *newDelegate = [QCocoaApplicationDelegate sharedDelegate];
@@ -316,10 +337,20 @@ QCocoaIntegration::QCocoaIntegration()
         [newDelegate setMenuLoader:qtMenuLoader];
     }
 
+    // The presentation options such as whether or not the dock and/or menu bar is
+    // hidden (automatically by the system) affects the main screen's available
+    // geometry. Since we're initializing the screens synchronously at application
+    // startup we need to ensure that the presentation options have been propagated
+    // to the screen before we read out its properties. Normally OS X does this in
+    // an asynchronous callback, but that's too late for us. We force the propagation
+    // by explicitly setting the presentation option to the magic 'default value',
+    // which will resolve to an actual value and result in screen invalidation.
+    cocoaApplication.presentationOptions = NSApplicationPresentationDefault;
     updateScreens();
 
     QMacInternalPasteboardMime::initializeMimeTypes();
     QCocoaMimeTypes::initializeMimeTypes();
+    QWindowSystemInterfacePrivate::TabletEvent::setPlatformSynthesizesMouse(false);
 }
 
 QCocoaIntegration::~QCocoaIntegration()
@@ -328,8 +359,8 @@ QCocoaIntegration::~QCocoaIntegration()
 
     qt_resetNSApplicationSendEvent();
 
-    QCocoaAutoReleasePool pool;
-    if (!QCoreApplication::testAttribute(Qt::AA_MacPluginApplication)) {
+    QMacAutoReleasePool pool;
+    if (!QCoreApplication::testAttribute(Qt::AA_PluginApplication)) {
         // remove the apple event handlers installed by QCocoaApplicationDelegate
         QCocoaApplicationDelegate *delegate = [QCocoaApplicationDelegate sharedDelegate];
         [delegate removeAppleEventHandlers];
@@ -345,7 +376,7 @@ QCocoaIntegration::~QCocoaIntegration()
 
     // Delete screens in reverse order to avoid crash in case of multiple screens
     while (!mScreens.isEmpty()) {
-        delete mScreens.takeLast();
+        destroyScreen(mScreens.takeLast());
     }
 
     clearToolbars();
@@ -356,6 +387,11 @@ QCocoaIntegration *QCocoaIntegration::instance()
     return mInstance;
 }
 
+QCocoaIntegration::Options QCocoaIntegration::options() const
+{
+    return mOptions;
+}
+
 /*!
     \brief Synchronizes the screen list, adds new screens, removes deleted ones
 */
@@ -402,14 +438,18 @@ void QCocoaIntegration::updateScreens()
         }
         siblings << screen;
     }
+
+    // Set virtual siblings list. All screens in mScreens are siblings, because we ignored the
+    // mirrors. Note that some of the screens we update the siblings list for here may be deleted
+    // below, but update anyway to keep the to-be-deleted screens out of the siblings list.
+    foreach (QCocoaScreen* screen, mScreens)
+        screen->setVirtualSiblings(siblings);
+
     // Now the leftovers in remainingScreens are no longer current, so we can delete them.
     foreach (QCocoaScreen* screen, remainingScreens) {
         mScreens.removeOne(screen);
-        delete screen;
+        destroyScreen(screen);
     }
-    // All screens in mScreens are siblings, because we ignored the mirrors.
-    foreach (QCocoaScreen* screen, mScreens)
-        screen->setVirtualSiblings(siblings);
 }
 
 QCocoaScreen *QCocoaIntegration::screenAtIndex(int index)
@@ -424,30 +464,38 @@ bool QCocoaIntegration::hasCapability(QPlatformIntegration::Capability cap) cons
 {
     switch (cap) {
     case ThreadedPixmaps:
+#ifndef QT_NO_OPENGL
     case OpenGL:
     case ThreadedOpenGL:
     case BufferQueueingOpenGL:
+#endif
     case WindowMasks:
     case MultipleWindows:
     case ForeignWindows:
     case RasterGLSurface:
+    case ApplicationState:
+    case ApplicationIcon:
         return true;
     default:
         return QPlatformIntegration::hasCapability(cap);
     }
 }
 
-
-
 QPlatformWindow *QCocoaIntegration::createPlatformWindow(QWindow *window) const
 {
     return new QCocoaWindow(window);
 }
 
+#ifndef QT_NO_OPENGL
 QPlatformOpenGLContext *QCocoaIntegration::createPlatformOpenGLContext(QOpenGLContext *context) const
 {
-    return new QCocoaGLContext(context->format(), context->shareHandle());
+    QCocoaGLContext *glContext = new QCocoaGLContext(context->format(),
+                                                     context->shareHandle(),
+                                                     context->nativeHandle());
+    context->setNativeHandle(glContext->nativeHandle());
+    return glContext;
 }
+#endif
 
 QPlatformBackingStore *QCocoaIntegration::createPlatformBackingStore(QWindow *window) const
 {
@@ -469,19 +517,17 @@ QCocoaNativeInterface *QCocoaIntegration::nativeInterface() const
     return mNativeInterface.data();
 }
 
-QCocoaInputContext *QCocoaIntegration::inputContext() const
+QPlatformInputContext *QCocoaIntegration::inputContext() const
 {
     return mInputContext.data();
 }
 
+#ifndef QT_NO_ACCESSIBILITY
 QCocoaAccessibility *QCocoaIntegration::accessibility() const
 {
-#ifndef QT_NO_ACCESSIBILITY
     return mAccessibility.data();
-#else
-    return 0;
-#endif
 }
+#endif
 
 QCocoaClipboard *QCocoaIntegration::clipboard() const
 {
@@ -514,8 +560,6 @@ QVariant QCocoaIntegration::styleHint(StyleHint hint) const
 {
     if (hint == QPlatformIntegration::FontSmoothingGamma)
         return 2.0;
-    if (hint == QPlatformIntegration::SynthesizeMouseFromTouchEvents)
-        return false;
 
     return QPlatformIntegration::styleHint(hint);
 }
@@ -554,4 +598,45 @@ void QCocoaIntegration::clearToolbars()
     mToolbars.clear();
 }
 
+void QCocoaIntegration::pushPopupWindow(QCocoaWindow *window)
+{
+    m_popupWindowStack.append(window);
+}
+
+QCocoaWindow *QCocoaIntegration::popPopupWindow()
+{
+    if (m_popupWindowStack.isEmpty())
+        return 0;
+    return m_popupWindowStack.takeLast();
+}
+
+QCocoaWindow *QCocoaIntegration::activePopupWindow() const
+{
+    if (m_popupWindowStack.isEmpty())
+        return 0;
+    return m_popupWindowStack.front();
+}
+
+QList<QCocoaWindow *> *QCocoaIntegration::popupWindowStack()
+{
+    return &m_popupWindowStack;
+}
+
+void QCocoaIntegration::setApplicationIcon(const QIcon &icon) const
+{
+    NSImage *image = nil;
+    if (!icon.isNull()) {
+        NSSize size = [[[NSApplication sharedApplication] dockTile] size];
+        QPixmap pixmap = icon.pixmap(size.width, size.height);
+        image = static_cast<NSImage *>(qt_mac_create_nsimage(pixmap));
+    }
+    [[NSApplication sharedApplication] setApplicationIconImage:image];
+    [image release];
+}
+
+void QCocoaIntegration::beep() const
+{
+    NSBeep();
+}
+
 QT_END_NAMESPACE
