@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the plugins of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -48,6 +46,7 @@
 #include <QtCore/QSysInfo>
 #include <QtGui/QGuiApplication>
 #include <qpa/qplatformnativeinterface.h>
+#include <QtPlatformHeaders/QWGLNativeContext>
 
 #include <algorithm>
 
@@ -142,8 +141,90 @@
 #define GL_CONTEXT_FLAG_DEBUG_BIT 0x00000002
 #endif
 
+// Common GL and WGL constants
+#define RESET_NOTIFICATION_STRATEGY_ARB 0x8256
+#define LOSE_CONTEXT_ON_RESET_ARB 0x8252
+
 QT_BEGIN_NAMESPACE
 
+QWindowsOpengl32DLL QOpenGLStaticContext::opengl32;
+
+void *QWindowsOpengl32DLL::resolve(const char *name)
+{
+#ifndef Q_OS_WINCE
+    void *proc = m_lib ? (void *) ::GetProcAddress(m_lib, name) : 0;
+#else
+    void *proc = m_lib ? (void *) ::GetProcAddress(m_lib, (const wchar_t *) QString::fromLatin1(name).utf16()) : 0;
+#endif
+
+    return proc;
+}
+
+bool QWindowsOpengl32DLL::init(bool softwareRendering)
+{
+    const QByteArray opengl32 = QByteArrayLiteral("opengl32.dll");
+    const QByteArray swopengl = QByteArrayLiteral("opengl32sw.dll");
+
+    QByteArray openglDll = qgetenv("QT_OPENGL_DLL");
+    if (openglDll.isEmpty())
+        openglDll = softwareRendering ? swopengl : opengl32;
+
+    openglDll = openglDll.toLower();
+    m_nonOpengl32 = openglDll != opengl32;
+
+    qCDebug(lcQpaGl) << "Qt: Using WGL and OpenGL from" << openglDll;
+
+    m_lib = ::LoadLibraryA(openglDll.constData());
+    if (!m_lib) {
+        qErrnoWarning(::GetLastError(), "Failed to load %s", openglDll.constData());
+        return false;
+    }
+
+    if (moduleIsNotOpengl32()) {
+        // Load opengl32.dll always. GDI functions like ChoosePixelFormat do
+        // GetModuleHandle for opengl32.dll and behave differently (and call back into
+        // opengl32) when the module is present. This is fine for dummy contexts and windows.
+        ::LoadLibraryA("opengl32.dll");
+    }
+
+    wglCreateContext = reinterpret_cast<HGLRC (WINAPI *)(HDC)>(resolve("wglCreateContext"));
+    wglDeleteContext = reinterpret_cast<BOOL (WINAPI *)(HGLRC)>(resolve("wglDeleteContext"));
+    wglGetCurrentContext = reinterpret_cast<HGLRC (WINAPI *)()>(resolve("wglGetCurrentContext"));
+    wglGetCurrentDC = reinterpret_cast<HDC (WINAPI *)()>(resolve("wglGetCurrentDC"));
+    wglGetProcAddress = reinterpret_cast<PROC (WINAPI *)(LPCSTR)>(resolve("wglGetProcAddress"));
+    wglMakeCurrent = reinterpret_cast<BOOL (WINAPI *)(HDC, HGLRC)>(resolve("wglMakeCurrent"));
+    wglShareLists = reinterpret_cast<BOOL (WINAPI *)(HGLRC, HGLRC)>(resolve("wglShareLists"));
+    wglSwapBuffers = reinterpret_cast<BOOL (WINAPI *)(HDC)>(resolve("wglSwapBuffers"));
+    wglSetPixelFormat = reinterpret_cast<BOOL (WINAPI *)(HDC, int, const PIXELFORMATDESCRIPTOR *)>(resolve("wglSetPixelFormat"));
+
+    glGetError = reinterpret_cast<GLenum (APIENTRY *)()>(resolve("glGetError"));
+    glGetIntegerv = reinterpret_cast<void (APIENTRY *)(GLenum , GLint *)>(resolve("glGetIntegerv"));
+    glGetString = reinterpret_cast<const GLubyte * (APIENTRY *)(GLenum )>(resolve("glGetString"));
+
+    return wglCreateContext && glGetError && glGetString;
+}
+
+BOOL QWindowsOpengl32DLL::swapBuffers(HDC dc)
+{
+    if (moduleIsNotOpengl32())
+        return wglSwapBuffers(dc);
+    else
+        return SwapBuffers(dc);
+}
+
+BOOL QWindowsOpengl32DLL::setPixelFormat(HDC dc, int pf, const PIXELFORMATDESCRIPTOR *pfd)
+{
+    if (moduleIsNotOpengl32())
+        return wglSetPixelFormat(dc, pf, pfd);
+    else
+        return SetPixelFormat(dc, pf, pfd);
+}
+
+QWindowsOpenGLContext *QOpenGLStaticContext::createContext(QOpenGLContext *context)
+{
+    return new QWindowsGLContext(this, context);
+}
+
 template <class MaskType, class FlagType> inline bool testFlag(MaskType mask, FlagType flag)
 {
     return (mask & MaskType(flag)) != 0;
@@ -162,47 +243,75 @@ static inline void initPixelFormatDescriptor(PIXELFORMATDESCRIPTOR *d)
     d->nVersion = 1;
 }
 
+#ifndef QT_NO_DEBUG_STREAM
 QDebug operator<<(QDebug d, const PIXELFORMATDESCRIPTOR &pd)
 {
-    QDebug nsp = d.nospace();
-    nsp << "PIXELFORMATDESCRIPTOR "
+    QDebugStateSaver saver(d);
+    d.nospace();
+    d << "PIXELFORMATDESCRIPTOR "
         << "dwFlags=" << hex << showbase << pd.dwFlags << dec << noshowbase;
-    if (pd.dwFlags & PFD_DRAW_TO_WINDOW) nsp << " PFD_DRAW_TO_WINDOW";
-    if (pd.dwFlags & PFD_DRAW_TO_BITMAP) nsp << " PFD_DRAW_TO_BITMAP";
-    if (pd.dwFlags & PFD_SUPPORT_GDI) nsp << " PFD_SUPPORT_GDI";
-    if (pd.dwFlags & PFD_SUPPORT_OPENGL) nsp << " PFD_SUPPORT_OPENGL";
-    if (pd.dwFlags & PFD_GENERIC_ACCELERATED) nsp << " PFD_GENERIC_ACCELERATED";
-    if (pd.dwFlags & PFD_SUPPORT_DIRECTDRAW) nsp << " PFD_SUPPORT_DIRECTDRAW";
-    if (pd.dwFlags & PFD_DIRECT3D_ACCELERATED) nsp << " PFD_DIRECT3D_ACCELERATED";
-    if (pd.dwFlags & PFD_SUPPORT_COMPOSITION) nsp << " PFD_SUPPORT_COMPOSITION";
-    if (pd.dwFlags & PFD_GENERIC_FORMAT) nsp << " PFD_GENERIC_FORMAT";
-    if (pd.dwFlags & PFD_NEED_PALETTE) nsp << " PFD_NEED_PALETTE";
-    if (pd.dwFlags & PFD_NEED_SYSTEM_PALETTE) nsp << " PFD_NEED_SYSTEM_PALETTE";
-    if (pd.dwFlags & PFD_DOUBLEBUFFER) nsp << " PFD_DOUBLEBUFFER";
-    if (pd.dwFlags & PFD_STEREO) nsp << " PFD_STEREO";
-    if (pd.dwFlags & PFD_SWAP_LAYER_BUFFERS) nsp << " PFD_SWAP_LAYER_BUFFERS";
-    if (hasGLOverlay(pd)) nsp << " overlay";
-    nsp << " iPixelType=" << pd.iPixelType << " cColorBits=" << pd.cColorBits
+    if (pd.dwFlags & PFD_DRAW_TO_WINDOW) d << " PFD_DRAW_TO_WINDOW";
+    if (pd.dwFlags & PFD_DRAW_TO_BITMAP) d << " PFD_DRAW_TO_BITMAP";
+    if (pd.dwFlags & PFD_SUPPORT_GDI) d << " PFD_SUPPORT_GDI";
+    if (pd.dwFlags & PFD_SUPPORT_OPENGL) d << " PFD_SUPPORT_OPENGL";
+    if (pd.dwFlags & PFD_GENERIC_ACCELERATED) d << " PFD_GENERIC_ACCELERATED";
+    if (pd.dwFlags & PFD_SUPPORT_DIRECTDRAW) d << " PFD_SUPPORT_DIRECTDRAW";
+    if (pd.dwFlags & PFD_DIRECT3D_ACCELERATED) d << " PFD_DIRECT3D_ACCELERATED";
+    if (pd.dwFlags & PFD_SUPPORT_COMPOSITION) d << " PFD_SUPPORT_COMPOSITION";
+    if (pd.dwFlags & PFD_GENERIC_FORMAT) d << " PFD_GENERIC_FORMAT";
+    if (pd.dwFlags & PFD_NEED_PALETTE) d << " PFD_NEED_PALETTE";
+    if (pd.dwFlags & PFD_NEED_SYSTEM_PALETTE) d << " PFD_NEED_SYSTEM_PALETTE";
+    if (pd.dwFlags & PFD_DOUBLEBUFFER) d << " PFD_DOUBLEBUFFER";
+    if (pd.dwFlags & PFD_STEREO) d << " PFD_STEREO";
+    if (pd.dwFlags & PFD_SWAP_LAYER_BUFFERS) d << " PFD_SWAP_LAYER_BUFFERS";
+    if (hasGLOverlay(pd)) d << " overlay";
+    d << " iPixelType=" << pd.iPixelType << " cColorBits=" << pd.cColorBits
         << " cRedBits=" << pd.cRedBits << " cRedShift=" << pd.cRedShift
         << " cGreenBits=" << pd.cGreenBits << " cGreenShift=" << pd.cGreenShift
         << " cBlueBits=" << pd.cBlueBits << " cBlueShift=" << pd.cBlueShift;
-    nsp  << " cDepthBits=" << pd.cDepthBits;
+    d  << " cDepthBits=" << pd.cDepthBits;
     if (pd.cStencilBits)
-        nsp << " cStencilBits=" << pd.cStencilBits;
+        d << " cStencilBits=" << pd.cStencilBits;
     if (pd.cAuxBuffers)
-        nsp << " cAuxBuffers=" << pd.cAuxBuffers;
-    nsp << " iLayerType=" << pd.iLayerType;
+        d << " cAuxBuffers=" << pd.cAuxBuffers;
+    d << " iLayerType=" << pd.iLayerType;
     if (pd.dwVisibleMask)
-        nsp << " dwVisibleMask=" << pd.dwVisibleMask;
+        d << " dwVisibleMask=" << pd.dwVisibleMask;
     if (pd.cAlphaBits)
-        nsp << " cAlphaBits=" << pd.cAlphaBits << " cAlphaShift=" << pd.cAlphaShift;
+        d << " cAlphaBits=" << pd.cAlphaBits << " cAlphaShift=" << pd.cAlphaShift;
     if (pd.cAccumBits)
-        nsp << " cAccumBits=" << pd.cAccumBits << " cAccumRedBits=" << pd.cAccumRedBits
+        d << " cAccumBits=" << pd.cAccumBits << " cAccumRedBits=" << pd.cAccumRedBits
         << " cAccumGreenBits=" << pd.cAccumGreenBits << " cAccumBlueBits=" << pd.cAccumBlueBits
         << " cAccumAlphaBits=" << pd.cAccumAlphaBits;
     return d;
 }
 
+QDebug operator<<(QDebug d, const QOpenGLStaticContext &s)
+{
+    QDebugStateSaver saver(d);
+    d.nospace();
+    d << "OpenGL: " << s.vendor << ',' << s.renderer << " default "
+        <<  s.defaultFormat;
+    if (s.extensions &  QOpenGLStaticContext::SampleBuffers)
+        d << ",SampleBuffers";
+    if (s.hasExtensions())
+        d << ", Extension-API present";
+    d << "\nExtensions: " << (s.extensionNames.count(' ') + 1);
+    if (QWindowsContext::verbose > 1)
+        d << s.extensionNames;
+    return d;
+}
+
+QDebug operator<<(QDebug d, const QWindowsOpenGLContextFormat &f)
+{
+    QDebugStateSaver saver(d);
+    d.nospace();
+    d << "ContextFormat: v" << (f.version >> 8) << '.' << (f.version & 0xFF)
+        << " profile: " << f.profile << " options: " << f.options;
+    return d;
+}
+#endif // !QT_NO_DEBUG_STREAM
+
 // Check whether an obtained PIXELFORMATDESCRIPTOR matches the request.
 static inline bool
     isAcceptableFormat(const QWindowsOpenGLAdditionalFormat &additional,
@@ -210,10 +319,11 @@ static inline bool
                        bool ignoreGLSupport = false) // ARB format may not contain it.
 {
     const bool pixmapRequested = testFlag(additional.formatFlags, QWindowsGLRenderToPixmap);
-    return (ignoreGLSupport || testFlag(pfd.dwFlags, PFD_SUPPORT_OPENGL))
-        && testFlag(pfd.dwFlags, PFD_DRAW_TO_BITMAP) == pixmapRequested
-        && hasGLOverlay(pfd) == testFlag(additional.formatFlags, QWindowsGLOverlay)
-        && (!pixmapRequested || pfd.cColorBits == additional.pixmapDepth);
+    const bool pixmapOk = !pixmapRequested || testFlag(pfd.dwFlags, PFD_DRAW_TO_BITMAP);
+    const bool colorOk =  !pixmapRequested || pfd.cColorBits == additional.pixmapDepth;
+    const bool glOk = ignoreGLSupport || testFlag(pfd.dwFlags, PFD_SUPPORT_OPENGL);
+    const bool overlayOk = hasGLOverlay(pfd) == testFlag(additional.formatFlags, QWindowsGLOverlay);
+    return pixmapOk && glOk && overlayOk && colorOk;
 }
 
 static void describeFormats(HDC hdc)
@@ -299,10 +409,23 @@ static PIXELFORMATDESCRIPTOR
 // over the available formats to find the best one.
 // Note: As of Windows 7, it seems direct-rendering is handled, so,
 // the code might be obsolete?
+//
+// NB! When using an implementation with a name different than opengl32.dll
+// this code path should not be used since it will result in a mess due to GDI
+// relying on and possibly calling back into functions in opengl32.dll (and not
+// the one we are using). This is not a problem usually since for Mesa, which
+// we are most likely to ship with a name other than opengl32.dll, the ARB code
+// path should work. Hence the early bail out below.
+//
 static int choosePixelFormat(HDC hdc, const QSurfaceFormat &format,
                             const QWindowsOpenGLAdditionalFormat &additional,
                             PIXELFORMATDESCRIPTOR *obtainedPfd)
 {
+    if (QOpenGLStaticContext::opengl32.moduleIsNotOpengl32()) {
+        qWarning("Attempted to use GDI functions with a non-opengl32.dll library");
+        return 0;
+    }
+
     // 1) Try ChoosePixelFormat().
     PIXELFORMATDESCRIPTOR requestedPfd = qPixelFormatFromSurfaceFormat(format, QWindowsGLDirectRendering);
     initPixelFormatDescriptor(obtainedPfd);
@@ -352,12 +475,12 @@ static int choosePixelFormat(HDC hdc, const QSurfaceFormat &format,
 
 static inline HGLRC createContext(HDC hdc, HGLRC shared)
 {
-    HGLRC result = wglCreateContext(hdc);
+    HGLRC result = QOpenGLStaticContext::opengl32.wglCreateContext(hdc);
     if (!result) {
         qErrnoWarning("%s: wglCreateContext failed.", __FUNCTION__);
         return 0;
     }
-    if (shared && !wglShareLists(shared, result))
+    if (shared && !QOpenGLStaticContext::opengl32.wglShareLists(shared, result))
         qErrnoWarning("%s: wglShareLists() failed.", __FUNCTION__);
     return result;
 }
@@ -465,6 +588,14 @@ static int choosePixelFormat(HDC hdc,
             break;
         if (iAttributes[samplesValuePosition] > 1) {
             iAttributes[samplesValuePosition] /= 2;
+        } else if (iAttributes[samplesValuePosition] == 1) {
+            // Fallback in case it is unable to initialize with any
+            // samples to avoid falling back to the GDI path
+            // NB: The sample attributes needs to be at the end for this
+            // to work correctly
+            iAttributes[samplesValuePosition - 1] = FALSE;
+            iAttributes[samplesValuePosition] = 0;
+            iAttributes[samplesValuePosition + 1] = 0;
         } else {
             break;
         }
@@ -615,6 +746,12 @@ static HGLRC createContext(const QOpenGLStaticContext &staticContext,
             break;
         }
     }
+
+    if (format.testOption(QSurfaceFormat::ResetNotification)) {
+        attributes[attribIndex++] = RESET_NOTIFICATION_STRATEGY_ARB;
+        attributes[attribIndex++] = LOSE_CONTEXT_ON_RESET_ARB;
+    }
+
     qCDebug(lcQpaGl) << __FUNCTION__ << "Creating context version"
         << majorVersion << '.' << minorVersion <<  attribIndex / 2 << "attributes";
 
@@ -623,7 +760,7 @@ static HGLRC createContext(const QOpenGLStaticContext &staticContext,
     if (!result) {
         QString message;
         QDebug(&message).nospace() << __FUNCTION__ << ": wglCreateContextAttribsARB() failed (GL error code: 0x"
-            << hex << glGetError() << dec << ") for format: " << format << ", shared context: " << shared;
+            << hex << staticContext.opengl32.glGetError() << dec << ") for format: " << format << ", shared context: " << shared;
         qErrnoWarning("%s", qPrintable(message));
     }
     return result;
@@ -648,16 +785,17 @@ static inline HGLRC createDummyGLContext(HDC dc)
     initPixelFormatDescriptor(&pixelFormDescriptor);
     pixelFormDescriptor.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_GENERIC_FORMAT;
     pixelFormDescriptor.iPixelType = PFD_TYPE_RGBA;
+    // Use the GDI variant, for the dummy this is fine, even when using something other than opengl32.dll.
     const int pixelFormat = ChoosePixelFormat(dc, &pixelFormDescriptor);
     if (!pixelFormat) {
         qErrnoWarning("%s: ChoosePixelFormat failed.", __FUNCTION__);
         return 0;
     }
-    if (!SetPixelFormat(dc, pixelFormat, &pixelFormDescriptor)) {
+    if (!QOpenGLStaticContext::opengl32.setPixelFormat(dc, pixelFormat, &pixelFormDescriptor)) {
         qErrnoWarning("%s: SetPixelFormat failed.", __FUNCTION__);
         return 0;
     }
-    HGLRC rc = wglCreateContext(dc);
+    HGLRC rc = QOpenGLStaticContext::opengl32.wglCreateContext(dc);
     if (!rc) {
         qErrnoWarning("%s: wglCreateContext failed.", __FUNCTION__);
         return 0;
@@ -668,8 +806,8 @@ static inline HGLRC createDummyGLContext(HDC dc)
 static inline QOpenGLContextData currentOpenGLContextData()
 {
     QOpenGLContextData result;
-    result.hdc = wglGetCurrentDC();
-    result.renderingContext = wglGetCurrentContext();
+    result.hdc = QOpenGLStaticContext::opengl32.wglGetCurrentDC();
+    result.renderingContext = QOpenGLStaticContext::opengl32.wglGetCurrentContext();
     return result;
 }
 
@@ -704,16 +842,11 @@ QWindowsOpenGLContextFormat QWindowsOpenGLContextFormat::current()
 {
     QWindowsOpenGLContextFormat result;
     const QByteArray version = QOpenGLStaticContext::getGlString(GL_VERSION);
-    const int majorDot = version.indexOf('.');
-    if (majorDot != -1) {
-        int minorDot = version.indexOf('.', majorDot + 1);
-        if (minorDot == -1)
-            minorDot = version.size();
-        result.version = (version.mid(0, majorDot).toInt() << 8)
-            + version.mid(majorDot + 1, minorDot - majorDot - 1).toInt();
-    } else {
+    int major, minor;
+    if (QPlatformOpenGLContext::parseOpenGLVersion(version, major, minor))
+        result.version = (major << 8) + minor;
+    else
         result.version = 0x0200;
-    }
     result.profile = QSurfaceFormat::NoProfile;
     if (result.version < 0x0300) {
         result.options |= QSurfaceFormat::DeprecatedFunctions;
@@ -721,16 +854,20 @@ QWindowsOpenGLContextFormat QWindowsOpenGLContextFormat::current()
     }
     // v3 onwards
     GLint value = 0;
-    glGetIntegerv(GL_CONTEXT_FLAGS, &value);
+    QOpenGLStaticContext::opengl32.glGetIntegerv(GL_CONTEXT_FLAGS, &value);
     if (!(value & GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT))
         result.options |= QSurfaceFormat::DeprecatedFunctions;
     if (value & GL_CONTEXT_FLAG_DEBUG_BIT)
         result.options |= QSurfaceFormat::DebugContext;
+    value = 0;
+    QOpenGLStaticContext::opengl32.glGetIntegerv(RESET_NOTIFICATION_STRATEGY_ARB, &value);
+    if (value == LOSE_CONTEXT_ON_RESET_ARB)
+        result.options |= QSurfaceFormat::ResetNotification;
     if (result.version < 0x0302)
         return result;
     // v3.2 onwards: Profiles
     value = 0;
-    glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &value);
+    QOpenGLStaticContext::opengl32.glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &value);
     if (value & GL_CONTEXT_CORE_PROFILE_BIT)
         result.profile = QSurfaceFormat::CoreProfile;
     else if (value & GL_CONTEXT_COMPATIBILITY_PROFILE_BIT)
@@ -749,14 +886,6 @@ void QWindowsOpenGLContextFormat::apply(QSurfaceFormat *format) const
         format->setOption(QSurfaceFormat::DeprecatedFunctions);
 }
 
-QDebug operator<<(QDebug d, const QWindowsOpenGLContextFormat &f)
-{
-    d.nospace() << "ContextFormat: v" << (f.version >> 8) << '.'
-                << (f.version & 0xFF) << " profile: " << f.profile
-                << " options: " << f.options;
-    return d;
-}
-
 /*!
     \class QOpenGLTemporaryContext
     \brief A temporary context that can be instantiated on the stack.
@@ -784,15 +913,15 @@ QOpenGLTemporaryContext::QOpenGLTemporaryContext() :
     m_previous(currentOpenGLContextData()),
     m_current(createDummyWindowOpenGLContextData())
 {
-    wglMakeCurrent(m_current.hdc, m_current.renderingContext);
+    QOpenGLStaticContext::opengl32.wglMakeCurrent(m_current.hdc, m_current.renderingContext);
 }
 
 QOpenGLTemporaryContext::~QOpenGLTemporaryContext()
 {
-    wglMakeCurrent(m_previous.hdc, m_previous.renderingContext);
+    QOpenGLStaticContext::opengl32.wglMakeCurrent(m_previous.hdc, m_previous.renderingContext);
     ReleaseDC(m_current.hwnd, m_current.hdc);
     DestroyWindow(m_current.hwnd);
-    wglDeleteContext(m_current.renderingContext);
+    QOpenGLStaticContext::opengl32.wglDeleteContext(m_current.renderingContext);
 }
 
 /*!
@@ -807,6 +936,11 @@ QOpenGLTemporaryContext::~QOpenGLTemporaryContext()
 
     Functions pending integration in the next version of OpenGL are post-fixed ARB.
 
+    No WGL or OpenGL functions are called directly from the windows plugin. Instead, the
+    static context loads opengl32.dll and resolves the necessary functions. This allows
+    building the plugin without linking to opengl32 and enables QT_OPENGL_DYNAMIC builds
+    where both the EGL and WGL (this) based implementation of the context are built.
+
     \note Initialization requires an active context (see create()).
 
     \sa QWindowsGLContext
@@ -822,11 +956,11 @@ QOpenGLStaticContext::QOpenGLStaticContext() :
     extensionNames(QOpenGLStaticContext::getGlString(GL_EXTENSIONS)),
     extensions(0),
     defaultFormat(QWindowsOpenGLContextFormat::current()),
-    wglGetPixelFormatAttribIVARB((WglGetPixelFormatAttribIVARB)wglGetProcAddress("wglGetPixelFormatAttribivARB")),
-    wglChoosePixelFormatARB((WglChoosePixelFormatARB)wglGetProcAddress("wglChoosePixelFormatARB")),
-    wglCreateContextAttribsARB((WglCreateContextAttribsARB)wglGetProcAddress("wglCreateContextAttribsARB")),
-    wglSwapInternalExt((WglSwapInternalExt)wglGetProcAddress("wglSwapIntervalEXT")),
-    wglGetSwapInternalExt((WglGetSwapInternalExt)wglGetProcAddress("wglGetSwapIntervalEXT"))
+    wglGetPixelFormatAttribIVARB((WglGetPixelFormatAttribIVARB)QOpenGLStaticContext::opengl32.wglGetProcAddress("wglGetPixelFormatAttribivARB")),
+    wglChoosePixelFormatARB((WglChoosePixelFormatARB)QOpenGLStaticContext::opengl32.wglGetProcAddress("wglChoosePixelFormatARB")),
+    wglCreateContextAttribsARB((WglCreateContextAttribsARB)QOpenGLStaticContext::opengl32.wglGetProcAddress("wglCreateContextAttribsARB")),
+    wglSwapInternalExt((WglSwapInternalExt)QOpenGLStaticContext::opengl32.wglGetProcAddress("wglSwapIntervalEXT")),
+    wglGetSwapInternalExt((WglGetSwapInternalExt)QOpenGLStaticContext::opengl32.wglGetProcAddress("wglGetSwapIntervalEXT"))
 {
     if (extensionNames.startsWith(SAMPLE_BUFFER_EXTENSION " ")
             || extensionNames.indexOf(" " SAMPLE_BUFFER_EXTENSION " ") != -1)
@@ -835,37 +969,27 @@ QOpenGLStaticContext::QOpenGLStaticContext() :
 
 QByteArray QOpenGLStaticContext::getGlString(unsigned int which)
 {
-    if (const GLubyte *s = glGetString(which))
-        return QByteArray((const char*)s);
+    if (const GLubyte *s = opengl32.glGetString(which))
+        return QByteArray(reinterpret_cast<const char*>(s));
     return QByteArray();
 }
 
-QOpenGLStaticContext *QOpenGLStaticContext::create()
+QOpenGLStaticContext *QOpenGLStaticContext::create(bool softwareRendering)
 {
+    if (!opengl32.init(softwareRendering)) {
+        qWarning("Failed to load and resolve WGL/OpenGL functions");
+        return 0;
+    }
+
     // We need a current context for wglGetProcAdress()/getGLString() to work.
     QScopedPointer<QOpenGLTemporaryContext> temporaryContext;
-    if (!wglGetCurrentContext())
+    if (!QOpenGLStaticContext::opengl32.wglGetCurrentContext())
         temporaryContext.reset(new QOpenGLTemporaryContext);
     QOpenGLStaticContext *result = new QOpenGLStaticContext;
     qCDebug(lcQpaGl) << __FUNCTION__ << *result;
     return result;
 }
 
-QDebug operator<<(QDebug d, const QOpenGLStaticContext &s)
-{
-    QDebug nsp = d.nospace();
-    nsp << "OpenGL: " << s.vendor << ',' << s.renderer << " default "
-        <<  s.defaultFormat;
-    if (s.extensions &  QOpenGLStaticContext::SampleBuffers)
-        nsp << ",SampleBuffers";
-    if (s.hasExtensions())
-        nsp << ", Extension-API present";
-    nsp  << "\nExtensions: " << (s.extensionNames.count(' ') + 1);
-    if (QWindowsContext::verbose > 1)
-        nsp <<  s.extensionNames;
-    return d;
-}
-
 /*!
     \class QWindowsGLContext
     \brief Open GL context.
@@ -881,15 +1005,63 @@ QDebug operator<<(QDebug d, const QOpenGLStaticContext &s)
     \ingroup qt-lighthouse-win
 */
 
-QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContext,
+QWindowsGLContext::QWindowsGLContext(QOpenGLStaticContext *staticContext,
                                      QOpenGLContext *context) :
     m_staticContext(staticContext),
     m_context(context),
     m_renderingContext(0),
     m_pixelFormat(0),
     m_extensionsUsed(false),
-    m_swapInterval(-1)
+    m_swapInterval(-1),
+    m_ownsContext(true),
+    m_getGraphicsResetStatus(0),
+    m_lost(false)
 {
+    if (!m_staticContext) // Something went very wrong. Stop here, isValid() will return false.
+        return;
+
+    QVariant nativeHandle = context->nativeHandle();
+    if (!nativeHandle.isNull()) {
+        // Adopt and existing context.
+        if (!nativeHandle.canConvert<QWGLNativeContext>()) {
+            qWarning("QWindowsGLContext: Requires a QWGLNativeContext");
+            return;
+        }
+        QWGLNativeContext handle = nativeHandle.value<QWGLNativeContext>();
+        HGLRC wglcontext = handle.context();
+        HWND wnd = handle.window();
+        if (!wglcontext || !wnd) {
+            qWarning("QWindowsGLContext: No context and window given");
+            return;
+        }
+
+        HDC dc = GetDC(wnd);
+        // A window with an associated pixel format is mandatory.
+        // When no SetPixelFormat() call has been made, the following will fail.
+        m_pixelFormat = GetPixelFormat(dc);
+        bool ok = m_pixelFormat != 0;
+        if (!ok)
+            qWarning("QWindowsGLContext: Failed to get pixel format");
+        ok = DescribePixelFormat(dc, m_pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &m_obtainedPixelFormatDescriptor);
+        if (!ok) {
+            qWarning("QWindowsGLContext: Failed to describe pixel format");
+        } else {
+            QWindowsOpenGLAdditionalFormat obtainedAdditional;
+            m_obtainedFormat = GDI::qSurfaceFormatFromPixelFormat(m_obtainedPixelFormatDescriptor, &obtainedAdditional);
+            m_renderingContext = wglcontext;
+            ok = updateObtainedParams(dc);
+        }
+
+        ReleaseDC(wnd, dc);
+
+        if (ok)
+            m_ownsContext = false;
+        else
+            m_renderingContext = 0;
+
+        return;
+    }
+
     QSurfaceFormat format = context->format();
     if (format.renderableType() == QSurfaceFormat::DefaultRenderableType)
         format.setRenderableType(QSurfaceFormat::OpenGL);
@@ -901,7 +1073,7 @@ QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContex
     static bool opengl32dll = false;
     if (!opengl32dll) {
         GLint params;
-        glGetIntegerv(GL_DEPTH_BITS, &params);
+        staticContext->opengl32.glGetIntegerv(GL_DEPTH_BITS, &params);
         opengl32dll = true;
     }
 
@@ -912,7 +1084,7 @@ QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContex
     HWND dummyWindow = 0;
     HDC hdc = 0;
     bool tryExtensions = false;
-    int obtainedSwapInternal = -1;
+    int obtainedSwapInterval = -1;
     do {
         dummyWindow = createDummyGLWindow();
         if (!dummyWindow)
@@ -954,7 +1126,7 @@ QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContex
             qWarning("%s: Unable find a suitable pixel format.", __FUNCTION__);
             break;
         }
-        if (!SetPixelFormat(hdc, m_pixelFormat, &m_obtainedPixelFormatDescriptor)) {
+        if (!QOpenGLStaticContext::opengl32.setPixelFormat(hdc, m_pixelFormat, &m_obtainedPixelFormatDescriptor)) {
             qErrnoWarning("SetPixelFormat failed.");
             break;
         }
@@ -978,18 +1150,16 @@ QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContex
         }
 
         // Query obtained parameters and apply swap interval.
-        if (!wglMakeCurrent(hdc, m_renderingContext)) {
-            qWarning("Failed to make context current.");
+        if (!updateObtainedParams(hdc, &obtainedSwapInterval))
             break;
-        }
 
-        QWindowsOpenGLContextFormat::current().apply(&m_obtainedFormat);
-
-        if (m_staticContext->wglGetSwapInternalExt)
-            obtainedSwapInternal = m_staticContext->wglGetSwapInternalExt();
-
-        wglMakeCurrent(0, 0);
     } while (false);
+
+    // Make the HGLRC retrievable via QOpenGLContext::nativeHandle().
+    // Do not provide the window since it is the dummy one and it is about to disappear.
+    if (m_renderingContext)
+        context->setNativeHandle(QVariant::fromValue<QWGLNativeContext>(QWGLNativeContext(m_renderingContext, 0)));
+
     if (hdc)
         ReleaseDC(dummyWindow, hdc);
     if (dummyWindow)
@@ -998,24 +1168,64 @@ QWindowsGLContext::QWindowsGLContext(const QOpenGLStaticContextPtr &staticContex
     qCDebug(lcQpaGl) << __FUNCTION__ << this << (tryExtensions ? "ARB" : "GDI")
         << " requested: " << context->format()
         << "\n    obtained #" << m_pixelFormat << (m_extensionsUsed ? "ARB" : "GDI") << m_obtainedFormat
-        << "\n    " << m_obtainedPixelFormatDescriptor << " swap interval: " << obtainedSwapInternal
+        << "\n    " << m_obtainedPixelFormatDescriptor << " swap interval: " << obtainedSwapInterval
         << "\n    default: " << m_staticContext->defaultFormat
         << "\n    HGLRC=" << m_renderingContext;
 }
 
 QWindowsGLContext::~QWindowsGLContext()
 {
-    if (m_renderingContext)
-        wglDeleteContext(m_renderingContext);
+    if (m_renderingContext && m_ownsContext)
+        QOpenGLStaticContext::opengl32.wglDeleteContext(m_renderingContext);
     releaseDCs();
 }
 
+bool QWindowsGLContext::updateObtainedParams(HDC hdc, int *obtainedSwapInterval)
+{
+    HGLRC prevContext = QOpenGLStaticContext::opengl32.wglGetCurrentContext();
+    HDC prevSurface = QOpenGLStaticContext::opengl32.wglGetCurrentDC();
+
+    if (!QOpenGLStaticContext::opengl32.wglMakeCurrent(hdc, m_renderingContext)) {
+        qWarning("Failed to make context current.");
+        return false;
+    }
+
+    QWindowsOpenGLContextFormat::current().apply(&m_obtainedFormat);
+
+    if (m_staticContext->wglGetSwapInternalExt && obtainedSwapInterval)
+        *obtainedSwapInterval = m_staticContext->wglGetSwapInternalExt();
+
+    bool hasRobustness = false;
+    if (m_obtainedFormat.majorVersion() < 3) {
+        const char *exts = reinterpret_cast<const char *>(QOpenGLStaticContext::opengl32.glGetString(GL_EXTENSIONS));
+        hasRobustness = exts && strstr(exts, "GL_ARB_robustness");
+    } else {
+        typedef const GLubyte * (APIENTRY *glGetStringi_t)(GLenum, GLuint);
+        glGetStringi_t glGetStringi = (glGetStringi_t) QOpenGLStaticContext::opengl32.wglGetProcAddress("glGetStringi");
+        if (glGetStringi) {
+            GLint n = 0;
+            QOpenGLStaticContext::opengl32.glGetIntegerv(GL_NUM_EXTENSIONS, &n);
+            for (GLint i = 0; i < n; ++i) {
+                const char *p = reinterpret_cast<const char *>(glGetStringi(GL_EXTENSIONS, i));
+                if (p && !strcmp(p, "GL_ARB_robustness")) {
+                    hasRobustness = true;
+                    break;
+                }
+            }
+        }
+    }
+    if (hasRobustness)
+        m_getGraphicsResetStatus = (GLenum (APIENTRY *)()) QOpenGLStaticContext::opengl32.wglGetProcAddress("glGetGraphicsResetStatusARB");
+
+    QOpenGLStaticContext::opengl32.wglMakeCurrent(prevSurface, prevContext);
+    return true;
+}
+
 void QWindowsGLContext::releaseDCs()
 {
-    const QOpenGLContextData *end = m_windowContexts.end();
-    for (const QOpenGLContextData *p = m_windowContexts.begin(); p < end; ++p)
-        ReleaseDC(p->hwnd, p->hdc);
-    m_windowContexts.resize(0);
+    for (const auto &e : m_windowContexts)
+        ReleaseDC(e.hwnd, e.hdc);
+    m_windowContexts.clear();
 }
 
 static inline QWindowsWindow *glWindowOf(QPlatformSurface *s)
@@ -1030,12 +1240,12 @@ static inline HWND handleOf(QPlatformSurface *s)
 
 // Find a window in a context list.
 static inline const QOpenGLContextData *
-    findByHWND(const Array<QOpenGLContextData> &data, HWND hwnd)
+    findByHWND(const std::vector<QOpenGLContextData> &data, HWND hwnd)
 {
-    const QOpenGLContextData *end = data.end();
-    for (const QOpenGLContextData *p = data.begin(); p < end; ++p)
-        if (p->hwnd == hwnd)
-            return p;
+    for (const auto &e : data) {
+        if (e.hwnd == hwnd)
+            return &e;
+    }
     return 0;
 }
 
@@ -1043,11 +1253,11 @@ void QWindowsGLContext::swapBuffers(QPlatformSurface *surface)
 {
     if (QWindowsContext::verbose > 1)
         qCDebug(lcQpaGl) << __FUNCTION__ << surface;
-    if (const QOpenGLContextData *contextData = findByHWND(m_windowContexts, handleOf(surface))) {
-        SwapBuffers(contextData->hdc);
-    } else {
+
+    if (const QOpenGLContextData *contextData = findByHWND(m_windowContexts, handleOf(surface)))
+        QOpenGLStaticContext::opengl32.swapBuffers(contextData->hdc);
+    else
         qWarning("%s: Cannot find window %p", __FUNCTION__, handleOf(surface));
-    }
 }
 
 bool QWindowsGLContext::makeCurrent(QPlatformSurface *surface)
@@ -1061,16 +1271,17 @@ bool QWindowsGLContext::makeCurrent(QPlatformSurface *surface)
 
     // Do we already have a DC entry for that window?
     QWindowsWindow *window = static_cast<QWindowsWindow *>(surface);
+    window->aboutToMakeCurrent();
     const HWND hwnd = window->handle();
     if (const QOpenGLContextData *contextData = findByHWND(m_windowContexts, hwnd)) {
         // Repeated calls to wglMakeCurrent when vsync is enabled in the driver will
         // often result in 100% cpuload. This check is cheap and avoids the problem.
         // This is reproducable on NVidia cards and Intel onboard chips.
-        if (wglGetCurrentContext() == contextData->renderingContext
-                && wglGetCurrentDC() == contextData->hdc) {
+        if (QOpenGLStaticContext::opengl32.wglGetCurrentContext() == contextData->renderingContext
+                && QOpenGLStaticContext::opengl32.wglGetCurrentDC() == contextData->hdc) {
             return true;
         }
-        return wglMakeCurrent(contextData->hdc, contextData->renderingContext);
+        return QOpenGLStaticContext::opengl32.wglMakeCurrent(contextData->hdc, contextData->renderingContext);
     }
     // Create a new entry.
     const QOpenGLContextData newContext(m_renderingContext, hwnd, GetDC(hwnd));
@@ -1079,7 +1290,7 @@ bool QWindowsGLContext::makeCurrent(QPlatformSurface *surface)
     // Initialize pixel format first time. This will apply to
     // the HWND as well and  must be done only once.
     if (!window->testFlag(QWindowsWindow::OpenGlPixelFormatInitialized)) {
-        if (!SetPixelFormat(newContext.hdc, m_pixelFormat, &m_obtainedPixelFormatDescriptor)) {
+        if (!QOpenGLStaticContext::opengl32.setPixelFormat(newContext.hdc, m_pixelFormat, &m_obtainedPixelFormatDescriptor)) {
             qErrnoWarning("%s: SetPixelFormat() failed", __FUNCTION__);
             ReleaseDC(newContext.hwnd, newContext.hdc);
             return false;
@@ -1088,9 +1299,18 @@ bool QWindowsGLContext::makeCurrent(QPlatformSurface *surface)
         if (m_obtainedFormat.swapBehavior() == QSurfaceFormat::DoubleBuffer)
             window->setFlag(QWindowsWindow::OpenGLDoubleBuffered);
     }
-    m_windowContexts.append(newContext);
+    m_windowContexts.push_back(newContext);
 
-    bool success = wglMakeCurrent(newContext.hdc, newContext.renderingContext);
+    m_lost = false;
+    bool success = QOpenGLStaticContext::opengl32.wglMakeCurrent(newContext.hdc, newContext.renderingContext);
+    if (!success) {
+        if (m_getGraphicsResetStatus && m_getGraphicsResetStatus()) {
+            m_lost = true;
+            qCDebug(lcQpaGl) << "makeCurrent(): context loss detected" << this;
+            // Drop the surface. Will recreate on the next makeCurrent.
+            window->invalidateSurface();
+        }
+    }
 
     // Set the swap interval
     if (m_staticContext->wglSwapInternalExt) {
@@ -1110,18 +1330,29 @@ void QWindowsGLContext::doneCurrent()
     if (QWindowsContext::verbose > 1)
         qCDebug(lcQpaGl) << __FUNCTION__ << this << m_windowContexts.size() << "contexts";
 #endif // DEBUG_GL
-    wglMakeCurrent(0, 0);
+    QOpenGLStaticContext::opengl32.wglMakeCurrent(0, 0);
     releaseDCs();
 }
 
-QWindowsGLContext::GL_Proc QWindowsGLContext::getProcAddress(const QByteArray &procName)
+QFunctionPointer QWindowsGLContext::getProcAddress(const char *procName)
 {
-    // TODO: Will that work with the calling conventions?
-    GL_Proc procAddress = reinterpret_cast<GL_Proc>(wglGetProcAddress(procName.constData()));
+    // Even though we use QFunctionPointer, it does not mean the function can be called.
+    // It will need to be cast to the proper function type with the correct calling
+    // convention. QFunctionPointer is nothing more than a glorified void* here.
+    QFunctionPointer procAddress = reinterpret_cast<QFunctionPointer>(QOpenGLStaticContext::opengl32.wglGetProcAddress(procName));
+
+    // We support AllGLFunctionsQueryable, which means this function must be able to
+    // return a function pointer even for functions that are in GL.h and exported
+    // normally from opengl32.dll. wglGetProcAddress() is not guaranteed to work for such
+    // functions, however in QT_OPENGL_DYNAMIC builds QOpenGLFunctions will just blindly
+    // call into here for _any_ OpenGL function.
+    if (!procAddress || procAddress == reinterpret_cast<void *>(0x1) || procAddress == reinterpret_cast<void *>(0x2)
+        || procAddress == reinterpret_cast<void *>(0x3) || procAddress == reinterpret_cast<void *>(-1))
+        procAddress = reinterpret_cast<QFunctionPointer>(QOpenGLStaticContext::opengl32.resolve(procName));
+
     if (QWindowsContext::verbose > 1)
-        qCDebug(lcQpaGl) << __FUNCTION__ <<  procName << wglGetCurrentContext() << "returns" << procAddress;
-    if (!procAddress)
-        qWarning("%s: Unable to resolve '%s'", __FUNCTION__, procName.constData());
+        qCDebug(lcQpaGl) << __FUNCTION__ <<  procName << QOpenGLStaticContext::opengl32.wglGetCurrentContext() << "returns" << procAddress;
+
     return procAddress;
 }
 
