@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtWidgets module of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -192,6 +190,31 @@ QT_BEGIN_NAMESPACE
 extern bool qt_sendSpontaneousEvent(QObject *, QEvent *);
 Q_WIDGETS_EXPORT extern bool qt_tab_all_widgets();
 
+/*!
+    \internal
+*/
+QGraphicsProxyWidgetPrivate::QGraphicsProxyWidgetPrivate()
+    : QGraphicsWidgetPrivate(),
+      dragDropWidget(nullptr),
+      posChangeMode(NoMode),
+      sizeChangeMode(NoMode),
+      visibleChangeMode(NoMode),
+      enabledChangeMode(NoMode),
+      styleChangeMode(NoMode),
+      paletteChangeMode(NoMode),
+      tooltipChangeMode(NoMode),
+      focusFromWidgetToProxy(false),
+      proxyIsGivingFocus(false)
+{
+}
+
+/*!
+    \internal
+*/
+QGraphicsProxyWidgetPrivate::~QGraphicsProxyWidgetPrivate()
+{
+}
+
 /*!
     \internal
 */
@@ -277,7 +300,7 @@ void QGraphicsProxyWidgetPrivate::sendWidgetMouseEvent(QGraphicsSceneMouseEvent
     // Send mouse event.
     QMouseEvent mouseEvent(type, pos, receiver->mapTo(receiver->topLevelWidget(), pos.toPoint()),
                            receiver->mapToGlobal(pos.toPoint()),
-                           event->button(), event->buttons(), event->modifiers());
+                           event->button(), event->buttons(), event->modifiers(), event->source());
 
     QWidget *embeddedMouseGrabberPtr = (QWidget *)embeddedMouseGrabber;
     QApplicationPrivate::sendMouseEvent(receiver, &mouseEvent, alienWidget, widget,
@@ -348,7 +371,7 @@ QWidget *QGraphicsProxyWidgetPrivate::findFocusChild(QWidget *child, bool next)
 
     // Run around the focus chain until we find a widget that can take tab focus.
     if (!child) {
-	child = next ? (QWidget *)widget : widget->d_func()->focus_prev;
+        child = next ? (QWidget *)widget : widget->d_func()->focus_prev;
     } else {
         child = next ? child->d_func()->focus_next : child->d_func()->focus_prev;
         if ((next && child == widget) || (!next && child == widget->d_func()->focus_prev)) {
@@ -356,11 +379,14 @@ QWidget *QGraphicsProxyWidgetPrivate::findFocusChild(QWidget *child, bool next)
         }
     }
 
+    if (!child)
+        return 0;
+
     QWidget *oldChild = child;
     uint focus_flag = qt_tab_all_widgets() ? Qt::TabFocus : Qt::StrongFocus;
     do {
         if (child->isEnabled()
-	    && child->isVisibleTo(widget)
+            && child->isVisibleTo(widget)
             && ((child->focusPolicy() & focus_flag) == focus_flag)
             && !(child->d_func()->extra && child->d_func()->extra->focus_proxy)) {
             return child;
@@ -376,6 +402,10 @@ QWidget *QGraphicsProxyWidgetPrivate::findFocusChild(QWidget *child, bool next)
 void QGraphicsProxyWidgetPrivate::_q_removeWidgetSlot()
 {
     Q_Q(QGraphicsProxyWidget);
+    if (!widget.isNull()) {
+        if (QWExtra *extra = widget->d_func()->extra)
+            extra->proxyWidget = 0;
+    }
     widget = 0;
     delete q;
 }
@@ -571,7 +601,8 @@ void QGraphicsProxyWidgetPrivate::setWidget_helper(QWidget *newWidget, bool auto
         resolvePalette(inheritedPaletteResolveMask);
         widget->update();
 
-        foreach (QGraphicsItem *child, q->childItems()) {
+        const auto childItems = q->childItems();
+        for (QGraphicsItem *child : childItems) {
             if (child->d_ptr->isProxyWidget()) {
                 QGraphicsProxyWidget *childProxy = static_cast<QGraphicsProxyWidget *>(child);
                 QWidget * parent = childProxy->widget();
@@ -880,6 +911,19 @@ bool QGraphicsProxyWidget::event(QEvent *event)
         break;
     }
 #endif
+    case QEvent::TouchBegin:
+    case QEvent::TouchUpdate:
+    case QEvent::TouchEnd: {
+        if (event->spontaneous())
+            qt_sendSpontaneousEvent(d->widget, event);
+        else
+            QApplication::sendEvent(d->widget, event);
+
+        if (event->isAccepted())
+            return true;
+
+        break;
+   }
     default:
         break;
     }
@@ -1156,7 +1200,7 @@ void QGraphicsProxyWidget::hoverMoveEvent(QGraphicsSceneHoverEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::hoverMoveEvent";
+    qDebug("QGraphicsProxyWidget::hoverMoveEvent");
 #endif
     // Ignore events on the window frame.
     if (!d->widget || !rect().contains(event->pos())) {
@@ -1196,7 +1240,7 @@ void QGraphicsProxyWidget::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::mouseMoveEvent";
+    qDebug("QGraphicsProxyWidget::mouseMoveEvent");
 #endif
     d->sendWidgetMouseEvent(event);
 }
@@ -1208,7 +1252,7 @@ void QGraphicsProxyWidget::mousePressEvent(QGraphicsSceneMouseEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::mousePressEvent";
+    qDebug("QGraphicsProxyWidget::mousePressEvent");
 #endif
     d->sendWidgetMouseEvent(event);
 }
@@ -1220,7 +1264,7 @@ void QGraphicsProxyWidget::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *event
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::mouseDoubleClickEvent";
+    qDebug("QGraphicsProxyWidget::mouseDoubleClickEvent");
 #endif
     d->sendWidgetMouseEvent(event);
 }
@@ -1233,7 +1277,7 @@ void QGraphicsProxyWidget::wheelEvent(QGraphicsSceneWheelEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::wheelEvent";
+    qDebug("QGraphicsProxyWidget::wheelEvent");
 #endif
     if (!d->widget)
         return;
@@ -1271,7 +1315,7 @@ void QGraphicsProxyWidget::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::mouseReleaseEvent";
+    qDebug("QGraphicsProxyWidget::mouseReleaseEvent");
 #endif
     d->sendWidgetMouseEvent(event);
 }
@@ -1283,7 +1327,7 @@ void QGraphicsProxyWidget::keyPressEvent(QKeyEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::keyPressEvent";
+    qDebug("QGraphicsProxyWidget::keyPressEvent");
 #endif
     d->sendWidgetKeyEvent(event);
 }
@@ -1295,7 +1339,7 @@ void QGraphicsProxyWidget::keyReleaseEvent(QKeyEvent *event)
 {
     Q_D(QGraphicsProxyWidget);
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::keyReleaseEvent";
+    qDebug("QGraphicsProxyWidget::keyReleaseEvent");
 #endif
     d->sendWidgetKeyEvent(event);
 }
@@ -1306,7 +1350,7 @@ void QGraphicsProxyWidget::keyReleaseEvent(QKeyEvent *event)
 void QGraphicsProxyWidget::focusInEvent(QFocusEvent *event)
 {
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::focusInEvent";
+    qDebug("QGraphicsProxyWidget::focusInEvent");
 #endif
     Q_D(QGraphicsProxyWidget);
 
@@ -1321,17 +1365,17 @@ void QGraphicsProxyWidget::focusInEvent(QFocusEvent *event)
 
     switch (event->reason()) {
     case Qt::TabFocusReason: {
-	if (QWidget *focusChild = d->findFocusChild(0, true))
+        if (QWidget *focusChild = d->findFocusChild(0, true))
             focusChild->setFocus(event->reason());
         break;
     }
     case Qt::BacktabFocusReason:
-	if (QWidget *focusChild = d->findFocusChild(0, false))
+        if (QWidget *focusChild = d->findFocusChild(0, false))
             focusChild->setFocus(event->reason());
         break;
     default:
-	if (d->widget && d->widget->focusWidget()) {
-	    d->widget->focusWidget()->setFocus(event->reason());
+        if (d->widget && d->widget->focusWidget()) {
+            d->widget->focusWidget()->setFocus(event->reason());
         }
         break;
     }
@@ -1345,7 +1389,7 @@ void QGraphicsProxyWidget::focusInEvent(QFocusEvent *event)
 void QGraphicsProxyWidget::focusOutEvent(QFocusEvent *event)
 {
 #ifdef GRAPHICSPROXYWIDGET_DEBUG
-    qDebug() << "QGraphicsProxyWidget::focusOutEvent";
+    qDebug("QGraphicsProxyWidget::focusOutEvent");
 #endif
     Q_D(QGraphicsProxyWidget);
     if (d->widget) {
@@ -1368,8 +1412,8 @@ bool QGraphicsProxyWidget::focusNextPrevChild(bool next)
     Qt::FocusReason reason = next ? Qt::TabFocusReason : Qt::BacktabFocusReason;
     QWidget *lastFocusChild = d->widget->focusWidget();
     if (QWidget *newFocusChild = d->findFocusChild(lastFocusChild, next)) {
-	newFocusChild->setFocus(reason);
-	return true;
+        newFocusChild->setFocus(reason);
+        return true;
     }
 
     return QGraphicsWidget::focusNextPrevChild(next);
