@@ -1,39 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2012 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -56,9 +43,12 @@ private slots:
     void createTest();
     void nullTest();
     void dataStreamTest();
+    void isTimeZoneIdAvailable();
     void availableTimeZoneIds();
     void stressTest();
     void windowsId();
+    void isValidId_data();
+    void isValidId();
     // Backend tests
     void utcTest();
     void icuTest();
@@ -347,6 +337,36 @@ void tst_QTimeZone::dataStreamTest()
     QCOMPARE(tz2.id(), tz1.id());
 }
 
+void tst_QTimeZone::isTimeZoneIdAvailable()
+{
+    QList<QByteArray> available = QTimeZone::availableTimeZoneIds();
+    foreach (const QByteArray &id, available)
+        QVERIFY(QTimeZone::isTimeZoneIdAvailable(id));
+
+    // a-z, A-Z, 0-9, '.', '-', '_' are valid chars
+    // Can't start with '-'
+    // Parts separated by '/', each part min 1 and max of 14 chars
+    QCOMPARE(QTimeZonePrivate::isValidId("az"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("AZ"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("09"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("a/z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("a.z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("a-z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("a_z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId(".z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("_z"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("12345678901234"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("12345678901234/12345678901234"), true);
+    QCOMPARE(QTimeZonePrivate::isValidId("a z"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("a\\z"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("a,z"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("/z"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("-z"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("123456789012345"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("123456789012345/12345678901234"), false);
+    QCOMPARE(QTimeZonePrivate::isValidId("12345678901234/123456789012345"), false);
+}
+
 void tst_QTimeZone::availableTimeZoneIds()
 {
     if (debug) {
@@ -490,6 +510,65 @@ void tst_QTimeZone::windowsId()
     QCOMPARE(QTimeZone::windowsIdToIanaIds(QByteArray(), QLocale::AnyCountry), list);
 }
 
+void tst_QTimeZone::isValidId_data()
+{
+#ifdef QT_BUILD_INTERNAL
+    QTest::addColumn<QByteArray>("input");
+    QTest::addColumn<bool>("valid");
+
+#define TESTSET(name, section, valid) \
+    QTest::newRow(name " front")  << QByteArray(section "/xyz/xyz")    << valid; \
+    QTest::newRow(name " middle") << QByteArray("xyz/" section "/xyz") << valid; \
+    QTest::newRow(name " back")   << QByteArray("xyz/xyz/" section)    << valid
+
+    TESTSET("empty", "", false);
+    TESTSET("minimal", "m", true);
+    TESTSET("maximal", "12345678901234", true);
+    TESTSET("too long", "123456789012345", false);
+
+    TESTSET("bad hyphen", "-hyphen", false);
+    TESTSET("good hyphen", "hy-phen", true);
+
+    TESTSET("valid char _", "_", true);
+    TESTSET("valid char .", ".", true);
+    TESTSET("valid char :", ":", true);
+    TESTSET("valid char +", "+", true);
+    TESTSET("valid char A", "A", true);
+    TESTSET("valid char Z", "Z", true);
+    TESTSET("valid char a", "a", true);
+    TESTSET("valid char z", "z", true);
+    TESTSET("valid char 0", "0", true);
+    TESTSET("valid char 9", "9", true);
+
+    TESTSET("invalid char ^", "^", false);
+    TESTSET("invalid char \"", "\"", false);
+    TESTSET("invalid char $", "$", false);
+    TESTSET("invalid char %", "%", false);
+    TESTSET("invalid char &", "&", false);
+    TESTSET("invalid char (", "(", false);
+    TESTSET("invalid char )", ")", false);
+    TESTSET("invalid char =", "=", false);
+    TESTSET("invalid char ?", "?", false);
+    TESTSET("invalid char ß", "ß", false);
+    TESTSET("invalid char \\x01", "\x01", false);
+    TESTSET("invalid char ' '", " ", false);
+
+#undef TESTSET
+#endif // QT_BUILD_INTERNAL
+}
+
+void tst_QTimeZone::isValidId()
+{
+#ifdef QT_BUILD_INTERNAL
+    QFETCH(QByteArray, input);
+    QFETCH(bool, valid);
+
+    QCOMPARE(QTimeZonePrivate::isValidId(input), valid);
+#else
+    QSKIP("This test requires a Qt -developer-build.");
+#endif
+}
+
 void tst_QTimeZone::utcTest()
 {
 #ifdef QT_BUILD_INTERNAL
@@ -683,16 +762,22 @@ void tst_QTimeZone::tzTest()
     // Warning: This could vary depending on age of TZ file!
 
     // Test low date uses first rule found
+    // Note: Depending on the OS in question, the database may be carrying the
+    // Local Mean Time. which for Berlin is 0:53:28
     QTimeZonePrivate::Data dat = tzp.data(-9999999999999);
     QCOMPARE(dat.atMSecsSinceEpoch, (qint64)-9999999999999);
-    QCOMPARE(dat.standardTimeOffset, 3600);
     QCOMPARE(dat.daylightTimeOffset, 0);
+    if (dat.abbreviation == "LMT") {
+        QCOMPARE(dat.standardTimeOffset, 3208);
+    } else {
+        QCOMPARE(dat.standardTimeOffset, 3600);
 
-    // Test previous to low value is invalid
-    dat = tzp.previousTransition(-9999999999999);
-    QCOMPARE(dat.atMSecsSinceEpoch, std::numeric_limits<qint64>::min());
-    QCOMPARE(dat.standardTimeOffset, std::numeric_limits<int>::min());
-    QCOMPARE(dat.daylightTimeOffset, std::numeric_limits<int>::min());
+        // Test previous to low value is invalid
+        dat = tzp.previousTransition(-9999999999999);
+        QCOMPARE(dat.atMSecsSinceEpoch, std::numeric_limits<qint64>::min());
+        QCOMPARE(dat.standardTimeOffset, std::numeric_limits<int>::min());
+        QCOMPARE(dat.daylightTimeOffset, std::numeric_limits<int>::min());
+    }
 
     dat = tzp.nextTransition(-9999999999999);
     QCOMPARE(dat.atMSecsSinceEpoch, (qint64)-2422054408000);
@@ -739,6 +824,24 @@ void tst_QTimeZone::tzTest()
     QCOMPARE(dat.offsetFromUtc, 3600);
     QCOMPARE(dat.standardTimeOffset, 3600);
     QCOMPARE(dat.daylightTimeOffset, 0);
+
+    // Test TZ timezone vs UTC timezone for fractionary negative offset
+    QTzTimeZonePrivate  tztz1("America/Caracas");
+    QUtcTimeZonePrivate tzutc1("UTC-04:30");
+    QVERIFY(tztz1.isValid());
+    QVERIFY(tzutc1.isValid());
+    QTzTimeZonePrivate::Data datatz1 = tztz1.data(std);
+    QTzTimeZonePrivate::Data datautc1 = tzutc1.data(std);
+    QCOMPARE(datatz1.offsetFromUtc, datautc1.offsetFromUtc);
+
+    // Test TZ timezone vs UTC timezone for fractionary positive offset
+    QTzTimeZonePrivate  tztz2("Asia/Calcutta");
+    QUtcTimeZonePrivate tzutc2("UTC+05:30");
+    QVERIFY(tztz2.isValid());
+    QVERIFY(tzutc2.isValid());
+    QTzTimeZonePrivate::Data datatz2 = tztz2.data(std);
+    QTzTimeZonePrivate::Data datautc2 = tzutc2.data(std);
+    QCOMPARE(datatz2.offsetFromUtc, datautc2.offsetFromUtc);
 #endif // Q_OS_UNIX
 }
 
