@@ -1,8 +1,8 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Copyright (C) 2013 BlackBerry Limited. All rights reserved.
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2016 BlackBerry Limited. All rights reserved.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtBluetooth module of the Qt Toolkit.
 **
@@ -11,30 +11,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -48,21 +46,27 @@
 #include "qbluetoothservicediscoveryagent.h"
 
 
-#include <QDebug>
+#include <QtCore/QLoggingCategory>
 #include <QSocketNotifier>
 
 QT_BEGIN_NAMESPACE
 
+Q_DECLARE_LOGGING_CATEGORY(QT_BT)
+
 /*!
     \class QBluetoothSocket
     \inmodule QtBluetooth
     \brief The QBluetoothSocket class enables connection to a Bluetooth device
     running a bluetooth server.
 
+    \since 5.2
+
     QBluetoothSocket supports two socket types, \l {QBluetoothServiceInfo::L2capProtocol}{L2CAP} and
     \l {QBluetoothServiceInfo::RfcommProtocol}{RFCOMM}.
 
     \l {QBluetoothServiceInfo::L2capProtocol}{L2CAP} is a low level datagram-oriented Bluetooth socket.
+    Android does not support \l {QBluetoothServiceInfo::L2capProtocol}{L2CAP} for socket
+    connections.
 
     \l {QBluetoothServiceInfo::RfcommProtocol}{RFCOMM} is a reliable, stream-oriented socket. RFCOMM
     sockets emulate an RS-232 serial port.
@@ -74,6 +78,9 @@ QT_BEGIN_NAMESPACE
     If the \l {QBluetoothServiceInfo::Protocol}{Protocol} is not supported on a platform, calling
     \l connectToService() will emit a \l {QBluetoothSocket::UnsupportedProtocolError}{UnsupportedProtocolError} error.
 
+    \note QBluetoothSocket does not support synchronous read and write operations. Functions such
+    as \l waitForReadyRead() and \l waitForBytesWritten() are not implemented. I/O operations should be
+    performed using \l readyRead(), \l read() and \l write().
 */
 
 /*!
@@ -103,6 +110,8 @@ QT_BEGIN_NAMESPACE
     \value NetworkError             Attempt to read or write from socket returned an error
     \value UnsupportedProtocolError The \l {QBluetoothServiceInfo::Protocol}{Protocol} is not
                                     supported on this platform.
+    \value OperationError           An operation was attempted while the socket was in a state
+                                    that did not permit it.
 */
 
 /*!
@@ -143,6 +152,10 @@ QT_BEGIN_NAMESPACE
     Aborts the current connection and resets the socket. Unlike disconnectFromService(), this
     function immediately closes the socket, discarding any pending data in the write buffer.
 
+    \note On Android, aborting the socket requires asynchronous interaction with Android threads.
+    Therefore the associated \l disconnected() and \l stateChanged() signals are delayed
+    until the threads have finished the closure.
+
     \sa disconnectFromService(), close()
 */
 
@@ -150,6 +163,11 @@ QT_BEGIN_NAMESPACE
     \fn void QBluetoothSocket::close()
 
     Disconnects the socket's connection with the device.
+
+    \note On Android, closing the socket requires asynchronous interaction with Android threads.
+    Therefore the associated \l disconnected() and \l stateChanged() signals are delayed
+    until the threads have finished the closure.
+
 */
 
 /*!
@@ -166,19 +184,30 @@ QT_BEGIN_NAMESPACE
     \fn QString QBluetoothSocket::localName() const
 
     Returns the name of the local device.
+
+    Although some platforms may differ the socket must generally be connected to guarantee
+    the return of a valid name. In particular, this is true when dealing with platforms
+    that support multiple local Bluetooth adapters.
 */
 
 /*!
     \fn QBluetoothAddress QBluetoothSocket::localAddress() const
 
     Returns the address of the local device.
+
+    Although some platforms may differ the socket must generally be connected to guarantee
+    the return of a valid address. In particular, this is true when dealing with platforms
+    that support multiple local Bluetooth adapters.
 */
 
 /*!
     \fn quint16 QBluetoothSocket::localPort() const
 
     Returns the port number of the local socket if available, otherwise returns 0.
-    On BlackBerry, this feature is not supported and returns 0.
+    Although some platforms may differ the socket must generally be connected to guarantee
+    the return of a valid port number.
+
+    On Android and OS X, this feature is not supported and returns 0.
 */
 
 /*!
@@ -197,7 +226,7 @@ QT_BEGIN_NAMESPACE
     \fn quint16 QBluetoothSocket::peerPort() const
 
     Return the port number of the peer socket if available, otherwise returns 0.
-    On BlackBerry, this feature is not supported.
+    On Android, this feature is not supported.
 */
 
 /*!
@@ -223,7 +252,7 @@ QBluetoothSocket::QBluetoothSocket(QBluetoothServiceInfo::Protocol socketType, Q
     Q_D(QBluetoothSocket);
     d->ensureNativeSocket(socketType);
 
-    setOpenMode(QIODevice::ReadWrite);
+    setOpenMode(QIODevice::NotOpen);
 }
 
 /*!
@@ -233,7 +262,7 @@ QBluetoothSocket::QBluetoothSocket(QObject *parent)
   : QIODevice(parent), d_ptr(new QBluetoothSocketPrivate)
 {
     d_ptr->q_ptr = this;
-    setOpenMode(QIODevice::ReadWrite);
+    setOpenMode(QIODevice::NotOpen);
 }
 
 /*!
@@ -277,51 +306,60 @@ qint64 QBluetoothSocket::bytesToWrite() const
 /*!
     Attempts to connect to the service described by \a service.
 
-    The socket is opened in the given \a openMode.
+    The socket is opened in the given \a openMode. The \l socketType() may change
+    depending on the protocol required by \a service.
 
-    For BlueZ, the socket first enters ConnectingState and attempts to connect to the device providing
+    The socket first enters ConnectingState and attempts to connect to the device providing
     \a service. If a connection is established, QBluetoothSocket enters ConnectedState and
     emits connected().
 
-    On QNX the service connection can be established directly using the UUID of the remote service.
+    At any point, the socket can emit error() to signal that an error occurred.
 
-    At any point, the socket can emit error() to siganl that an error occurred.
+    Note that most platforms require a pairing prior to connecting to the remote device. Otherwise
+    the connection process may fail.
 
     \sa state(), disconnectFromService()
 */
 void QBluetoothSocket::connectToService(const QBluetoothServiceInfo &service, OpenMode openMode)
 {
     Q_D(QBluetoothSocket);
-    setOpenMode(openMode);
 
-#ifdef QT_QNX_BLUETOOTH
-    if (socketType() != QBluetoothServiceInfo::RfcommProtocol) {
-        d->socketError = QBluetoothSocket::UnsupportedProtocolError;
+    if (state() != QBluetoothSocket::UnconnectedState && state() != QBluetoothSocket::ServiceLookupState) {
+        qCWarning(QT_BT)  << "QBluetoothSocket::connectToService called on busy socket";
+        d->errorString = QBluetoothSocket::tr("Trying to connect while connection is in progress");
+        setSocketError(QBluetoothSocket::OperationError);
+        return;
+    }
+#if defined(QT_ANDROID_BLUETOOTH)
+    if (!d->ensureNativeSocket(service.socketProtocol())) {
         d->errorString = tr("Socket type not supported");
-        Q_EMIT error(d->socketError);
+        setSocketError(QBluetoothSocket::UnsupportedProtocolError);
         return;
     }
     d->connectToService(service.device().address(), service.serviceUuid(), openMode);
 #else
     if (service.protocolServiceMultiplexer() > 0) {
         if (!d->ensureNativeSocket(QBluetoothServiceInfo::L2capProtocol)) {
-            emit error(UnknownSocketError);
+            d->errorString = tr("Unknown socket error");
+            setSocketError(UnknownSocketError);
             return;
         }
         d->connectToService(service.device().address(), service.protocolServiceMultiplexer(), openMode);
     } else if (service.serverChannel() > 0) {
         if (!d->ensureNativeSocket(QBluetoothServiceInfo::RfcommProtocol)) {
-            emit error(UnknownSocketError);
+            d->errorString = tr("Unknown socket error");
+            setSocketError(UnknownSocketError);
             return;
         }
         d->connectToService(service.device().address(), service.serverChannel(), openMode);
     } else {
         // try doing service discovery to see if we can find the socket
-        if(service.serviceUuid().isNull()){
-            qWarning() << "No port, no PSM, and no UUID provided, unable to connect";
+        if (service.serviceUuid().isNull()
+                && !service.serviceClassUuids().contains(QBluetoothUuid::SerialPort)) {
+            qCWarning(QT_BT) << "No port, no PSM, and no UUID provided, unable to connect";
             return;
         }
-        //qDebug() << "Need a port/psm, doing discovery";
+        qCDebug(QT_BT) << "Need a port/psm, doing discovery";
         doDeviceDiscovery(service, openMode);
     }
 #endif
@@ -333,25 +371,38 @@ void QBluetoothSocket::connectToService(const QBluetoothServiceInfo &service, Op
 
     The socket is opened in the given \a openMode.
 
-    For BlueZ, the socket first enters the ServiceLookupState and queries the connection parameters for
+    For BlueZ, the socket first enters the \l ServiceLookupState and queries the connection parameters for
     \a uuid. If the service parameters are successfully retrieved the socket enters
     ConnectingState, and attempts to connect to \a address. If a connection is established,
     QBluetoothSocket enters Connected State and emits connected().
 
-    On BlackBerry, the service connection can be established directly using the UUID of the remote service.
+    On Android, the service connection can directly be established
+    using the UUID of the remote service. Therefore the platforms does not require
+    the \l ServiceLookupState and \l socketType() is always set to
+    \l QBluetoothServiceInfo::RfcommProtocol.
 
     At any point, the socket can emit error() to signal that an error occurred.
 
+    Note that most platforms require a pairing prior to connecting to the remote device. Otherwise
+    the connection process may fail.
+
     \sa state(), disconnectFromService()
 */
 void QBluetoothSocket::connectToService(const QBluetoothAddress &address, const QBluetoothUuid &uuid, OpenMode openMode)
 {
-#ifdef QT_QNX_BLUETOOTH
     Q_D(QBluetoothSocket);
-    if (socketType() != QBluetoothServiceInfo::RfcommProtocol) {
-        d->socketError = QBluetoothSocket::UnsupportedProtocolError;
+
+    if (state() != QBluetoothSocket::UnconnectedState) {
+        qCWarning(QT_BT)  << "QBluetoothSocket::connectToService called on busy socket";
+        d->errorString = QBluetoothSocket::tr("Trying to connect while connection is in progress");
+        setSocketError(QBluetoothSocket::OperationError);
+        return;
+    }
+
+#if defined(QT_ANDROID_BLUETOOTH)
+    if (!d->ensureNativeSocket(QBluetoothServiceInfo::RfcommProtocol)) {
         d->errorString = tr("Socket type not supported");
-        Q_EMIT error(d->socketError);
+        setSocketError(QBluetoothSocket::UnsupportedProtocolError);
         return;
     }
     d->connectToService(address, uuid, openMode);
@@ -374,30 +425,43 @@ void QBluetoothSocket::connectToService(const QBluetoothAddress &address, const
 
     At any point, the socket can emit error() to signal that an error occurred.
 
-    On BlackBerry, a connection to a service can not be established using a port. Calling this function
+    On Android, a connection to a service can not be established using a port. Calling this function
     will emit a \l {QBluetoothSocket::ServiceNotFoundError}{ServiceNotFoundError}
 
+    Note that most platforms require a pairing prior to connecting to the remote device. Otherwise
+    the connection process may fail.
+
     \sa state(), disconnectFromService()
 */
 void QBluetoothSocket::connectToService(const QBluetoothAddress &address, quint16 port, OpenMode openMode)
 {
     Q_D(QBluetoothSocket);
-#ifdef QT_QNX_BLUETOOTH
+#if defined(QT_ANDROID_BLUETOOTH)
     Q_UNUSED(port);
     Q_UNUSED(openMode);
     Q_UNUSED(address);
-    d->socketError = QBluetoothSocket::ServiceNotFoundError;
-    d->errorString = tr("Connecting to port is not supported on QNX");
-    Q_EMIT error(d->socketError);
-    qWarning("Connecting to port is not supported");
+    d->errorString = tr("Connecting to port is not supported");
+    setSocketError(QBluetoothSocket::ServiceNotFoundError);
+    qCWarning(QT_BT) << "Connecting to port is not supported";
 #else
+    if (state() != QBluetoothSocket::UnconnectedState) {
+        qCWarning(QT_BT)  << "QBluetoothSocket::connectToService called on busy socket";
+        d->errorString = QBluetoothSocket::tr("Trying to connect while connection is in progress");
+        setSocketError(QBluetoothSocket::OperationError);
+        return;
+    }
+
     setOpenMode(openMode);
     d->connectToService(address, port, openMode);
 #endif
 }
 
 /*!
-    Returns the socket type.
+    Returns the socket type. The socket automatically adjusts to the protocol
+    offered by the remote service.
+
+    Android only support \l{QBluetoothServiceInfo::RfcommProtocol}{RFCOMM}
+    based sockets.
 */
 QBluetoothServiceInfo::Protocol QBluetoothSocket::socketType() const
 {
@@ -432,6 +496,60 @@ QString QBluetoothSocket::errorString() const
     return d->errorString;
 }
 
+/*!
+    Sets the preferred security parameter for the connection attempt to
+    \a flags. This value is incorporated when calling \l connectToService().
+    Therefore it is required to reconnect to change this parameter for an
+    existing connection.
+
+    On Bluez this property is set to QBluetooth::Authorization by default.
+
+    On OS X, this value is ignored as the platform does not permit access
+    to the security parameter of the socket. By default the platform prefers
+    secure/encrypted connections though and therefore this function always
+    returns \l QBluetooth::Secure.
+
+    Android only supports two levels of security (secure and non-secure). If this flag is set to
+    \l QBluetooth::NoSecurity the socket object will not employ any authentication or encryption.
+    Any other security flag combination will trigger a secure Bluetooth connection.
+    This flag is set to \l QBluetooth::Secure by default.
+
+    \note A secure connection requires a pairing between the two devices. On
+    some platforms, the pairing is automatically initiated during the establishment
+    of the connection. Other platforms require the application to manually trigger
+    the pairing before attempting to connect.
+
+    \sa preferredSecurityFlags()
+
+    \since 5.6
+*/
+void QBluetoothSocket::setPreferredSecurityFlags(QBluetooth::SecurityFlags flags)
+{
+    Q_D(QBluetoothSocket);
+    if (d->secFlags != flags)
+        d->secFlags = flags;
+}
+
+/*!
+    Returns the security parameters used for the initial connection
+    attempt.
+
+    The security parameters may be renegotiated between the two parties
+    during or after the connection has been established. If such a change happens
+    it is not reflected in the value of this flag.
+
+    On OS X, this flag is always set to \l QBluetooth::Secure.
+
+    \sa setPreferredSecurityFlags()
+
+    \since 5.6
+*/
+QBluetooth::SecurityFlags QBluetoothSocket::preferredSecurityFlags() const
+{
+    Q_D(const QBluetoothSocket);
+    return d->secFlags;
+}
+
 /*!
     Sets the socket state to \a state.
 */
@@ -482,13 +600,15 @@ void QBluetoothSocket::doDeviceDiscovery(const QBluetoothServiceInfo &service, O
     Q_D(QBluetoothSocket);
 
     setSocketState(QBluetoothSocket::ServiceLookupState);
-    //qDebug() << "Starting discovery";
+    qCDebug(QT_BT) << "Starting Bluetooth Socket discovery";
 
     if(d->discoveryAgent) {
+        d->discoveryAgent->stop();
         delete d->discoveryAgent;
     }
 
-    d->discoveryAgent = new QBluetoothServiceDiscoveryAgent(service.device().address(),this);
+    d->discoveryAgent = new QBluetoothServiceDiscoveryAgent(this);
+    d->discoveryAgent->setRemoteAddress(service.device().address());
 
     //qDebug() << "Got agent";
 
@@ -497,16 +617,17 @@ void QBluetoothSocket::doDeviceDiscovery(const QBluetoothServiceInfo &service, O
 
     d->openMode = openMode;
 
+    QList<QBluetoothUuid> filterUuids = service.serviceClassUuids();
     if(!service.serviceUuid().isNull())
-        d->discoveryAgent->setUuidFilter(service.serviceUuid());
+        filterUuids.append(service.serviceUuid());
 
-    if(!service.serviceClassUuids().isEmpty())
-        d->discoveryAgent->setUuidFilter(service.serviceClassUuids());
+    if (!filterUuids.isEmpty())
+        d->discoveryAgent->setUuidFilter(filterUuids);
 
     // we have to ID the service somehow
     Q_ASSERT(!d->discoveryAgent->uuidFilter().isEmpty());
 
-    //qDebug() << "UUID filter" << d->discoveryAgent->uuidFilter();
+    qCDebug(QT_BT) << "UUID filter" << d->discoveryAgent->uuidFilter();
 
     d->discoveryAgent->start(QBluetoothServiceDiscoveryAgent::FullDiscovery);
 }
@@ -514,21 +635,25 @@ void QBluetoothSocket::doDeviceDiscovery(const QBluetoothServiceInfo &service, O
 void QBluetoothSocket::serviceDiscovered(const QBluetoothServiceInfo &service)
 {
     Q_D(QBluetoothSocket);
-    //qDebug() << "FOUND SERVICE!" << service;
-    if(service.protocolServiceMultiplexer() != 0 || service.serverChannel() != 0) {
+    qCDebug(QT_BT) << "FOUND SERVICE!" << service;
+    if (service.protocolServiceMultiplexer() > 0 || service.serverChannel() > 0) {
         connectToService(service, d->openMode);
         d->discoveryAgent->deleteLater();
         d->discoveryAgent = 0;
+    } else {
+        qCDebug(QT_BT) << "Could not find port/psm for potential remote service";
     }
 }
 
 void QBluetoothSocket::discoveryFinished()
 {
-    //qDebug() << "Socket discovery finished";
+    qCDebug(QT_BT) << "Socket discovery finished";
     Q_D(QBluetoothSocket);
-    if(d->discoveryAgent){
-        //qDebug() << "Didn't find any";
-        emit error(QBluetoothSocket::ServiceNotFoundError);
+    if (d->discoveryAgent){
+        qCDebug(QT_BT) << "Didn't find any";
+        d->errorString = tr("Service cannot be found");
+        setSocketError(ServiceNotFoundError);
+        setSocketState(QBluetoothSocket::UnconnectedState);
         d->discoveryAgent->deleteLater();
         d->discoveryAgent = 0;
     }
@@ -536,16 +661,31 @@ void QBluetoothSocket::discoveryFinished()
 
 void QBluetoothSocket::abort()
 {
+    if (state() == UnconnectedState)
+        return;
+
     Q_D(QBluetoothSocket);
+    setOpenMode(QIODevice::NotOpen);
+
+    if (state() == ServiceLookupState && d->discoveryAgent) {
+        d->discoveryAgent->disconnect();
+        d->discoveryAgent->stop();
+        d->discoveryAgent = 0;
+    }
+
+    setSocketState(ClosingState);
     d->abort();
+
+#ifndef QT_ANDROID_BLUETOOTH
+    //Android closes when the Java event loop comes around
     setSocketState(QBluetoothSocket::UnconnectedState);
+    emit disconnected();
+#endif
 }
 
 void QBluetoothSocket::disconnectFromService()
 {
-    // TODO: is this all we need to do?
-    Q_D(QBluetoothSocket);
-    d->close();
+    close();
 }
 
 QString QBluetoothSocket::localName() const
@@ -587,6 +727,13 @@ quint16 QBluetoothSocket::peerPort() const
 qint64 QBluetoothSocket::writeData(const char *data, qint64 maxSize)
 {
     Q_D(QBluetoothSocket);
+
+    if (!data || maxSize <= 0) {
+        d_ptr->errorString = tr("Invalid data/data size");
+        setSocketError(QBluetoothSocket::OperationError);
+        return -1;
+    }
+
     return d->writeData(data, maxSize);
 }
 
@@ -598,12 +745,27 @@ qint64 QBluetoothSocket::readData(char *data, qint64 maxSize)
 
 void QBluetoothSocket::close()
 {
+    if (state() == UnconnectedState)
+        return;
+
     Q_D(QBluetoothSocket);
+    setOpenMode(QIODevice::NotOpen);
+
+    if (state() == ServiceLookupState && d->discoveryAgent) {
+        d->discoveryAgent->disconnect();
+        d->discoveryAgent->stop();
+        d->discoveryAgent = 0;
+    }
+
     setSocketState(ClosingState);
 
     d->close();
 
+#ifndef QT_ANDROID_BLUETOOTH
+    //Android closes when the Java event loop comes around
     setSocketState(UnconnectedState);
+    emit disconnected();
+#endif
 }
 
 /*!
@@ -623,6 +785,7 @@ bool QBluetoothSocket::setSocketDescriptor(int socketDescriptor, QBluetoothServi
 
 /*!
   Returns the platform-specific socket descriptor, if available.
+  This function returns -1 if the descriptor is not available or an error has occurred.
 */
 
 int QBluetoothSocket::socketDescriptor() const
