@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtBluetooth module of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -45,12 +43,13 @@
 #include "bluez/manager_p.h"
 #include "bluez/adapter_p.h"
 #include "bluez/device_p.h"
+#include "bluez/objectmanager_p.h"
+#include <QtBluetooth/QBluetoothLocalDevice>
+#include "bluez/bluez_data_p.h"
 
 #include <qplatformdefs.h>
 
-#include <bluetooth/bluetooth.h>
-#include <bluetooth/rfcomm.h>
-#include <bluetooth/l2cap.h>
+#include <QtCore/QLoggingCategory>
 
 #include <errno.h>
 #include <unistd.h>
@@ -60,14 +59,19 @@
 
 QT_BEGIN_NAMESPACE
 
+Q_DECLARE_LOGGING_CATEGORY(QT_BT_BLUEZ)
+
 QBluetoothSocketPrivate::QBluetoothSocketPrivate()
     : socket(-1),
       socketType(QBluetoothServiceInfo::UnknownProtocol),
       state(QBluetoothSocket::UnconnectedState),
+      socketError(QBluetoothSocket::NoSocketError),
       readNotifier(0),
       connectWriteNotifier(0),
       connecting(false),
-      discoveryAgent(0)
+      discoveryAgent(0),
+      secFlags(QBluetooth::Authorization),
+      lowEnergySocketType(0)
 {
 }
 
@@ -113,9 +117,9 @@ bool QBluetoothSocketPrivate::ensureNativeSocket(QBluetoothServiceInfo::Protocol
 
     Q_Q(QBluetoothSocket);
     readNotifier = new QSocketNotifier(socket, QSocketNotifier::Read);
-    QObject::connect(readNotifier, SIGNAL(activated(int)), q, SLOT(_q_readNotify()));
-    connectWriteNotifier = new QSocketNotifier(socket, QSocketNotifier::Write, q);    
-    QObject::connect(connectWriteNotifier, SIGNAL(activated(int)), q, SLOT(_q_writeNotify()));
+    QObject::connect(readNotifier, SIGNAL(activated(int)), this, SLOT(_q_readNotify()));
+    connectWriteNotifier = new QSocketNotifier(socket, QSocketNotifier::Write, q);
+    QObject::connect(connectWriteNotifier, SIGNAL(activated(int)), this, SLOT(_q_writeNotify()));
 
     connectWriteNotifier->setEnabled(false);
     readNotifier->setEnabled(false);
@@ -127,9 +131,35 @@ bool QBluetoothSocketPrivate::ensureNativeSocket(QBluetoothServiceInfo::Protocol
 void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address, quint16 port, QIODevice::OpenMode openMode)
 {
     Q_Q(QBluetoothSocket);
-    Q_UNUSED(openMode);
     int result = -1;
 
+    if (socket == -1 && !ensureNativeSocket(socketType)) {
+        errorString = QBluetoothSocket::tr("Unknown socket error");
+        q->setSocketError(QBluetoothSocket::UnknownSocketError);
+        return;
+    }
+
+    // apply preferred security level
+    // ignore QBluetooth::Authentication -> not used anymore by kernel
+    struct bt_security security;
+    memset(&security, 0, sizeof(security));
+
+    if (secFlags & QBluetooth::Authorization)
+        security.level = BT_SECURITY_LOW;
+    if (secFlags & QBluetooth::Encryption)
+        security.level = BT_SECURITY_MEDIUM;
+    if (secFlags & QBluetooth::Secure)
+        security.level = BT_SECURITY_HIGH;
+
+    if (setsockopt(socket, SOL_BLUETOOTH, BT_SECURITY,
+                   &security, sizeof(security)) != 0) {
+        qCWarning(QT_BT_BLUEZ) << "Failed to set socket option, closing socket for safety" << errno;
+        qCWarning(QT_BT_BLUEZ) << "Error: " << qt_error_string(errno);
+        errorString = QBluetoothSocket::tr("Cannot set connection security level");
+        q->setSocketError(QBluetoothSocket::UnknownSocketError);
+        return;
+    }
+
     if (socketType == QBluetoothServiceInfo::RfcommProtocol) {
         sockaddr_rc addr;
 
@@ -148,7 +178,22 @@ void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address,
 
         memset(&addr, 0, sizeof(addr));
         addr.l2_family = AF_BLUETOOTH;
-        addr.l2_psm = port;
+        // This is an ugly hack but the socket class does what's needed already.
+        // For L2CP GATT we need a channel rather than a socket and the LE address type
+        // We don't want to make this public API offering for now especially since
+        // only Linux (of all platforms supported by this library) supports this type
+        // of socket.
+
+#if defined(QT_BLUEZ_BLUETOOTH) && !defined(QT_BLUEZ_NO_BTLE)
+        if (lowEnergySocketType) {
+            addr.l2_cid = htobs(port);
+            addr.l2_bdaddr_type = lowEnergySocketType;
+        } else {
+            addr.l2_psm = htobs(port);
+        }
+#else
+        addr.l2_psm = htobs(port);
+#endif
 
         convertAddress(address.toUInt64(), addr.l2_bdaddr.b);
 
@@ -161,8 +206,9 @@ void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address,
     if (result >= 0 || (result == -1 && errno == EINPROGRESS)) {
         connecting = true;
         q->setSocketState(QBluetoothSocket::ConnectingState);
+        q->setOpenMode(openMode);
     } else {
-        errorString = QString::fromLocal8Bit(strerror(errno));
+        errorString = qt_error_string(errno);
         q->setSocketError(QBluetoothSocket::UnknownSocketError);
     }
 }
@@ -175,8 +221,8 @@ void QBluetoothSocketPrivate::_q_writeNotify()
         len = sizeof(errorno);
         ::getsockopt(socket, SOL_SOCKET, SO_ERROR, &errorno, (socklen_t*)&len);
         if(errorno) {
-            errorString = QString::fromLocal8Bit(strerror(errorno));
-            emit q->error(QBluetoothSocket::UnknownSocketError);
+            errorString = qt_error_string(errorno);
+            q->setSocketError(QBluetoothSocket::UnknownSocketError);
             return;
         }
 
@@ -193,13 +239,12 @@ void QBluetoothSocketPrivate::_q_writeNotify()
         }
 
         char buf[1024];
-        Q_Q(QBluetoothSocket);
 
         int size = txBuffer.read(buf, 1024);
 
         if (::write(socket, buf, size) != size) {
-            socketError = QBluetoothSocket::NetworkError;
-            emit q->error(socketError);
+            errorString = QBluetoothSocket::tr("Network Error");
+            q->setSocketError(QBluetoothSocket::NetworkError);
         }
         else {
             emit q->bytesWritten(size);
@@ -212,34 +257,30 @@ void QBluetoothSocketPrivate::_q_writeNotify()
             connectWriteNotifier->setEnabled(false);
             this->close();
         }
-    }    
+    }
 }
 
-// TODO: move to private backend?
-
 void QBluetoothSocketPrivate::_q_readNotify()
 {
     Q_Q(QBluetoothSocket);
     char *writePointer = buffer.reserve(QPRIVATELINEARBUFFER_BUFFERSIZE);
 //    qint64 readFromDevice = q->readData(writePointer, QPRIVATELINEARBUFFER_BUFFERSIZE);
     int readFromDevice = ::read(socket, writePointer, QPRIVATELINEARBUFFER_BUFFERSIZE);
+    buffer.chop(QPRIVATELINEARBUFFER_BUFFERSIZE - (readFromDevice < 0 ? 0 : readFromDevice));
     if(readFromDevice <= 0){
         int errsv = errno;
         readNotifier->setEnabled(false);
         connectWriteNotifier->setEnabled(false);
-        errorString = QString::fromLocal8Bit(strerror(errsv));
-        qWarning() << Q_FUNC_INFO << socket << "error:" << readFromDevice << errorString;
-        if(errsv == EHOSTDOWN)
-            emit q->error(QBluetoothSocket::HostNotFoundError);
-        else
-            emit q->error(QBluetoothSocket::UnknownSocketError);
+        errorString = qt_error_string(errsv);
+        qCWarning(QT_BT_BLUEZ) << Q_FUNC_INFO << socket << "error:" << readFromDevice << errorString;
+        if (errsv == EHOSTDOWN)
+            q->setSocketError(QBluetoothSocket::HostNotFoundError);
+        else if (errsv != ECONNRESET) // The other side closing the connection is not an error.
+            q->setSocketError(QBluetoothSocket::UnknownSocketError);
 
         q->disconnectFromService();
-        q->setSocketState(QBluetoothSocket::UnconnectedState);        
     }
     else {
-        buffer.chop(QPRIVATELINEARBUFFER_BUFFERSIZE - (readFromDevice < 0 ? 0 : readFromDevice));
-
         emit q->readyRead();
     }
 }
@@ -255,39 +296,17 @@ void QBluetoothSocketPrivate::abort()
     // we don't call disconnectFromService or
     // QBluetoothSocket::close
     QT_CLOSE(socket);
-
-    Q_Q(QBluetoothSocket);
-    emit q->disconnected();
+    socket = -1;
 }
 
 QString QBluetoothSocketPrivate::localName() const
 {
-    if (!m_localName.isEmpty())
-        return m_localName;
-
     const QBluetoothAddress address = localAddress();
     if (address.isNull())
         return QString();
 
-    OrgBluezManagerInterface manager(QLatin1String("org.bluez"), QLatin1String("/"),
-                                     QDBusConnection::systemBus());
-
-    QDBusPendingReply<QDBusObjectPath> reply = manager.FindAdapter(address.toString());
-    reply.waitForFinished();
-    if (reply.isError())
-        return QString();
-
-    OrgBluezAdapterInterface adapter(QLatin1String("org.bluez"), reply.value().path(),
-                                     QDBusConnection::systemBus());
-
-    QDBusPendingReply<QVariantMap> properties = adapter.GetProperties();
-    properties.waitForFinished();
-    if (properties.isError())
-        return QString();
-
-    m_localName = properties.value().value(QLatin1String("Name")).toString();
-
-    return m_localName;
+    QBluetoothLocalDevice device(address);
+    return device.name();
 }
 
 QBluetoothAddress QBluetoothSocketPrivate::localAddress() const
@@ -296,20 +315,14 @@ QBluetoothAddress QBluetoothSocketPrivate::localAddress() const
         sockaddr_rc addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.rc_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.rc_bdaddr.b));
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.l2_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.l2_bdaddr.b));
     }
 
     return QBluetoothAddress();
@@ -336,9 +349,6 @@ quint16 QBluetoothSocketPrivate::localPort() const
 
 QString QBluetoothSocketPrivate::peerName() const
 {
-    if (!m_peerName.isEmpty())
-        return m_peerName;
-
     quint64 bdaddr;
 
     if (socketType == QBluetoothServiceInfo::RfcommProtocol) {
@@ -348,7 +358,7 @@ QString QBluetoothSocketPrivate::peerName() const
         if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) < 0)
             return QString();
 
-        convertAddress(addr.rc_bdaddr.b, bdaddr);
+        convertAddress(addr.rc_bdaddr.b, &bdaddr);
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
@@ -356,49 +366,73 @@ QString QBluetoothSocketPrivate::peerName() const
         if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) < 0)
             return QString();
 
-        convertAddress(addr.l2_bdaddr.b, bdaddr);
+        convertAddress(addr.l2_bdaddr.b, &bdaddr);
     } else {
-        qWarning("peerName() called on socket of known type");
+        qCWarning(QT_BT_BLUEZ) << "peerName() called on socket of unknown type";
         return QString();
     }
 
     const QString peerAddress = QBluetoothAddress(bdaddr).toString();
     const QString localAdapter = localAddress().toString();
 
-    OrgBluezManagerInterface manager(QLatin1String("org.bluez"), QLatin1String("/"),
-                                     QDBusConnection::systemBus());
-
-    QDBusPendingReply<QDBusObjectPath> reply = manager.FindAdapter(localAdapter);
-    reply.waitForFinished();
-    if (reply.isError())
-        return QString();
-
-    OrgBluezAdapterInterface adapter(QLatin1String("org.bluez"), reply.value().path(),
-                                     QDBusConnection::systemBus());
-
-    QDBusPendingReply<QDBusObjectPath> deviceObjectPath = adapter.CreateDevice(peerAddress);
-    deviceObjectPath.waitForFinished();
-    if (deviceObjectPath.isError()) {
-        if (deviceObjectPath.error().name() != QLatin1String("org.bluez.Error.AlreadyExists"))
+    if (isBluez5()) {
+        OrgFreedesktopDBusObjectManagerInterface manager(QStringLiteral("org.bluez"),
+                                                         QStringLiteral("/"),
+                                                         QDBusConnection::systemBus());
+        QDBusPendingReply<ManagedObjectList> reply = manager.GetManagedObjects();
+        reply.waitForFinished();
+        if (reply.isError())
             return QString();
 
-        deviceObjectPath = adapter.FindDevice(peerAddress);
+        ManagedObjectList managedObjectList = reply.value();
+        for (ManagedObjectList::const_iterator it = managedObjectList.constBegin(); it != managedObjectList.constEnd(); ++it) {
+            const InterfaceList &ifaceList = it.value();
+
+            for (InterfaceList::const_iterator jt = ifaceList.constBegin(); jt != ifaceList.constEnd(); ++jt) {
+                const QString &iface = jt.key();
+                const QVariantMap &ifaceValues = jt.value();
+
+                if (iface == QStringLiteral("org.bluez.Device1")) {
+                    if (ifaceValues.value(QStringLiteral("Address")).toString() == peerAddress)
+                        return ifaceValues.value(QStringLiteral("Alias")).toString();
+                }
+            }
+        }
+        return QString();
+    } else {
+        OrgBluezManagerInterface manager(QStringLiteral("org.bluez"), QStringLiteral("/"),
+                                         QDBusConnection::systemBus());
+
+        QDBusPendingReply<QDBusObjectPath> reply = manager.FindAdapter(localAdapter);
+        reply.waitForFinished();
+        if (reply.isError())
+            return QString();
+
+        OrgBluezAdapterInterface adapter(QStringLiteral("org.bluez"), reply.value().path(),
+                                         QDBusConnection::systemBus());
+
+        QDBusPendingReply<QDBusObjectPath> deviceObjectPath = adapter.FindDevice(peerAddress);
         deviceObjectPath.waitForFinished();
-        if (deviceObjectPath.isError())
+        if (deviceObjectPath.isError()) {
+            if (deviceObjectPath.error().name() != QStringLiteral("org.bluez.Error.DoesNotExist"))
+                return QString();
+
+            deviceObjectPath = adapter.CreateDevice(peerAddress);
+            deviceObjectPath.waitForFinished();
+            if (deviceObjectPath.isError())
+                return QString();
+        }
+
+        OrgBluezDeviceInterface device(QStringLiteral("org.bluez"), deviceObjectPath.value().path(),
+                                       QDBusConnection::systemBus());
+
+        QDBusPendingReply<QVariantMap> properties = device.GetProperties();
+        properties.waitForFinished();
+        if (properties.isError())
             return QString();
+
+        return properties.value().value(QStringLiteral("Alias")).toString();
     }
-
-    OrgBluezDeviceInterface device(QLatin1String("org.bluez"), deviceObjectPath.value().path(),
-                                   QDBusConnection::systemBus());
-
-    QDBusPendingReply<QVariantMap> properties = device.GetProperties();
-    properties.waitForFinished();
-    if (properties.isError())
-        return QString();
-
-    m_peerName = properties.value().value(QLatin1String("Alias")).toString();
-
-    return m_peerName;
 }
 
 QBluetoothAddress QBluetoothSocketPrivate::peerAddress() const
@@ -407,20 +441,14 @@ QBluetoothAddress QBluetoothSocketPrivate::peerAddress() const
         sockaddr_rc addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.rc_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.rc_bdaddr.b));
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.l2_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.l2_bdaddr.b));
     }
 
     return QBluetoothAddress();
@@ -448,10 +476,18 @@ quint16 QBluetoothSocketPrivate::peerPort() const
 qint64 QBluetoothSocketPrivate::writeData(const char *data, qint64 maxSize)
 {
     Q_Q(QBluetoothSocket);
+
+    if (state != QBluetoothSocket::ConnectedState) {
+        errorString = QBluetoothSocket::tr("Cannot write while not connected");
+        q->setSocketError(QBluetoothSocket::OperationError);
+        return -1;
+    }
+
     if (q->openMode() & QIODevice::Unbuffered) {
         if (::write(socket, data, maxSize) != maxSize) {
-            socketError = QBluetoothSocket::NetworkError;
-            emit q->error(socketError);
+            errorString = QBluetoothSocket::tr("Network Error");
+            q->setSocketError(QBluetoothSocket::NetworkError);
+            return -1;
         }
 
         emit q->bytesWritten(maxSize);
@@ -461,11 +497,11 @@ qint64 QBluetoothSocketPrivate::writeData(const char *data, qint64 maxSize)
     else {
 
         if(!connectWriteNotifier)
-            return 0;
+            return -1;
 
         if(txBuffer.size() == 0) {
-            connectWriteNotifier->setEnabled(true);        
-            QMetaObject::invokeMethod(q, "_q_writeNotify", Qt::QueuedConnection);
+            connectWriteNotifier->setEnabled(true);
+            QMetaObject::invokeMethod(this, "_q_writeNotify", Qt::QueuedConnection);
         }
 
         char *txbuf = txBuffer.reserve(maxSize);
@@ -477,39 +513,28 @@ qint64 QBluetoothSocketPrivate::writeData(const char *data, qint64 maxSize)
 
 qint64 QBluetoothSocketPrivate::readData(char *data, qint64 maxSize)
 {
-    if(!buffer.isEmpty()){
+    Q_Q(QBluetoothSocket);
+
+    if (state != QBluetoothSocket::ConnectedState) {
+        errorString = QBluetoothSocket::tr("Cannot read while not connected");
+        q->setSocketError(QBluetoothSocket::OperationError);
+        return -1;
+    }
+
+    if (!buffer.isEmpty()) {
         int i = buffer.read(data, maxSize);
         return i;
-
     }
+
     return 0;
 }
 
 void QBluetoothSocketPrivate::close()
 {
-    Q_Q(QBluetoothSocket);
-
-    // Only go through closing if the socket was fully opened
-    if(state == QBluetoothSocket::ConnectedState)
-        q->setSocketState(QBluetoothSocket::ClosingState);
-
-    if(txBuffer.size() > 0 &&
-       state == QBluetoothSocket::ClosingState){
+    if (txBuffer.size() > 0)
         connectWriteNotifier->setEnabled(true);
-    }
-    else {
-
-        delete readNotifier;
-        readNotifier = 0;
-        delete connectWriteNotifier;
-        connectWriteNotifier = 0;
-
-        // We are disconnected now, so go to unconnected.
-        q->setSocketState(QBluetoothSocket::UnconnectedState);
-        emit q->disconnected();
-        ::close(socket);
-    }
-
+    else
+        abort();
 }
 
 bool QBluetoothSocketPrivate::setSocketDescriptor(int socketDescriptor, QBluetoothServiceInfo::Protocol socketType_,
@@ -530,9 +555,9 @@ bool QBluetoothSocketPrivate::setSocketDescriptor(int socketDescriptor, QBluetoo
         fcntl(socket, F_SETFL, flags | O_NONBLOCK);
 
     readNotifier = new QSocketNotifier(socket, QSocketNotifier::Read);
-    QObject::connect(readNotifier, SIGNAL(activated(int)), q, SLOT(_q_readNotify()));
+    QObject::connect(readNotifier, SIGNAL(activated(int)), this, SLOT(_q_readNotify()));
     connectWriteNotifier = new QSocketNotifier(socket, QSocketNotifier::Write, q);
-    QObject::connect(connectWriteNotifier, SIGNAL(activated(int)), q, SLOT(_q_writeNotify()));
+    QObject::connect(connectWriteNotifier, SIGNAL(activated(int)), this, SLOT(_q_writeNotify()));
 
     q->setSocketState(socketState);
     q->setOpenMode(openMode);
@@ -540,11 +565,6 @@ bool QBluetoothSocketPrivate::setSocketDescriptor(int socketDescriptor, QBluetoo
     return true;
 }
 
-int QBluetoothSocketPrivate::socketDescriptor() const
-{
-    return socket;
-}
-
 qint64 QBluetoothSocketPrivate::bytesAvailable() const
 {
     return buffer.size();
