@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtQml module of the Qt Toolkit.
 **
@@ -10,30 +10,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -50,10 +48,8 @@
 #include "qqml.h"
 #include "qqmlengine.h"
 #include "qqmlbinding_p.h"
-#include "qqmlglobal_p.h"
-#include "qqmlscript_p.h"
-#include <private/qqmlprofilerservice_p.h>
-#include <private/qqmlenginedebugservice_p.h>
+#include <private/qqmldebugconnector_p.h>
+#include <private/qqmldebugserviceinterfaces_p.h>
 #include "qqmlincubator.h"
 #include "qqmlincubator_p.h"
 #include <private/qqmljavascriptexpression_p.h>
@@ -63,6 +59,8 @@
 #include <private/qv4functionobject_p.h>
 #include <private/qv4script_p.h>
 #include <private/qv4scopedvalue_p.h>
+#include <private/qv4objectiterator_p.h>
+#include <private/qv4qobjectwrapper_p.h>
 
 #include <QStack>
 #include <QStringList>
@@ -71,23 +69,6 @@
 #include <qqmlinfo.h>
 #include "qqmlmemoryprofiler_p.h"
 
-#define INITIALPROPERTIES_SOURCE \
-        "(function(object, values) {"\
-            "try {"\
-                "for (var property in values) {" \
-                    "try {"\
-                        "var properties = property.split(\".\");"\
-                        "var o = object;"\
-                        "for (var ii = 0; ii < properties.length - 1; ++ii) {"\
-                            "o = o[properties[ii]];"\
-                        "}"\
-                        "o[properties[properties.length - 1]] = values[property];"\
-                    "} catch(e) {}"\
-                "}"\
-            "} catch(e) {}"\
-        "})"
-
-
 namespace {
     QThreadStorage<int> creationDepth;
 }
@@ -97,46 +78,17 @@ QT_BEGIN_NAMESPACE
 class QQmlComponentExtension : public QV8Engine::Deletable
 {
 public:
-    QQmlComponentExtension(QV8Engine *);
+    QQmlComponentExtension(QV4::ExecutionEngine *v4);
     virtual ~QQmlComponentExtension();
 
     QV4::PersistentValue incubationProto;
 };
-V8_DEFINE_EXTENSION(QQmlComponentExtension, componentExtension);
-
-/*
-    Try to do what's necessary for a reasonable display of the type
-    name, but no more (just enough for the client to do more extensive cleanup).
-
-    Should only be called when debugging is enabled.
-*/
-static inline QString buildTypeNameForDebug(const QMetaObject *metaObject)
-{
-    static const QString qmlMarker(QLatin1String("_QML"));
-    static const QChar underscore(QLatin1Char('_'));
-    static const QChar asterisk(QLatin1Char('*'));
-    QQmlType *type = QQmlMetaType::qmlType(metaObject);
-    QString typeName = type ? type->qmlTypeName() : QString::fromUtf8(metaObject->className());
-    if (!type) {
-        //### optimize further?
-        int marker = typeName.indexOf(qmlMarker);
-        if (marker != -1 && marker < typeName.count() - 1) {
-            if (typeName[marker + 1] == underscore) {
-                const QString className = typeName.left(marker) + asterisk;
-                type = QQmlMetaType::qmlType(QMetaType::type(className.toUtf8()));
-                if (type)
-                    typeName = type->qmlTypeName();
-            }
-        }
-    }
-    return typeName;
-}
+V4_DEFINE_EXTENSION(QQmlComponentExtension, componentExtension);
 
 /*!
     \class QQmlComponent
     \since 5.0
     \inmodule QtQml
-    \mainclass
 
     \brief The QQmlComponent class encapsulates a QML component definition
 
@@ -167,6 +119,26 @@ static inline QString buildTypeNameForDebug(const QMetaObject *metaObject)
     int width = item->width();  // width = 200
     \endcode
 
+    To create instances of a component in code where a QQmlEngine instance is
+    not available, you can use \l qmlContext() or \l qmlEngine(). For example,
+    in the scenario below, child items are being created within a QQuickItem
+    subclass:
+
+    \code
+    void MyCppItem::init()
+    {
+        QQmlEngine *engine = qmlEngine(this);
+        // Or:
+        // QQmlEngine *engine = qmlContext(this)->engine();
+        QQmlComponent component(engine, QUrl::fromLocalFile("MyItem.qml"));
+        QQuickItem *childItem = qobject_cast<QQuickItem*>(component.create());
+        childItem->setParentItem(this);
+    }
+    \endcode
+
+    Note that these functions will return \c null when called inside the
+    constructor of a QObject subclass, as the instance will not yet have
+    a context nor engine.
 
     \section2 Network Components
 
@@ -265,22 +237,22 @@ static inline QString buildTypeNameForDebug(const QMetaObject *metaObject)
     \table
     \row
     \li MyItem.qml
-    \li main.qml
-    \row
     \li \snippet qml/component/MyItem.qml 0
+    \row
+    \li main.qml
     \li \snippet qml/component/main.qml 0
     \endtable
 */
 
 /*!
-    \qmlattachedsignal Component::onCompleted()
+    \qmlattachedsignal Component::completed()
 
-    Emitted after component "startup" has completed. This can be used to
+    Emitted after the object has been instantiated. This can be used to
     execute script code at startup, once the full QML environment has been
     established.
 
-    The \c {Component::onCompleted} attached property can be declared on
-    any object. The order of running the \c onCompleted scripts is
+    The corresponding handler is \c onCompleted. It can be declared on
+    any object. The order of running the \c onCompleted handlers is
     undefined.
 
     \qml
@@ -294,16 +266,15 @@ static inline QString buildTypeNameForDebug(const QMetaObject *metaObject)
 */
 
 /*!
-    \qmlattachedsignal Component::onDestruction()
+    \qmlattachedsignal Component::destruction()
 
-    Emitted as the component begins destruction. This can be used to undo
-    work done in the onCompleted signal, or other imperative code in your
-    application.
+    Emitted as the object begins destruction. This can be used to undo
+    work done in response to the \l {completed}{completed()} signal, or other
+    imperative code in your application.
 
-    The \c {Component::onDestruction} attached property can be declared on
-    any object. However, it applies to the destruction of the component as
-    a whole, and not the destruction of the specific object. The order of
-    running the \c onDestruction scripts is undefined.
+    The corresponding handler is \c onDestruction. It can be declared on
+    any object. The order of running the \c onDestruction handlers is
+    undefined.
 
     \qml
     Rectangle {
@@ -325,7 +296,7 @@ static inline QString buildTypeNameForDebug(const QMetaObject *metaObject)
     \value Null This QQmlComponent has no data. Call loadUrl() or setData() to add QML content.
     \value Ready This QQmlComponent is ready and create() may be called.
     \value Loading This QQmlComponent is loading network data.
-    \value Error An error has occurred. Call errors() to retrieve a list of \{QQmlError}{errors}.
+    \value Error An error has occurred. Call errors() to retrieve a list of \l {QQmlError}{errors}.
 */
 
 /*!
@@ -408,6 +379,13 @@ QQmlComponent::~QQmlComponent()
 
     if (d->state.completePending) {
         qWarning("QQmlComponent: Component destroyed while completion pending");
+
+        if (isError()) {
+            qWarning() << "This may have been caused by one of the following errors:";
+            foreach (const QQmlError &error, d->state.errors)
+                qWarning().nospace().noquote() << QLatin1String("    ") << error;
+        }
+
         d->completeCreate();
     }
 
@@ -602,7 +580,7 @@ QQmlComponent::QQmlComponent(QQmlEngine *engine, QQmlCompiledData *cc, int start
     d->cc = cc;
     cc->addref();
     d->start = start;
-    d->url = cc->url;
+    d->url = cc->url();
     d->progress = 1.0;
 }
 
@@ -694,9 +672,9 @@ void QQmlComponentPrivate::loadUrl(const QUrl &newUrl, QQmlComponent::Compilatio
         emit q->progressChanged(progress);
     }
 
-    QQmlDataLoader::Mode loaderMode = (mode == QQmlComponent::Asynchronous)
-            ? QQmlDataLoader::Asynchronous
-            : QQmlDataLoader::PreferSynchronous;
+    QQmlTypeLoader::Mode loaderMode = (mode == QQmlComponent::Asynchronous)
+            ? QQmlTypeLoader::Asynchronous
+            : QQmlTypeLoader::PreferSynchronous;
 
     QQmlTypeData *data = QQmlEnginePrivate::get(engine)->typeLoader.getType(url, loaderMode);
 
@@ -715,7 +693,7 @@ void QQmlComponentPrivate::loadUrl(const QUrl &newUrl, QQmlComponent::Compilatio
 }
 
 /*!
-    Return the list of errors that occurred during the last compile or create
+    Returns the list of errors that occurred during the last compile or create
     operation. An empty list is returned if isError() is not set.
 */
 QList<QQmlError> QQmlComponent::errors() const
@@ -790,6 +768,11 @@ QQmlComponent::QQmlComponent(QQmlComponentPrivate &dd, QObject *parent)
 
     The ownership of the returned object instance is transferred to the caller.
 
+    If the object being created from this component is a visual item, it must
+    have a visual parent, which can be set by calling
+    QQuickItem::setParentItem(). See \l {Concepts - Visual Parent in Qt Quick}
+    for more details.
+
     \sa QQmlEngine::ObjectOwnership
 */
 QObject *QQmlComponent::create(QQmlContext *context)
@@ -808,8 +791,8 @@ QObject *QQmlComponent::create(QQmlContext *context)
 
 /*!
     This method provides advanced control over component instance creation.
-    In general, programmers should use QQmlComponent::create() to create a
-    component.
+    In general, programmers should use QQmlComponent::create() to create object
+    instances.
 
     Create an object instance from this component. Returns 0 if creation
     failed. \a publicContext specifies the context within which to create the object
@@ -884,26 +867,16 @@ QQmlComponentPrivate::beginCreate(QQmlContextData *context)
 
     QQmlEnginePrivate *enginePriv = QQmlEnginePrivate::get(engine);
 
-    if (enginePriv->inProgressCreations == 0) {
-        // only track root, since further ones might not be properly nested
-        profiler = new QQmlObjectCreatingProfiler();
-    }
-
     enginePriv->inProgressCreations++;
     state.errors.clear();
     state.completePending = true;
 
     enginePriv->referenceScarceResources();
     QObject *rv = 0;
-    if (enginePriv->useNewCompiler) {
-        state.creator = new QmlObjectCreator(context, cc);
-        rv = state.creator->create(start);
-        if (!rv)
-            state.errors = state.creator->errors;
-    } else {
-        state.vme.init(context, cc, start, creationContext);
-        rv = state.vme.execute(&state.errors);
-    }
+    state.creator.reset(new QQmlObjectCreator(context, cc, creationContext));
+    rv = state.creator->create(start);
+    if (!rv)
+        state.errors = state.creator->errors;
     enginePriv->dereferenceScarceResources();
 
     if (rv) {
@@ -920,17 +893,11 @@ QQmlComponentPrivate::beginCreate(QQmlContextData *context)
         depthIncreased = false;
     }
 
-    if (enginePriv->isDebugging && rv) {
+    QQmlEngineDebugService *service = QQmlDebugConnector::service<QQmlEngineDebugService>();
+    if (service && rv) {
         if (!context->isInternal)
             context->asQQmlContextPrivate()->instances.append(rv);
-        QQmlEngineDebugService::instance()->objectCreated(engine, rv);
-
-        if (profiler && profiler->enabled) {
-            profiler->setTypeName(buildTypeNameForDebug(rv->metaObject()));
-            QQmlData *data = QQmlData::get(rv);
-            Q_ASSERT(data);
-            profiler->setLocation(cc->url, data->lineNumber, data->columnNumber);
-        }
+        service->objectCreated(engine, rv);
     }
 
     return rv;
@@ -943,22 +910,20 @@ void QQmlComponentPrivate::beginDeferred(QQmlEnginePrivate *enginePriv,
     state->errors.clear();
     state->completePending = true;
 
-    if (enginePriv->useNewCompiler) {
-        // ###
-    } else {
-        state->vme.initDeferred(object);
-        state->vme.execute(&state->errors);
-    }
+    QQmlData *ddata = QQmlData::get(object);
+    Q_ASSERT(ddata->deferredData);
+    QQmlData::DeferredData *deferredData = ddata->deferredData;
+    QQmlContextData *creationContext = 0;
+    state->creator.reset(new QQmlObjectCreator(deferredData->context->parent, deferredData->compiledData, creationContext));
+    if (!state->creator->populateDeferredProperties(object))
+        state->errors << state->creator->errors;
 }
 
 void QQmlComponentPrivate::complete(QQmlEnginePrivate *enginePriv, ConstructionState *state)
 {
     if (state->completePending) {
-        if (enginePriv->useNewCompiler) {
-            state->creator->finalize();
-        } else {
-            state->vme.complete();
-        }
+        QQmlInstantiationInterrupt interrupt;
+        state->creator->finalize(interrupt);
 
         state->completePending = false;
 
@@ -995,9 +960,6 @@ void QQmlComponentPrivate::completeCreate()
     if (state.completePending) {
         QQmlEnginePrivate *ep = QQmlEnginePrivate::get(engine);
         complete(ep, &state);
-
-        delete profiler;
-        profiler = 0;
     }
 
     if (depthIncreased) {
@@ -1032,10 +994,8 @@ QQmlComponentAttached *QQmlComponent::qmlAttachedProperties(QObject *obj)
         return a;
 
     QQmlEnginePrivate *p = QQmlEnginePrivate::get(engine);
-    if (p->activeVME) { // XXX should only be allowed during begin
-        a->add(&p->activeVME->componentAttached);
-    } else if (p->activeObjectCreator) {
-        a->add(&p->activeObjectCreator->componentAttached);
+    if (p->activeObjectCreator) { // XXX should only be allowed during begin
+        a->add(p->activeObjectCreator->componentAttachment());
     } else {
         QQmlData *d = QQmlData::get(obj);
         Q_ASSERT(d);
@@ -1100,18 +1060,33 @@ void QQmlComponent::create(QQmlIncubator &incubator, QQmlContext *context,
 
     p->compiledData = d->cc;
     p->compiledData->addref();
-    p->vme.init(contextData, d->cc, d->start, d->creationContext);
+    p->creator.reset(new QQmlObjectCreator(contextData, d->cc, d->creationContext, p.data()));
+    p->subComponentToCreate = d->start;
 
     enginePriv->incubate(incubator, forContextData);
 }
 
 class QQmlComponentIncubator;
 
-class QmlIncubatorObject : public QV4::Object
+namespace QV4 {
+
+namespace Heap {
+
+struct QmlIncubatorObject : Object {
+    QmlIncubatorObject(QQmlIncubator::IncubationMode = QQmlIncubator::Asynchronous);
+    QScopedPointer<QQmlComponentIncubator> incubator;
+    QPointer<QObject> parent;
+    QV4::Value valuemap;
+    QV4::Value statusChanged;
+    Pointer<Heap::QmlContext> qmlContext;
+};
+
+}
+
+struct QmlIncubatorObject : public QV4::Object
 {
-    Q_MANAGED
-public:
-    QmlIncubatorObject(QV8Engine *engine, QQmlIncubator::IncubationMode = QQmlIncubator::Asynchronous);
+    V4_OBJECT2(QmlIncubatorObject, Object)
+    V4_NEEDS_DESTROY
 
     static QV4::ReturnedValue method_get_statusChanged(QV4::CallContext *ctx);
     static QV4::ReturnedValue method_set_statusChanged(QV4::CallContext *ctx);
@@ -1119,39 +1094,37 @@ public:
     static QV4::ReturnedValue method_get_object(QV4::CallContext *ctx);
     static QV4::ReturnedValue method_forceCompletion(QV4::CallContext *ctx);
 
-    static void destroy(Managed *that);
-    static void markObjects(Managed *that, QV4::ExecutionEngine *e);
-
-    QScopedPointer<QQmlComponentIncubator> incubator;
-    QV8Engine *v8;
-    QPointer<QObject> parent;
-    QV4::SafeValue valuemap;
-    QV4::SafeValue qmlGlobal;
-    QV4::SafeValue m_statusChanged;
+    static void markObjects(QV4::Heap::Base *that, QV4::ExecutionEngine *e);
 
     void statusChanged(QQmlIncubator::Status);
     void setInitialState(QObject *);
 };
 
-DEFINE_MANAGED_VTABLE(QmlIncubatorObject);
+}
+
+DEFINE_OBJECT_VTABLE(QV4::QmlIncubatorObject);
 
 class QQmlComponentIncubator : public QQmlIncubator
 {
 public:
-    QQmlComponentIncubator(QmlIncubatorObject *inc, IncubationMode mode)
+    QQmlComponentIncubator(QV4::Heap::QmlIncubatorObject *inc, IncubationMode mode)
         : QQmlIncubator(mode)
         , incubatorObject(inc)
     {}
 
     virtual void statusChanged(Status s) {
-        incubatorObject->statusChanged(s);
+        QV4::Scope scope(incubatorObject->internalClass->engine);
+        QV4::Scoped<QV4::QmlIncubatorObject> i(scope, incubatorObject);
+        i->statusChanged(s);
     }
 
     virtual void setInitialState(QObject *o) {
-        incubatorObject->setInitialState(o);
+        QV4::Scope scope(incubatorObject->internalClass->engine);
+        QV4::Scoped<QV4::QmlIncubatorObject> i(scope, incubatorObject);
+        i->setInitialState(o);
     }
 
-    QmlIncubatorObject *incubatorObject;
+    QV4::Heap::QmlIncubatorObject *incubatorObject;
 };
 
 
@@ -1219,6 +1192,49 @@ static void QQmlComponent_setQmlParent(QObject *me, QObject *parent)
     \sa incubateObject()
 */
 
+
+static void setInitialProperties(QV4::ExecutionEngine *engine, QV4::QmlContext *qmlContext, const QV4::Value &o, const QV4::Value &v)
+{
+    QV4::Scope scope(engine);
+    QV4::ScopedObject object(scope);
+    QV4::ScopedObject valueMap(scope, v);
+    QV4::ObjectIterator it(scope, valueMap, QV4::ObjectIterator::EnumerableOnly|QV4::ObjectIterator::WithProtoChain);
+    QV4::ScopedString name(scope);
+    QV4::ScopedValue val(scope);
+    if (engine->hasException)
+        return;
+
+    QV4::ExecutionContextSaver saver(scope);
+    engine->pushContext(qmlContext);
+
+    while (1) {
+        name = it.nextPropertyNameAsString(val);
+        if (!name)
+            break;
+        object = o;
+        const QStringList properties = name->toQString().split(QLatin1Char('.'));
+        for (int i = 0; i < properties.length() - 1; ++i) {
+            name = engine->newString(properties.at(i));
+            object = object->get(name);
+            if (engine->hasException || !object) {
+                break;
+            }
+        }
+        if (engine->hasException || !object) {
+            engine->hasException = false;
+            continue;
+        }
+        name = engine->newString(properties.last());
+        object->put(name, val);
+        if (engine->hasException) {
+            engine->hasException = false;
+            continue;
+        }
+    }
+
+    engine->hasException = false;
+}
+
 /*!
     \internal
 */
@@ -1241,7 +1257,7 @@ void QQmlComponent::createObject(QQmlV4Function *args)
 
     if (args->length() >= 2) {
         QV4::ScopedValue v(scope, (*args)[1]);
-        if (!v->asObject() || v->asArrayObject()) {
+        if (!v->as<QV4::Object>() || v->as<QV4::ArrayObject>()) {
             qmlInfo(this) << tr("createObject: value is not an object");
             args->setReturnValue(QV4::Encode::null());
             return;
@@ -1265,14 +1281,8 @@ void QQmlComponent::createObject(QQmlV4Function *args)
     Q_ASSERT(object->isObject());
 
     if (!valuemap->isUndefined()) {
-        QV4::ScopedObject qmlglobal(scope, args->qmlGlobal());
-        QV4::ScopedValue f(scope, QV4::Script::evaluate(v4, QString::fromLatin1(INITIALPROPERTIES_SOURCE), qmlglobal));
-        Q_ASSERT(f->asFunctionObject());
-        QV4::ScopedCallData callData(scope, 2);
-        callData->thisObject = v4->globalObject;
-        callData->args[0] = object;
-        callData->args[1] = valuemap;
-        f->asFunctionObject()->call(callData);
+        QV4::Scoped<QV4::QmlContext> qmlContext(scope, v4->qmlContext());
+        setInitialProperties(v4, qmlContext, object, valuemap);
     }
 
     d->completeCreate();
@@ -1281,10 +1291,7 @@ void QQmlComponent::createObject(QQmlV4Function *args)
     QQmlData::get(rv)->explicitIndestructibleSet = false;
     QQmlData::get(rv)->indestructible = false;
 
-    if (!rv)
-        args->setReturnValue(QV4::Encode::null());
-    else
-        args->setReturnValue(object.asReturnedValue());
+    args->setReturnValue(object->asReturnedValue());
 }
 
 /*!
@@ -1367,7 +1374,7 @@ void QQmlComponent::incubateObject(QQmlV4Function *args)
     if (args->length() >= 2) {
         QV4::ScopedValue v(scope, (*args)[1]);
         if (v->isNull()) {
-        } else if (!v->asObject() || v->asArrayObject()) {
+        } else if (!v->as<QV4::Object>() || v->as<QV4::ArrayObject>()) {
             qmlInfo(this) << tr("createObject: value is not an object");
             args->setReturnValue(QV4::Encode::null());
             return;
@@ -1385,19 +1392,18 @@ void QQmlComponent::incubateObject(QQmlV4Function *args)
             mode = QQmlIncubator::AsynchronousIfNested;
     }
 
-    QQmlComponentExtension *e = componentExtension(args->engine());
+    QQmlComponentExtension *e = componentExtension(args->v4engine());
 
-    QV4::Scoped<QmlIncubatorObject> r(scope, new (v4->memoryManager) QmlIncubatorObject(args->engine(), mode));
+    QV4::Scoped<QV4::QmlIncubatorObject> r(scope, v4->memoryManager->allocObject<QV4::QmlIncubatorObject>(mode));
     QV4::ScopedObject p(scope, e->incubationProto.value());
-    r->setPrototype(p.getPointer());
+    r->setPrototype(p);
 
-    if (!valuemap->isUndefined()) {
-        r->valuemap = valuemap;
-        r->qmlGlobal = args->qmlGlobal();
-    }
-    r->parent = parent;
+    if (!valuemap->isUndefined())
+        r->d()->valuemap = valuemap;
+    r->d()->qmlContext = v4->qmlContext();
+    r->d()->parent = parent;
 
-    QQmlIncubator *incubator = r.getPointer()->incubator.data();
+    QQmlIncubator *incubator = r->d()->incubator.data();
     create(*incubator, creationContext());
 
     if (incubator->status() == QQmlIncubator::Null) {
@@ -1408,93 +1414,83 @@ void QQmlComponent::incubateObject(QQmlV4Function *args)
 }
 
 // XXX used by QSGLoader
-void QQmlComponentPrivate::initializeObjectWithInitialProperties(const QV4::ValueRef qmlGlobal, const QV4::ValueRef valuemap, QObject *toCreate)
+void QQmlComponentPrivate::initializeObjectWithInitialProperties(QV4::QmlContext *qmlContext, const QV4::Value &valuemap, QObject *toCreate)
 {
     QQmlEnginePrivate *ep = QQmlEnginePrivate::get(engine);
-    QV8Engine *v8engine = ep->v8engine();
-    QV4::ExecutionEngine *v4engine = QV8Engine::getV4(v8engine);
+    QV4::ExecutionEngine *v4engine = QV8Engine::getV4(ep->v8engine());
     QV4::Scope scope(v4engine);
 
     QV4::ScopedValue object(scope, QV4::QObjectWrapper::wrap(v4engine, toCreate));
-    Q_ASSERT(object->asObject());
+    Q_ASSERT(object->as<QV4::Object>());
 
-    if (!valuemap->isUndefined()) {
-        QV4::ScopedObject qmlGlobalObj(scope, qmlGlobal);
-        QV4::Scoped<QV4::FunctionObject>  f(scope, QV4::Script::evaluate(QV8Engine::getV4(v8engine),
-                                                                    QString::fromLatin1(INITIALPROPERTIES_SOURCE), qmlGlobalObj));
-        QV4::ScopedCallData callData(scope, 2);
-        callData->thisObject = v4engine->globalObject;
-        callData->args[0] = object;
-        callData->args[1] = valuemap;
-        f->call(callData);
-    }
+    if (!valuemap.isUndefined())
+        setInitialProperties(v4engine, qmlContext, object, valuemap);
 }
 
-QQmlComponentExtension::QQmlComponentExtension(QV8Engine *engine)
+QQmlComponentExtension::QQmlComponentExtension(QV4::ExecutionEngine *v4)
 {
-    QV4::ExecutionEngine *v4 = QV8Engine::getV4(engine);
     QV4::Scope scope(v4);
-    QV4::Scoped<QV4::Object> proto(scope, v4->newObject());
+    QV4::ScopedObject proto(scope, v4->newObject());
     proto->defineAccessorProperty(QStringLiteral("onStatusChanged"),
-                                  QmlIncubatorObject::method_get_statusChanged, QmlIncubatorObject::method_set_statusChanged);
-    proto->defineAccessorProperty(QStringLiteral("status"), QmlIncubatorObject::method_get_status, 0);
-    proto->defineAccessorProperty(QStringLiteral("object"), QmlIncubatorObject::method_get_object, 0);
-    proto->defineDefaultProperty(QStringLiteral("forceCompletion"), QmlIncubatorObject::method_forceCompletion);
+                                  QV4::QmlIncubatorObject::method_get_statusChanged, QV4::QmlIncubatorObject::method_set_statusChanged);
+    proto->defineAccessorProperty(QStringLiteral("status"), QV4::QmlIncubatorObject::method_get_status, 0);
+    proto->defineAccessorProperty(QStringLiteral("object"), QV4::QmlIncubatorObject::method_get_object, 0);
+    proto->defineDefaultProperty(QStringLiteral("forceCompletion"), QV4::QmlIncubatorObject::method_forceCompletion);
 
-    incubationProto = proto;
+    incubationProto.set(v4, proto);
 }
 
-QV4::ReturnedValue QmlIncubatorObject::method_get_object(QV4::CallContext *ctx)
+QV4::ReturnedValue QV4::QmlIncubatorObject::method_get_object(QV4::CallContext *ctx)
 {
     QV4::Scope scope(ctx);
-    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->callData->thisObject.as<QmlIncubatorObject>());
+    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->thisObject().as<QmlIncubatorObject>());
     if (!o)
-        return ctx->throwTypeError();
+        return ctx->engine()->throwTypeError();
 
-    return QV4::QObjectWrapper::wrap(ctx->engine, o->incubator->object());
+    return QV4::QObjectWrapper::wrap(ctx->d()->engine, o->d()->incubator->object());
 }
 
-QV4::ReturnedValue QmlIncubatorObject::method_forceCompletion(QV4::CallContext *ctx)
+QV4::ReturnedValue QV4::QmlIncubatorObject::method_forceCompletion(QV4::CallContext *ctx)
 {
     QV4::Scope scope(ctx);
-    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->callData->thisObject.as<QmlIncubatorObject>());
+    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->thisObject().as<QmlIncubatorObject>());
     if (!o)
-        return ctx->throwTypeError();
+        return ctx->engine()->throwTypeError();
 
-    o->incubator->forceCompletion();
+    o->d()->incubator->forceCompletion();
 
     return QV4::Encode::undefined();
 }
 
-QV4::ReturnedValue QmlIncubatorObject::method_get_status(QV4::CallContext *ctx)
+QV4::ReturnedValue QV4::QmlIncubatorObject::method_get_status(QV4::CallContext *ctx)
 {
     QV4::Scope scope(ctx);
-    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->callData->thisObject.as<QmlIncubatorObject>());
+    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->thisObject().as<QmlIncubatorObject>());
     if (!o)
-        return ctx->throwTypeError();
+        return ctx->engine()->throwTypeError();
 
-    return QV4::Encode(o->incubator->status());
+    return QV4::Encode(o->d()->incubator->status());
 }
 
-QV4::ReturnedValue QmlIncubatorObject::method_get_statusChanged(QV4::CallContext *ctx)
+QV4::ReturnedValue QV4::QmlIncubatorObject::method_get_statusChanged(QV4::CallContext *ctx)
 {
     QV4::Scope scope(ctx);
-    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->callData->thisObject.as<QmlIncubatorObject>());
+    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->thisObject().as<QmlIncubatorObject>());
     if (!o)
-        return ctx->throwTypeError();
+        return ctx->engine()->throwTypeError();
 
-    return o->m_statusChanged.asReturnedValue();
+    return o->d()->statusChanged.asReturnedValue();
 }
 
-QV4::ReturnedValue QmlIncubatorObject::method_set_statusChanged(QV4::CallContext *ctx)
+QV4::ReturnedValue QV4::QmlIncubatorObject::method_set_statusChanged(QV4::CallContext *ctx)
 {
     QV4::Scope scope(ctx);
-    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->callData->thisObject.as<QmlIncubatorObject>());
-    if (!o || ctx->callData->argc < 1)
-        return ctx->throwTypeError();
+    QV4::Scoped<QmlIncubatorObject> o(scope, ctx->thisObject().as<QmlIncubatorObject>());
+    if (!o || ctx->argc() < 1)
+        return ctx->engine()->throwTypeError();
 
 
-    o->m_statusChanged = ctx->callData->args[0];
+    o->d()->statusChanged = ctx->args()[0];
     return QV4::Encode::undefined();
 }
 
@@ -1502,75 +1498,58 @@ QQmlComponentExtension::~QQmlComponentExtension()
 {
 }
 
-QmlIncubatorObject::QmlIncubatorObject(QV8Engine *engine, QQmlIncubator::IncubationMode m)
-    : Object(QV8Engine::getV4(engine))
+QV4::Heap::QmlIncubatorObject::QmlIncubatorObject(QQmlIncubator::IncubationMode m)
+    : valuemap(QV4::Primitive::undefinedValue())
+    , statusChanged(QV4::Primitive::undefinedValue())
+    , qmlContext(0)
 {
     incubator.reset(new QQmlComponentIncubator(this, m));
-    v8 = engine;
-    vtbl = &static_vtbl;
-
-    valuemap = QV4::Primitive::undefinedValue();
-    qmlGlobal = QV4::Primitive::undefinedValue();
-    m_statusChanged = QV4::Primitive::undefinedValue();
 }
 
-void QmlIncubatorObject::setInitialState(QObject *o)
+void QV4::QmlIncubatorObject::setInitialState(QObject *o)
 {
-    QQmlComponent_setQmlParent(o, parent);
+    QQmlComponent_setQmlParent(o, d()->parent);
 
-    if (!valuemap.isUndefined()) {
-        QV4::ExecutionEngine *v4 = QV8Engine::getV4(v8);
+    if (!d()->valuemap.isUndefined()) {
+        QV4::ExecutionEngine *v4 = engine();
         QV4::Scope scope(v4);
-
-        QV4::Scoped<QV4::FunctionObject> f(scope, QV4::Script::evaluate(v4, QString::fromLatin1(INITIALPROPERTIES_SOURCE), qmlGlobal));
-        QV4::ScopedCallData callData(scope, 2);
-        callData->thisObject = v4->globalObject;
-        callData->args[0] = QV4::QObjectWrapper::wrap(v4, o);
-        callData->args[1] = valuemap;
-        f->call(callData);
+        QV4::ScopedObject obj(scope, QV4::QObjectWrapper::wrap(v4, o));
+        QV4::Scoped<QV4::QmlContext> qmlCtxt(scope, d()->qmlContext);
+        setInitialProperties(v4, qmlCtxt, obj, d()->valuemap);
     }
 }
 
-void QmlIncubatorObject::destroy(Managed *that)
+void QV4::QmlIncubatorObject::markObjects(QV4::Heap::Base *that, QV4::ExecutionEngine *e)
 {
-    QmlIncubatorObject *o = that->as<QmlIncubatorObject>();
-    Q_ASSERT(o);
-    o->~QmlIncubatorObject();
-}
-
-void QmlIncubatorObject::markObjects(QV4::Managed *that, QV4::ExecutionEngine *e)
-{
-    QmlIncubatorObject *o = that->as<QmlIncubatorObject>();
-    Q_ASSERT(o);
+    QmlIncubatorObject::Data *o = static_cast<QmlIncubatorObject::Data *>(that);
     o->valuemap.mark(e);
-    o->qmlGlobal.mark(e);
-    o->m_statusChanged.mark(e);
+    o->statusChanged.mark(e);
+    if (o->qmlContext)
+        o->qmlContext->mark(e);
     Object::markObjects(that, e);
 }
 
-void QmlIncubatorObject::statusChanged(QQmlIncubator::Status s)
+void QV4::QmlIncubatorObject::statusChanged(QQmlIncubator::Status s)
 {
-    QV4::Scope scope(QV8Engine::getV4(v8));
+    QV4::Scope scope(engine());
     // hold the incubated object in a scoped value to prevent it's destruction before this method returns
-    QV4::ScopedObject incubatedObject(scope, QV4::QObjectWrapper::wrap(scope.engine, incubator->object()));
+    QV4::ScopedObject incubatedObject(scope, QV4::QObjectWrapper::wrap(scope.engine, d()->incubator->object()));
 
     if (s == QQmlIncubator::Ready) {
-        Q_ASSERT(QQmlData::get(incubator->object()));
-        QQmlData::get(incubator->object())->explicitIndestructibleSet = false;
-        QQmlData::get(incubator->object())->indestructible = false;
+        Q_ASSERT(QQmlData::get(d()->incubator->object()));
+        QQmlData::get(d()->incubator->object())->explicitIndestructibleSet = false;
+        QQmlData::get(d()->incubator->object())->indestructible = false;
     }
 
-    QV4::ScopedFunctionObject f(scope, m_statusChanged);
+    QV4::ScopedFunctionObject f(scope, d()->statusChanged);
     if (f) {
-        QV4::ExecutionContext *ctx = scope.engine->current;
         QV4::ScopedCallData callData(scope, 1);
         callData->thisObject = this;
         callData->args[0] = QV4::Primitive::fromUInt32(s);
         f->call(callData);
         if (scope.hasException()) {
-            ctx->catchException();
-            QQmlError error = QV4::ExecutionEngine::catchExceptionAsQmlError(ctx);
-            QQmlEnginePrivate::warning(QQmlEnginePrivate::get(v8->engine()), error);
+            QQmlError error = scope.engine->catchExceptionAsQmlError();
+            QQmlEnginePrivate::warning(QQmlEnginePrivate::get(scope.engine->qmlEngine()), error);
         }
     }
 }
