@@ -1,39 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
-** This file is part of the QtQml module of the Qt Toolkit.
+** This file is part of the QtQuick module of the Qt Toolkit.
 **
 ** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -48,8 +46,6 @@
 
 #include <private/qqmlcomponent_p.h>
 
-#include <private/qv8_p.h>
-
 QT_BEGIN_NAMESPACE
 
 static const QQuickItemPrivate::ChangeTypes watchedChanges
@@ -108,6 +104,7 @@ void QQuickLoaderPrivate::clear()
         component->deleteLater();
         component = 0;
     }
+    componentStrongReference.clear();
     source = QUrl();
 
     if (item) {
@@ -160,7 +157,7 @@ qreal QQuickLoaderPrivate::getImplicitHeight() const
 /*!
     \qmltype Loader
     \instantiates QQuickLoader
-    \inqmlmodule QtQuick 2
+    \inqmlmodule QtQuick
     \ingroup qtquick-dynamic
     \inherits Item
 
@@ -226,7 +223,7 @@ qreal QQuickLoaderPrivate::getImplicitHeight() const
     loads \c MyItem.qml, and is able to receive the \c message signal from
     the loaded item through a \l Connections object:
 
-    \table
+    \table 70%
     \row
     \li application.qml
     \li MyItem.qml
@@ -272,7 +269,6 @@ qreal QQuickLoaderPrivate::getImplicitHeight() const
 
     In some cases you may wish to use a Loader within a view delegate to improve delegate
     loading performance. This works well in most cases, but there is one important issue to
-    be aware of related to the \l{QtQuick2::Component#creation-context}{creation context} of a Component.
 
     In the following example, the \c index context property inserted by the ListView into \c delegateComponent's
     context will be inaccessible to Text, as the Loader will use the creation context of \c myComponent as the parent
@@ -312,7 +308,7 @@ QQuickLoader::~QQuickLoader()
 }
 
 /*!
-    \qmlproperty bool QtQuick2::Loader::active
+    \qmlproperty bool QtQuick::Loader::active
     This property is \c true if the Loader is currently active.
     The default value for this property is \c true.
 
@@ -375,7 +371,7 @@ void QQuickLoader::setActive(bool newVal)
 
 
 /*!
-    \qmlproperty url QtQuick2::Loader::source
+    \qmlproperty url QtQuick::Loader::source
     This property holds the URL of the QML component to instantiate.
 
     Since \c {QtQuick 2.0}, Loader is able to load any type of object; it
@@ -435,7 +431,7 @@ void QQuickLoader::loadFromSource()
 }
 
 /*!
-    \qmlproperty Component QtQuick2::Loader::sourceComponent
+    \qmlproperty Component QtQuick::Loader::sourceComponent
     This property holds the \l{Component} to instantiate.
 
     \qml
@@ -450,8 +446,7 @@ void QQuickLoader::loadFromSource()
     }
     \endqml
 
-    To unload the currently loaded object, set this property to an empty string
-    or \c undefined.
+    To unload the currently loaded object, set this property to \c undefined.
 
     Since \c {QtQuick 2.0}, Loader is able to load any type of object; it
     is not restricted to Item types.
@@ -474,6 +469,10 @@ void QQuickLoader::setSourceComponent(QQmlComponent *comp)
     d->clear();
 
     d->component = comp;
+    if (comp) {
+        if (QQmlData *ddata = QQmlData::get(comp))
+            d->componentStrongReference = ddata->jsWrapper;
+    }
     d->loadingFromSource = false;
 
     if (d->active)
@@ -503,7 +502,7 @@ void QQuickLoader::loadFromSourceComponent()
 }
 
 /*!
-    \qmlmethod object QtQuick2::Loader::setSource(url source, object properties)
+    \qmlmethod object QtQuick::Loader::setSource(url source, object properties)
 
     Creates an object instance of the given \a source component that will have
     the given \a properties. The \a properties argument is optional.  The instance
@@ -522,7 +521,7 @@ void QQuickLoader::loadFromSourceComponent()
     is changed after calling this function but prior to setting the loader \l active.
 
     Example:
-    \table
+    \table 70%
     \row
     \li
     \qml
@@ -551,11 +550,13 @@ void QQuickLoader::loadFromSourceComponent()
     Item {
         Loader {
             id: squareLoader
-            onLoaded: console.log(squareLoader.item.width); // prints [10], not [30]
+            onLoaded: console.log(squareLoader.item.width);
+            // prints [10], not [30]
         }
 
         Component.onCompleted: {
-            squareLoader.setSource("ExampleComponent.qml", { "color": "blue" });
+            squareLoader.setSource("ExampleComponent.qml",
+                                 { "color": "blue" });
             // will trigger the onLoaded code when complete.
         }
     }
@@ -564,34 +565,31 @@ void QQuickLoader::loadFromSourceComponent()
 
     \sa source, active
 */
-void QQuickLoader::setSource(QQmlV8Function *args)
+void QQuickLoader::setSource(QQmlV4Function *args)
 {
     Q_ASSERT(args);
     Q_D(QQuickLoader);
 
     bool ipvError = false;
-    args->returnValue(v8::Undefined());
-    v8::Handle<v8::Object> ipv = d->extractInitialPropertyValues(args, this, &ipvError);
+    args->setReturnValue(QV4::Encode::undefined());
+    QV4::Scope scope(args->v4engine());
+    QV4::ScopedValue ipv(scope, d->extractInitialPropertyValues(args, this, &ipvError));
     if (ipvError)
         return;
 
     d->clear();
     QUrl sourceUrl = d->resolveSourceUrl(args);
-    if (!ipv.IsEmpty()) {
+    if (!ipv->isUndefined()) {
         d->disposeInitialPropertyValues();
-        d->initialPropertyValues = qPersistentNew(ipv);
-        d->qmlGlobalForIpv = qPersistentNew(args->qmlGlobal());
+        d->initialPropertyValues.set(args->v4engine(), ipv);
     }
+    d->qmlCallingContext.set(scope.engine, scope.engine->qmlContext());
 
     setSource(sourceUrl, false); // already cleared and set ipv above.
 }
 
 void QQuickLoaderPrivate::disposeInitialPropertyValues()
 {
-    if (!initialPropertyValues.IsEmpty())
-        qPersistentDispose(initialPropertyValues);
-    if (!qmlGlobalForIpv.IsEmpty())
-        qPersistentDispose(qmlGlobalForIpv);
 }
 
 void QQuickLoaderPrivate::load()
@@ -645,12 +643,17 @@ void QQuickLoaderPrivate::setInitialState(QObject *obj)
         itemContext = 0;
     }
 
-    if (initialPropertyValues.IsEmpty())
+    if (initialPropertyValues.isUndefined())
         return;
 
     QQmlComponentPrivate *d = QQmlComponentPrivate::get(component);
     Q_ASSERT(d && d->engine);
-    d->initializeObjectWithInitialProperties(qmlGlobalForIpv, initialPropertyValues, obj);
+    QV4::ExecutionEngine *v4 = QV8Engine::getV4(d->engine);
+    Q_ASSERT(v4);
+    QV4::Scope scope(v4);
+    QV4::ScopedValue ipv(scope, initialPropertyValues.value());
+    QV4::Scoped<QV4::QmlContext> qmlContext(scope, qmlCallingContext.value());
+    d->initializeObjectWithInitialProperties(qmlContext, ipv, obj);
 }
 
 void QQuickLoaderIncubator::statusChanged(Status status)
@@ -722,7 +725,7 @@ void QQuickLoaderPrivate::_q_sourceLoaded()
 }
 
 /*!
-    \qmlproperty enumeration QtQuick2::Loader::status
+    \qmlproperty enumeration QtQuick::Loader::status
 
     This property holds the status of QML loading.  It can be one of:
     \list
@@ -812,15 +815,17 @@ void QQuickLoader::componentComplete()
 }
 
 /*!
-    \qmlsignal QtQuick2::Loader::onLoaded()
+    \qmlsignal QtQuick::Loader::loaded()
 
-    This handler is called when the \l status becomes \c Loader.Ready, or on successful
+    This signal is emitted when the \l status becomes \c Loader.Ready, or on successful
     initial load.
+
+    The corresponding handler is \c onLoaded.
 */
 
 
 /*!
-\qmlproperty real QtQuick2::Loader::progress
+\qmlproperty real QtQuick::Loader::progress
 
 This property holds the progress of loading QML data from the network, from
 0.0 (nothing loaded) to 1.0 (finished).  Most QML files are quite small, so
@@ -842,7 +847,7 @@ qreal QQuickLoader::progress() const
 }
 
 /*!
-\qmlproperty bool QtQuick2::Loader::asynchronous
+\qmlproperty bool QtQuick::Loader::asynchronous
 
 This property holds whether the component will be instantiated asynchronously.
 
@@ -855,6 +860,11 @@ likelihood of glitches in animation.  When loading asynchronously the status
 will change to Loader.Loading.  Once the entire component has been created, the
 \l item will be available and the status will change to Loader.Ready.
 
+Changing the value of this property to \c false while an asynchronous load is in
+progress will force immediate, synchronous completion.  This allows beginning an
+asynchronous load and then forcing completion if the Loader content must be
+accessed before the asynchronous load has completed.
+
 To avoid seeing the items loading progressively set \c visible appropriately, e.g.
 
 \code
@@ -881,6 +891,19 @@ void QQuickLoader::setAsynchronous(bool a)
         return;
 
     d->asynchronous = a;
+
+    if (!d->asynchronous && isComponentComplete() && d->active) {
+        if (d->loadingFromSource && d->component && d->component->isLoading()) {
+            // Force a synchronous component load
+            QUrl currentSource = d->source;
+            d->clear();
+            d->source = currentSource;
+            loadFromSource();
+        } else if (d->incubator && d->incubator->isLoading()) {
+            d->incubator->forceCompletion();
+        }
+    }
+
     emit asynchronousChanged();
 }
 
@@ -906,7 +929,7 @@ void QQuickLoaderPrivate::_q_updateSize(bool loaderGeometryChanged)
 }
 
 /*!
-    \qmlproperty object QtQuick2::Loader::item
+    \qmlproperty object QtQuick::Loader::item
     This property holds the top-level object that is currently loaded.
 
     Since \c {QtQuick 2.0}, Loader can load any object type.
@@ -926,33 +949,35 @@ void QQuickLoader::geometryChanged(const QRectF &newGeometry, const QRectF &oldG
     QQuickItem::geometryChanged(newGeometry, oldGeometry);
 }
 
-QUrl QQuickLoaderPrivate::resolveSourceUrl(QQmlV8Function *args)
+QUrl QQuickLoaderPrivate::resolveSourceUrl(QQmlV4Function *args)
 {
-    QV8Engine *v8engine = args->engine();
-    QString arg = v8engine->toString((*args)[0]->ToString());
+    QV4::Scope scope(args->v4engine());
+    QV4::ScopedValue v(scope, (*args)[0]);
+    QString arg = v->toQString();
     if (arg.isEmpty())
         return QUrl();
 
-    QQmlContextData *context = args->context();
+    QQmlContextData *context = scope.engine->callingQmlContext();
     Q_ASSERT(context);
     return context->resolvedUrl(QUrl(arg));
 }
 
-v8::Handle<v8::Object> QQuickLoaderPrivate::extractInitialPropertyValues(QQmlV8Function *args, QObject *loader, bool *error)
+QV4::ReturnedValue QQuickLoaderPrivate::extractInitialPropertyValues(QQmlV4Function *args, QObject *loader, bool *error)
 {
-    v8::Local<v8::Object> valuemap;
-    if (args->Length() >= 2) {
-        v8::Local<v8::Value> v = (*args)[1];
-        if (!v->IsObject() || v->IsArray()) {
+    QV4::Scope scope(args->v4engine());
+    QV4::ScopedValue valuemap(scope, QV4::Primitive::undefinedValue());
+    if (args->length() >= 2) {
+        QV4::ScopedValue v(scope, (*args)[1]);
+        if (!v->isObject() || v->as<QV4::ArrayObject>()) {
             *error = true;
             qmlInfo(loader) << QQuickLoader::tr("setSource: value is not an object");
         } else {
             *error = false;
-            valuemap = v8::Local<v8::Object>::Cast(v);
+            valuemap = v;
         }
     }
 
-    return valuemap;
+    return valuemap->asReturnedValue();
 }
 
 #include <moc_qquickloader_p.cpp>
