@@ -1,39 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -44,13 +31,15 @@
 #include <QtQuick/qquickitem.h>
 #include <QtQuick/qquickwindow.h>
 #include <QtQuick/qquickview.h>
-#include <QtWidgets/QGraphicsSceneMouseEvent>
 #include "private/qquickfocusscope_p.h"
 #include "private/qquickitem_p.h"
 #include <qpa/qwindowsysteminterface.h>
 #include <QDebug>
 #include <QTimer>
+#include <QQmlEngine>
 #include "../../shared/util.h"
+#include "../shared/viewtestutil.h"
+#include <QSignalSpy>
 
 class TestItem : public QQuickItem
 {
@@ -168,6 +157,18 @@ private slots:
 
     void acceptedMouseButtons();
 
+    void visualParentOwnership();
+    void visualParentOwnershipWindow();
+
+    void testSGInvalidate();
+
+    void objectChildTransform();
+
+    void contains_data();
+    void contains();
+
+    void childAt();
+
 private:
 
     enum PaintOrderOp {
@@ -272,7 +273,7 @@ void tst_qquickitem::simpleFocus()
     QQuickWindow window;
     ensureFocus(&window);
 
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem *l1c1 = new TestItem(window.contentItem());
     QQuickItem *l1c2 = new TestItem(window.contentItem());
@@ -323,7 +324,7 @@ void tst_qquickitem::scopedFocus()
 {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem *l1c1 = new TestItem(window.contentItem());
     QQuickItem *l1c2 = new TestItem(window.contentItem());
@@ -403,7 +404,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem *item = new TestItem;
 
@@ -423,7 +424,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem *item = new TestItem(window.contentItem());
 
@@ -452,7 +453,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem *tree = new TestItem;
     QQuickItem *c1 = new TestItem(tree);
@@ -476,7 +477,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *tree = new TestFocusScope;
     QQuickItem *c1 = new TestItem(tree);
     QQuickItem *c2 = new TestItem(tree);
@@ -504,7 +505,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *tree = new TestFocusScope;
     QQuickItem *c1 = new TestItem(tree);
     QQuickItem *c2 = new TestItem(tree);
@@ -530,7 +531,7 @@ void tst_qquickitem::addedToWindow()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
     QQuickItem *tree = new TestFocusScope;
     QQuickItem *c1 = new TestItem(tree);
@@ -570,7 +571,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
 
     FocusState focusState;
@@ -592,7 +593,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
     QQuickItem *child2 = new TestItem(window.contentItem());
 
@@ -613,7 +614,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
     QQuickItem *child2 = new TestFocusScope(window.contentItem());
     QQuickItem *item = new TestItem(child);
@@ -635,7 +636,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
     QQuickItem *child2 = new TestFocusScope(window.contentItem());
     QQuickItem *item = new TestItem(child2);
@@ -657,7 +658,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *child = new TestItem(window.contentItem());
     QQuickItem *child2 = new TestFocusScope(window.contentItem());
     QQuickItem *item = new TestItem(child2);
@@ -683,7 +684,7 @@ void tst_qquickitem::changeParent()
     {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     QQuickItem *item = new TestFocusScope(window.contentItem());
     QQuickItem *child = new TestItem(item);
     QQuickItem *child2 = new TestItem;
@@ -722,7 +723,7 @@ void tst_qquickitem::multipleFocusClears()
     view.setSource(testFileUrl("multipleFocusClears.qml"));
     view.show();
     ensureFocus(&view);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &view);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &view);
 }
 
 void tst_qquickitem::focusSubItemInNonFocusScope()
@@ -753,7 +754,7 @@ void tst_qquickitem::parentItemWithFocus()
 {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
     {
     QQuickItem parent;
     QQuickItem child;
@@ -852,7 +853,7 @@ void tst_qquickitem::reparentFocusedItem()
 {
     QQuickWindow window;
     ensureFocus(&window);
-    QTRY_VERIFY(QGuiApplication::focusWindow() == &window);
+    QTRY_COMPARE(QGuiApplication::focusWindow(), &window);
 
     QQuickItem parent(window.contentItem());
     QQuickItem child(&parent);
@@ -879,18 +880,18 @@ void tst_qquickitem::reparentFocusedItem()
 void tst_qquickitem::constructor()
 {
     QScopedPointer<QQuickItem> root(new QQuickItem);
-    QVERIFY(root->parent() == 0);
-    QVERIFY(root->parentItem() == 0);
+    QVERIFY(!root->parent());
+    QVERIFY(!root->parentItem());
 
     QQuickItem *child1 = new QQuickItem(root.data());
-    QVERIFY(child1->parent() == root.data());
-    QVERIFY(child1->parentItem() == root.data());
+    QCOMPARE(child1->parent(), root.data());
+    QCOMPARE(child1->parentItem(), root.data());
     QCOMPARE(root->childItems().count(), 1);
     QCOMPARE(root->childItems().at(0), child1);
 
     QQuickItem *child2 = new QQuickItem(root.data());
-    QVERIFY(child2->parent() == root.data());
-    QVERIFY(child2->parentItem() == root.data());
+    QCOMPARE(child2->parent(), root.data());
+    QCOMPARE(child2->parentItem(), root.data());
     QCOMPARE(root->childItems().count(), 2);
     QCOMPARE(root->childItems().at(0), child1);
     QCOMPARE(root->childItems().at(1), child2);
@@ -899,41 +900,41 @@ void tst_qquickitem::constructor()
 void tst_qquickitem::setParentItem()
 {
     QQuickItem *root = new QQuickItem;
-    QVERIFY(root->parent() == 0);
-    QVERIFY(root->parentItem() == 0);
+    QVERIFY(!root->parent());
+    QVERIFY(!root->parentItem());
 
     QQuickItem *child1 = new QQuickItem;
-    QVERIFY(child1->parent() == 0);
-    QVERIFY(child1->parentItem() == 0);
+    QVERIFY(!child1->parent());
+    QVERIFY(!child1->parentItem());
 
     child1->setParentItem(root);
-    QVERIFY(child1->parent() == 0);
-    QVERIFY(child1->parentItem() == root);
+    QVERIFY(!child1->parent());
+    QCOMPARE(child1->parentItem(), root);
     QCOMPARE(root->childItems().count(), 1);
     QCOMPARE(root->childItems().at(0), child1);
 
     QQuickItem *child2 = new QQuickItem;
-    QVERIFY(child2->parent() == 0);
-    QVERIFY(child2->parentItem() == 0);
+    QVERIFY(!child2->parent());
+    QVERIFY(!child2->parentItem());
     child2->setParentItem(root);
-    QVERIFY(child2->parent() == 0);
-    QVERIFY(child2->parentItem() == root);
+    QVERIFY(!child2->parent());
+    QCOMPARE(child2->parentItem(), root);
     QCOMPARE(root->childItems().count(), 2);
     QCOMPARE(root->childItems().at(0), child1);
     QCOMPARE(root->childItems().at(1), child2);
 
     child1->setParentItem(0);
-    QVERIFY(child1->parent() == 0);
-    QVERIFY(child1->parentItem() == 0);
+    QVERIFY(!child1->parent());
+    QVERIFY(!child1->parentItem());
     QCOMPARE(root->childItems().count(), 1);
     QCOMPARE(root->childItems().at(0), child2);
 
     delete root;
 
-    QVERIFY(child1->parent() == 0);
-    QVERIFY(child1->parentItem() == 0);
-    QVERIFY(child2->parent() == 0);
-    QVERIFY(child2->parentItem() == 0);
+    QVERIFY(!child1->parent());
+    QVERIFY(!child1->parentItem());
+    QVERIFY(!child2->parent());
+    QVERIFY(!child2->parentItem());
 
     delete child1;
     delete child2;
@@ -1151,6 +1152,30 @@ void tst_qquickitem::enabledFocus()
     QCOMPARE(child2.hasFocus(), false);
     QCOMPARE(child2.hasActiveFocus(), false);
     QCOMPARE(window.activeFocusItem(), static_cast<QQuickItem *>(&child1));
+
+    child2.setFocus(true);
+    QCOMPARE(root.isEnabled(), true);
+    QCOMPARE(root.hasFocus(), true);
+    QCOMPARE(root.hasActiveFocus(), true);
+    QCOMPARE(child1.isEnabled(), true);
+    QCOMPARE(child1.hasFocus(), false);
+    QCOMPARE(child1.hasActiveFocus(), false);
+    QCOMPARE(child2.isEnabled(), false);
+    QCOMPARE(child2.hasFocus(), true);
+    QCOMPARE(child2.hasActiveFocus(), false);
+    QCOMPARE(window.activeFocusItem(), static_cast<QQuickItem *>(&root));
+
+    root.setEnabled(false);
+    QCOMPARE(root.isEnabled(), false);
+    QCOMPARE(root.hasFocus(), true);
+    QCOMPARE(root.hasActiveFocus(), false);
+    QCOMPARE(child1.isEnabled(), false);
+    QCOMPARE(child1.hasFocus(), false);
+    QCOMPARE(child1.hasActiveFocus(), false);
+    QCOMPARE(child2.isEnabled(), false);
+    QCOMPARE(child2.hasFocus(), true);
+    QCOMPARE(child2.hasActiveFocus(), false);
+    QCOMPARE(window.activeFocusItem(), window.contentItem());
 }
 
 static inline QByteArray msgItem(const QQuickItem *item)
@@ -1162,8 +1187,9 @@ static inline QByteArray msgItem(const QQuickItem *item)
 
 void tst_qquickitem::mouseGrab()
 {
-#if defined(Q_OS_WIN) && defined(QT_OPENGL_ES_2)
-    QSKIP("Fails in the CI for ANGLE builds on Windows, QTBUG-32664");
+#ifdef Q_OS_WIN
+    if (QOpenGLContext::openGLModuleType() == QOpenGLContext::LibGLES)
+        QSKIP("Fails in the CI for ANGLE builds on Windows, QTBUG-32664");
 #endif
     QQuickWindow window;
     window.setFramePosition(QPoint(100, 100));
@@ -1285,9 +1311,13 @@ void tst_qquickitem::touchEventAcceptIgnore()
         item->touchEventReached = false;
 
         bool accepted = window.event(&event);
+        QQuickTouchUtils::flush(&window);
 
         QVERIFY(item->touchEventReached);
-        QCOMPARE(accepted && event.isAccepted(), itemSupportsTouch);
+
+        // always true because QtQuick always eats touch events so as to not
+        // allow QtGui to synthesise them for us.
+        QCOMPARE(accepted && event.isAccepted(), true);
     }
     {
         QTouchEvent::TouchPoint point;
@@ -1305,9 +1335,13 @@ void tst_qquickitem::touchEventAcceptIgnore()
         item->touchEventReached = false;
 
         bool accepted = window.event(&event);
+        QQuickTouchUtils::flush(&window);
 
         QCOMPARE(item->touchEventReached, itemSupportsTouch);
-        QCOMPARE(accepted && event.isAccepted(), itemSupportsTouch);
+
+        // always true because QtQuick always eats touch events so as to not
+        // allow QtGui to synthesise them for us.
+        QCOMPARE(accepted && event.isAccepted(), true);
     }
     {
         QTouchEvent::TouchPoint point;
@@ -1325,9 +1359,13 @@ void tst_qquickitem::touchEventAcceptIgnore()
         item->touchEventReached = false;
 
         bool accepted = window.event(&event);
+        QQuickTouchUtils::flush(&window);
 
         QCOMPARE(item->touchEventReached, itemSupportsTouch);
-        QCOMPARE(accepted && event.isAccepted(), itemSupportsTouch);
+
+        // always true because QtQuick always eats touch events so as to not
+        // allow QtGui to synthesise them for us.
+        QCOMPARE(accepted && event.isAccepted(), true);
     }
 }
 
@@ -1722,6 +1760,254 @@ void tst_qquickitem::acceptedMouseButtons()
     QCOMPARE(item.releaseCount, 3);
 }
 
+static void gc(QQmlEngine &engine)
+{
+    engine.collectGarbage();
+    QCoreApplication::sendPostedEvents(0, QEvent::DeferredDelete);
+    QCoreApplication::processEvents();
+}
+
+void tst_qquickitem::visualParentOwnership()
+{
+    QQuickView view;
+    view.setSource(testFileUrl("visualParentOwnership.qml"));
+
+    QQuickItem *root = qobject_cast<QQuickItem*>(view.rootObject());
+    QVERIFY(root);
+
+    QVariant newObject;
+    {
+        QVERIFY(QMetaObject::invokeMethod(root, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> newItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!newItem.isNull());
+
+        QVERIFY(!newItem->parent());
+        QVERIFY(!newItem->parentItem());
+
+        newItem->setParentItem(root);
+
+        gc(*view.engine());
+
+        QVERIFY(!newItem.isNull());
+        newItem->setParentItem(0);
+
+        gc(*view.engine());
+        QVERIFY(newItem.isNull());
+    }
+    {
+        QVERIFY(QMetaObject::invokeMethod(root, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> firstItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!firstItem.isNull());
+
+        firstItem->setParentItem(root);
+
+        QVERIFY(QMetaObject::invokeMethod(root, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> secondItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!firstItem.isNull());
+
+        secondItem->setParentItem(firstItem);
+
+        gc(*view.engine());
+
+        delete firstItem;
+
+        root->setProperty("keepAliveProperty", newObject);
+
+        gc(*view.engine());
+        QVERIFY(!secondItem.isNull());
+
+        root->setProperty("keepAliveProperty", QVariant());
+
+        gc(*view.engine());
+        QVERIFY(secondItem.isNull());
+    }
+}
+
+void tst_qquickitem::visualParentOwnershipWindow()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("visualParentOwnershipWindow.qml"));
+
+    QQuickWindow *window = qobject_cast<QQuickWindow*>(component.create());
+    QVERIFY(window);
+    QQuickItem *root = window->contentItem();
+
+    QVariant newObject;
+    {
+        QVERIFY(QMetaObject::invokeMethod(window, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> newItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!newItem.isNull());
+
+        QVERIFY(!newItem->parent());
+        QVERIFY(!newItem->parentItem());
+
+        newItem->setParentItem(root);
+
+        gc(engine);
+
+        QVERIFY(!newItem.isNull());
+        newItem->setParentItem(0);
+
+        gc(engine);
+        QVERIFY(newItem.isNull());
+    }
+    {
+        QVERIFY(QMetaObject::invokeMethod(window, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> firstItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!firstItem.isNull());
+
+        firstItem->setParentItem(root);
+
+        QVERIFY(QMetaObject::invokeMethod(window, "createItemWithoutParent", Q_RETURN_ARG(QVariant, newObject)));
+        QPointer<QQuickItem> secondItem = qvariant_cast<QQuickItem*>(newObject);
+        QVERIFY(!firstItem.isNull());
+
+        secondItem->setParentItem(firstItem);
+
+        gc(engine);
+
+        delete firstItem;
+
+        window->setProperty("keepAliveProperty", newObject);
+
+        gc(engine);
+        QVERIFY(!secondItem.isNull());
+
+        window->setProperty("keepAliveProperty", QVariant());
+
+        gc(engine);
+        QVERIFY(secondItem.isNull());
+    }
+}
+
+class InvalidatedItem : public QQuickItem {
+    Q_OBJECT
+signals:
+    void invalidated();
+public slots:
+    void invalidateSceneGraph() { emit invalidated(); }
+};
+
+void tst_qquickitem::testSGInvalidate()
+{
+    for (int i=0; i<2; ++i) {
+        QScopedPointer<QQuickView> view(new QQuickView());
+
+        InvalidatedItem *item = new InvalidatedItem();
+
+        int expected = 0;
+        if (i == 0) {
+            // First iteration, item has contents and should get signals
+            expected = 1;
+            item->setFlag(QQuickItem::ItemHasContents, true);
+        } else {
+            // Second iteration, item does not have content and will not get signals
+        }
+
+        QSignalSpy invalidateSpy(item, SIGNAL(invalidated()));
+        item->setParentItem(view->contentItem());
+        view->show();
+
+        QVERIFY(QTest::qWaitForWindowExposed(view.data()));
+
+        delete view.take();
+        QCOMPARE(invalidateSpy.size(), expected);
+    }
+}
+
+void tst_qquickitem::objectChildTransform()
+{
+    QQuickView view;
+    view.setSource(testFileUrl("objectChildTransform.qml"));
+
+    QQuickItem *root = qobject_cast<QQuickItem*>(view.rootObject());
+    QVERIFY(root);
+
+    root->setProperty("source", QString());
+    // Shouldn't crash.
+}
+
+void tst_qquickitem::contains_data()
+{
+    QTest::addColumn<int>("x");
+    QTest::addColumn<int>("y");
+    QTest::addColumn<bool>("contains");
+
+    QTest::newRow("(0, 0) = false") << 0 << 0 << false;
+    QTest::newRow("(50, 0) = false") << 50 << 0 << false;
+    QTest::newRow("(0, 50) = false") << 0 << 50 << false;
+    QTest::newRow("(50, 50) = true") << 50 << 50 << true;
+    QTest::newRow("(100, 100) = true") << 100 << 100 << true;
+    QTest::newRow("(150, 150) = false") << 150 << 150 << false;
+}
+
+void tst_qquickitem::contains()
+{
+    // Tests that contains works, but also checks that mapToItem/mapFromItem
+    // return the correct type (point or rect, not a JS object with those properties),
+    // as this is a common combination of calls.
+
+    QFETCH(int, x);
+    QFETCH(int, y);
+    QFETCH(bool, contains);
+
+    QQuickView view;
+    view.setSource(testFileUrl("contains.qml"));
+
+    QQuickItem *root = qobject_cast<QQuickItem*>(view.rootObject());
+    QVERIFY(root);
+
+    QVariant result = false;
+    QVERIFY(QMetaObject::invokeMethod(root, "childContainsViaMapToItem",
+        Q_RETURN_ARG(QVariant, result), Q_ARG(QVariant, qreal(x)), Q_ARG(QVariant, qreal(y))));
+    QCOMPARE(result.toBool(), contains);
+
+    result = false;
+    QVERIFY(QMetaObject::invokeMethod(root, "childContainsViaMapFromItem",
+        Q_RETURN_ARG(QVariant, result), Q_ARG(QVariant, qreal(x)), Q_ARG(QVariant, qreal(y))));
+    QCOMPARE(result.toBool(), contains);
+}
+
+void tst_qquickitem::childAt()
+{
+    QQuickView view;
+    view.setSource(testFileUrl("childAtRectangle.qml"));
+    QQuickItem *root = qobject_cast<QQuickItem*>(view.rootObject());
+
+    int found = 0;
+    for (int i = 0; i < 16; i++)
+    {
+        if (root->childAt(i, 0))
+            found++;
+    }
+    QCOMPARE(found, 16);
+
+    found = 0;
+    for (int i = 0; i < 16; i++)
+    {
+        if (root->childAt(0, i))
+            found++;
+    }
+    QCOMPARE(found, 16);
+
+    found = 0;
+    for (int i = 0; i < 2; i++)
+    {
+        if (root->childAt(18 + i, 0))
+            found++;
+    }
+    QCOMPARE(found, 1);
+
+    found = 0;
+    for (int i = 0; i < 16; i++)
+    {
+        if (root->childAt(18, i))
+            found++;
+    }
+    QCOMPARE(found, 1);
+
+    QVERIFY(!root->childAt(19,19));
+}
 
 QTEST_MAIN(tst_qquickitem)
 
