@@ -1,39 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -49,9 +36,11 @@
 #include <QtQml/qqmlcontext.h>
 #include <QtQml/qqmlengine.h>
 #include "../../shared/util.h"
+#include "../shared/viewtestutil.h"
 #include <QtGui/qstylehints.h>
 #include <QtGui/QCursor>
 #include <QtGui/QScreen>
+#include <qpa/qwindowsysteminterface.h>
 
 // Initialize view, set Url, center in available geometry, move mouse away if desired
 static bool initView(QQuickView &v, const QUrl &url, bool moveMouseOut, QByteArray *errorMessage)
@@ -81,19 +70,28 @@ static bool initView(QQuickView &v, const QUrl &url, bool moveMouseOut, QByteArr
 class tst_QQuickMouseArea: public QQmlDataTest
 {
     Q_OBJECT
+public:
+    tst_QQuickMouseArea()
+        : device(nullptr)
+    {}
+
 private slots:
+    void initTestCase() Q_DECL_OVERRIDE;
     void dragProperties();
     void resetDrag();
     void dragging_data() { acceptedButton_data(); }
     void dragging();
+    void dragSmoothed();
     void dragThreshold();
     void invalidDrag_data() { rejectedButton_data(); }
     void invalidDrag();
+    void cancelDragging();
     void setDragOnPressed();
     void updateMouseAreaPosOnClick();
     void updateMouseAreaPosOnResize();
     void noOnClickedWithPressAndHold();
     void onMousePressRejected();
+    void pressedCanceledOnWindowDeactivate_data();
     void pressedCanceledOnWindowDeactivate();
     void doubleClick_data() { acceptedButton_data(); }
     void doubleClick();
@@ -121,15 +119,29 @@ private slots:
     void moveAndReleaseWithoutPress();
     void nestedStopAtBounds();
     void nestedStopAtBounds_data();
+    void containsPress_data();
+    void containsPress();
+    void ignoreBySource();
 
 private:
     void acceptedButton_data();
     void rejectedButton_data();
+    QTouchDevice *device;
 };
 
 Q_DECLARE_METATYPE(Qt::MouseButton)
 Q_DECLARE_METATYPE(Qt::MouseButtons)
 
+void tst_QQuickMouseArea::initTestCase()
+{
+    QQmlDataTest::initTestCase();
+    if (!device) {
+        device = new QTouchDevice;
+        device->setType(QTouchDevice::TouchScreen);
+        QWindowSystemInterface::registerTouchDevice(device);
+    }
+}
+
 void tst_QQuickMouseArea::acceptedButton_data()
 {
     QTest::addColumn<Qt::MouseButtons>("acceptedButtons");
@@ -172,7 +184,7 @@ void tst_QQuickMouseArea::dragProperties()
     // target
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
     QQuickItem *rootItem = qobject_cast<QQuickItem*>(window.rootObject());
     QVERIFY(rootItem != 0);
     QSignalSpy targetSpy(drag, SIGNAL(targetChanged()));
@@ -270,14 +282,14 @@ void tst_QQuickMouseArea::resetDrag()
     // target
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
     QQuickItem *rootItem = qobject_cast<QQuickItem*>(window.rootObject());
     QVERIFY(rootItem != 0);
     QSignalSpy targetSpy(drag, SIGNAL(targetChanged()));
     QVERIFY(drag->target() != 0);
     window.rootContext()->setContextProperty("haveTarget", QVariant(false));
     QCOMPARE(targetSpy.count(),1);
-    QVERIFY(drag->target() == 0);
+    QVERIFY(!drag->target());
 }
 
 void tst_QQuickMouseArea::dragging()
@@ -303,7 +315,7 @@ void tst_QQuickMouseArea::dragging()
     // target
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
 
     QVERIFY(!drag->active());
 
@@ -333,6 +345,50 @@ void tst_QQuickMouseArea::dragging()
     QCOMPARE(blackRect->y(), 61.0);
 }
 
+void tst_QQuickMouseArea::dragSmoothed()
+{
+    QQuickView window;
+    QByteArray errorMessage;
+    QVERIFY2(initView(window, testFileUrl("dragging.qml"), true, &errorMessage), errorMessage.constData());
+
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QVERIFY(window.rootObject() != 0);
+
+    QQuickMouseArea *mouseRegion = window.rootObject()->findChild<QQuickMouseArea*>("mouseregion");
+    QQuickDrag *drag = mouseRegion->drag();
+    drag->setThreshold(5);
+
+    mouseRegion->setAcceptedButtons(Qt::LeftButton);
+    QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
+    QVERIFY(blackRect != 0);
+    QCOMPARE(blackRect, drag->target());
+    QVERIFY(!drag->active());
+    QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(100,100));
+    QVERIFY(!drag->active());
+    QTest::mouseMove(&window, QPoint(100, 102), 50);
+    QTest::mouseMove(&window, QPoint(100, 106), 50);
+    QTest::mouseMove(&window, QPoint(100, 122), 50);
+    QTRY_COMPARE(blackRect->x(), 50.0);
+    QTRY_COMPARE(blackRect->y(), 66.0);
+    QTest::mouseRelease(&window, Qt::LeftButton, 0, QPoint(100,122));
+
+    // reset rect position
+    blackRect->setX(50.0);
+    blackRect->setY(50.0);
+
+    // now try with smoothed disabled
+    drag->setSmoothed(false);
+    QVERIFY(!drag->active());
+    QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(100,100));
+    QVERIFY(!drag->active());
+    QTest::mouseMove(&window, QPoint(100, 102), 50);
+    QTest::mouseMove(&window, QPoint(100, 106), 50);
+    QTest::mouseMove(&window, QPoint(100, 122), 50);
+    QTRY_COMPARE(blackRect->x(), 50.0);
+    QTRY_COMPARE(blackRect->y(), 72.0);
+    QTest::mouseRelease(&window, Qt::LeftButton, 0, QPoint(100, 122));
+}
 
 void tst_QQuickMouseArea::dragThreshold()
 {
@@ -352,7 +408,7 @@ void tst_QQuickMouseArea::dragThreshold()
     mouseRegion->setAcceptedButtons(Qt::LeftButton);
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
     QVERIFY(!drag->active());
     QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(100,100));
     QVERIFY(!drag->active());
@@ -410,7 +466,7 @@ void tst_QQuickMouseArea::invalidDrag()
     // target
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
 
     QVERIFY(!drag->active());
 
@@ -440,6 +496,57 @@ void tst_QQuickMouseArea::invalidDrag()
     QCOMPARE(blackRect->y(), 50.0);
 }
 
+void tst_QQuickMouseArea::cancelDragging()
+{
+    QQuickView window;
+    QByteArray errorMessage;
+    QVERIFY2(initView(window, testFileUrl("dragging.qml"), true, &errorMessage), errorMessage.constData());
+
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QVERIFY(window.rootObject() != 0);
+
+    QQuickMouseArea *mouseRegion = window.rootObject()->findChild<QQuickMouseArea*>("mouseregion");
+    QQuickDrag *drag = mouseRegion->drag();
+    QVERIFY(mouseRegion != 0);
+    QVERIFY(drag != 0);
+
+    mouseRegion->setAcceptedButtons(Qt::LeftButton);
+
+    // target
+    QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
+    QVERIFY(blackRect != 0);
+    QCOMPARE(blackRect, drag->target());
+
+    QVERIFY(!drag->active());
+
+    QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(100,100));
+
+    QVERIFY(!drag->active());
+    QCOMPARE(blackRect->x(), 50.0);
+    QCOMPARE(blackRect->y(), 50.0);
+
+    QTest::mouseMove(&window, QPoint(111,111), 50);
+    QTest::mouseMove(&window, QPoint(116,116), 50);
+    QTest::mouseMove(&window, QPoint(122,122), 50);
+
+    QTRY_VERIFY(drag->active());
+    QTRY_COMPARE(blackRect->x(), 61.0);
+    QCOMPARE(blackRect->y(), 61.0);
+
+    mouseRegion->QQuickItem::ungrabMouse();
+    QTRY_VERIFY(!drag->active());
+    QCOMPARE(blackRect->x(), 61.0);
+    QCOMPARE(blackRect->y(), 61.0);
+
+    QTest::mouseMove(&window, QPoint(132,132), 50);
+    QTRY_VERIFY(!drag->active());
+    QCOMPARE(blackRect->x(), 61.0);
+    QCOMPARE(blackRect->y(), 61.0);
+
+    QTest::mouseRelease(&window, Qt::LeftButton, 0, QPoint(122,122));
+}
+
 void tst_QQuickMouseArea::setDragOnPressed()
 {
     QQuickView window;
@@ -563,7 +670,7 @@ void tst_QQuickMouseArea::noOnClickedWithPressAndHold()
         QMouseEvent pressEvent(QEvent::MouseButtonPress, QPoint(100, 100), Qt::LeftButton, Qt::LeftButton, 0);
         QGuiApplication::sendEvent(&window, &pressEvent);
 
-        QVERIFY(mouseArea->pressedButtons() == Qt::LeftButton);
+        QCOMPARE(mouseArea->pressedButtons(), Qt::LeftButton);
         QVERIFY(!window.rootObject()->property("clicked").toBool());
         QVERIFY(!window.rootObject()->property("held").toBool());
 
@@ -640,8 +747,19 @@ void tst_QQuickMouseArea::onMousePressRejected()
     QVERIFY(!window.rootObject()->property("mr1_canceled").toBool());
     QVERIFY(!window.rootObject()->property("mr2_released").toBool());
 }
+
+void tst_QQuickMouseArea::pressedCanceledOnWindowDeactivate_data()
+{
+    QTest::addColumn<bool>("doubleClick");
+    QTest::newRow("simple click") << false;
+    QTest::newRow("double click") << true;
+}
+
+
 void tst_QQuickMouseArea::pressedCanceledOnWindowDeactivate()
 {
+    QFETCH(bool, doubleClick);
+
     QQuickView window;
     QByteArray errorMessage;
     QVERIFY2(initView(window, testFileUrl("pressedCanceled.qml"), true, &errorMessage), errorMessage.constData());
@@ -650,14 +768,41 @@ void tst_QQuickMouseArea::pressedCanceledOnWindowDeactivate()
     QVERIFY(window.rootObject() != 0);
     QVERIFY(!window.rootObject()->property("pressed").toBool());
     QVERIFY(!window.rootObject()->property("canceled").toBool());
-    QVERIFY(!window.rootObject()->property("released").toBool());
+
+    int expectedRelease = 0;
+    int expectedClicks = 0;
+    QCOMPARE(window.rootObject()->property("released").toInt(), expectedRelease);
+    QCOMPARE(window.rootObject()->property("clicked").toInt(), expectedClicks);
+
 
     QMouseEvent pressEvent(QEvent::MouseButtonPress, QPoint(100, 100), Qt::LeftButton, Qt::LeftButton, 0);
+    QMouseEvent releaseEvent(QEvent::MouseButtonRelease, QPoint(100, 100), Qt::LeftButton, Qt::LeftButton, 0);
+
     QGuiApplication::sendEvent(&window, &pressEvent);
 
     QVERIFY(window.rootObject()->property("pressed").toBool());
     QVERIFY(!window.rootObject()->property("canceled").toBool());
-    QVERIFY(!window.rootObject()->property("released").toBool());
+    QCOMPARE(window.rootObject()->property("released").toInt(), expectedRelease);
+    QCOMPARE(window.rootObject()->property("clicked").toInt(), expectedClicks);
+
+    if (doubleClick) {
+        QGuiApplication::sendEvent(&window, &releaseEvent);
+        QVERIFY(!window.rootObject()->property("pressed").toBool());
+        QVERIFY(!window.rootObject()->property("canceled").toBool());
+        QCOMPARE(window.rootObject()->property("released").toInt(), ++expectedRelease);
+        QCOMPARE(window.rootObject()->property("clicked").toInt(), ++expectedClicks);
+
+        QGuiApplication::sendEvent(&window, &pressEvent);
+        QMouseEvent pressEvent2(QEvent::MouseButtonDblClick, QPoint(100, 100), Qt::LeftButton, Qt::LeftButton, 0);
+        QGuiApplication::sendEvent(&window, &pressEvent2);
+
+        QVERIFY(window.rootObject()->property("pressed").toBool());
+        QVERIFY(!window.rootObject()->property("canceled").toBool());
+        QCOMPARE(window.rootObject()->property("released").toInt(), expectedRelease);
+        QCOMPARE(window.rootObject()->property("clicked").toInt(), expectedClicks);
+        QCOMPARE(window.rootObject()->property("doubleClicked").toInt(), 1);
+    }
+
 
     QWindow *secondWindow = qvariant_cast<QWindow*>(window.rootObject()->property("secondWindow"));
     secondWindow->setProperty("visible", true);
@@ -665,22 +810,24 @@ void tst_QQuickMouseArea::pressedCanceledOnWindowDeactivate()
 
     QVERIFY(!window.rootObject()->property("pressed").toBool());
     QVERIFY(window.rootObject()->property("canceled").toBool());
-    QVERIFY(!window.rootObject()->property("released").toBool());
+    QCOMPARE(window.rootObject()->property("released").toInt(), expectedRelease);
+    QCOMPARE(window.rootObject()->property("clicked").toInt(), expectedClicks);
 
     //press again
     QGuiApplication::sendEvent(&window, &pressEvent);
     QVERIFY(window.rootObject()->property("pressed").toBool());
     QVERIFY(!window.rootObject()->property("canceled").toBool());
-    QVERIFY(!window.rootObject()->property("released").toBool());
+    QCOMPARE(window.rootObject()->property("released").toInt(), expectedRelease);
+    QCOMPARE(window.rootObject()->property("clicked").toInt(), expectedClicks);
 
     QTest::qWait(200);
 
     //release
-    QMouseEvent releaseEvent(QEvent::MouseButtonRelease, QPoint(100, 100), Qt::LeftButton, Qt::LeftButton, 0);
     QGuiApplication::sendEvent(&window, &releaseEvent);
     QVERIFY(!window.rootObject()->property("pressed").toBool());
     QVERIFY(!window.rootObject()->property("canceled").toBool());
-    QVERIFY(window.rootObject()->property("released").toBool());
+    QCOMPARE(window.rootObject()->property("released").toInt(), ++expectedRelease);
+    QCOMPARE(window.rootObject()->property("clicked").toInt(), ++expectedClicks);
 }
 
 void tst_QQuickMouseArea::doubleClick()
@@ -881,7 +1028,6 @@ void tst_QQuickMouseArea::preventStealing()
 
 void tst_QQuickMouseArea::clickThrough()
 {
-    QSKIP("QTBUG-23976 Unstable");
     //With no handlers defined click, doubleClick and PressAndHold should propagate to those with handlers
     QScopedPointer<QQuickView> window(new QQuickView);
     QByteArray errorMessage;
@@ -1018,6 +1164,19 @@ void tst_QQuickMouseArea::clickThrough()
 
     QCOMPARE(window->rootObject()->property("clicksEnabled").toInt(), 2);
     QCOMPARE(window->rootObject()->property("clicksDisabled").toInt(), 1); //disabled, shouldn't increment
+
+    window.reset(new QQuickView);
+
+    //QTBUG-49100
+    QVERIFY2(initView(*window.data(), testFileUrl("qtbug49100.qml"), true, &errorMessage), errorMessage.constData());
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window.data()));
+    QVERIFY(window->rootObject() != 0);
+
+    QTest::mousePress(window.data(), Qt::LeftButton, 0, QPoint(100,100));
+    QTest::mouseRelease(window.data(), Qt::LeftButton, 0, QPoint(100,100));
+
+    QVERIFY(window->rootObject() != 0);
 }
 
 void tst_QQuickMouseArea::hoverPosition()
@@ -1137,7 +1296,7 @@ void tst_QQuickMouseArea::disableAfterPress()
     // target
     QQuickItem *blackRect = window.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
 
     QVERIFY(!drag->active());
 
@@ -1430,7 +1589,7 @@ void tst_QQuickMouseArea::changeAxis()
     // target
     QQuickItem *blackRect = view.rootObject()->findChild<QQuickItem*>("blackrect");
     QVERIFY(blackRect != 0);
-    QVERIFY(blackRect == drag->target());
+    QCOMPARE(blackRect, drag->target());
 
     QVERIFY(!drag->active());
 
@@ -1588,6 +1747,163 @@ void tst_QQuickMouseArea::nestedStopAtBounds()
     QTest::mouseRelease(&view, Qt::LeftButton, 0, position);
 }
 
+void tst_QQuickMouseArea::containsPress_data()
+{
+    QTest::addColumn<bool>("hoverEnabled");
+
+    QTest::newRow("hover enabled") << true;
+    QTest::newRow("hover disaabled") << false;
+}
+
+void tst_QQuickMouseArea::containsPress()
+{
+    QFETCH(bool, hoverEnabled);
+
+    QQuickView window;
+    QByteArray errorMessage;
+    QVERIFY2(initView(window, testFileUrl("containsPress.qml"), true, &errorMessage), errorMessage.constData());
+    window.show();
+    window.requestActivate();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QQuickItem *root = window.rootObject();
+    QVERIFY(root != 0);
+
+    QQuickMouseArea *mouseArea = window.rootObject()->findChild<QQuickMouseArea*>("mouseArea");
+    QVERIFY(mouseArea != 0);
+
+    QSignalSpy containsPressSpy(mouseArea, SIGNAL(containsPressChanged()));
+
+    mouseArea->setHoverEnabled(hoverEnabled);
+
+    QTest::mouseMove(&window, QPoint(22,33));
+    QCOMPARE(mouseArea->hovered(), false);
+    QCOMPARE(mouseArea->pressed(), false);
+    QCOMPARE(mouseArea->containsPress(), false);
+
+    QTest::mouseMove(&window, QPoint(200,200));
+    QCOMPARE(mouseArea->hovered(), hoverEnabled);
+    QCOMPARE(mouseArea->pressed(), false);
+    QCOMPARE(mouseArea->containsPress(), false);
+
+    QTest::mousePress(&window, Qt::LeftButton, Qt::NoModifier, QPoint(200,200));
+    QCOMPARE(mouseArea->hovered(), true);
+    QTRY_COMPARE(mouseArea->pressed(), true);
+    QCOMPARE(mouseArea->containsPress(), true);
+    QCOMPARE(containsPressSpy.count(), 1);
+
+    QTest::mouseMove(&window, QPoint(22,33));
+    QCOMPARE(mouseArea->hovered(), false);
+    QCOMPARE(mouseArea->pressed(), true);
+    QCOMPARE(mouseArea->containsPress(), false);
+    QCOMPARE(containsPressSpy.count(), 2);
+
+    QTest::mouseMove(&window, QPoint(200,200));
+    QCOMPARE(mouseArea->hovered(), true);
+    QCOMPARE(mouseArea->pressed(), true);
+    QCOMPARE(mouseArea->containsPress(), true);
+    QCOMPARE(containsPressSpy.count(), 3);
+
+    QTest::mouseRelease(&window, Qt::LeftButton, Qt::NoModifier, QPoint(200,200));
+    QCOMPARE(mouseArea->hovered(), hoverEnabled);
+    QCOMPARE(mouseArea->pressed(), false);
+    QCOMPARE(mouseArea->containsPress(), false);
+    QCOMPARE(containsPressSpy.count(), 4);
+}
+
+void tst_QQuickMouseArea::ignoreBySource()
+{
+    QQuickView window;
+    QByteArray errorMessage;
+    QVERIFY2(initView(window, testFileUrl("ignoreBySource.qml"), true, &errorMessage), errorMessage.constData());
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+    QVERIFY(window.rootObject());
+
+    QQuickItem *root = qobject_cast<QQuickItem*>(window.rootObject());
+    QVERIFY(root);
+
+    QQuickMouseArea *mouseArea = root->findChild<QQuickMouseArea*>("mousearea");
+    QVERIFY(mouseArea);
+
+    QQuickFlickable *flickable = root->findChild<QQuickFlickable*>("flickable");
+    QVERIFY(flickable);
+
+    // MouseArea should grab the press because it's interested in non-synthesized mouse events
+    QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(80, 80));
+    QVERIFY(window.mouseGrabberItem() == mouseArea);
+    // That was a real mouse event
+    QVERIFY(root->property("lastEventSource").toInt() == Qt::MouseEventNotSynthesized);
+
+    // Flickable content should not move
+    QTest::mouseMove(&window,QPoint(69,69));
+    QTest::mouseMove(&window,QPoint(58,58));
+    QTest::mouseMove(&window,QPoint(47,47));
+    QCOMPARE(flickable->contentX(), 0.);
+    QCOMPARE(flickable->contentY(), 0.);
+
+    QTest::mouseRelease(&window, Qt::LeftButton, 0, QPoint(47, 47));
+
+    // Now try touch events and confirm that MouseArea ignores them, while Flickable does its thing
+
+    QTest::touchEvent(&window, device).press(0, QPoint(80, 80), &window);
+    QQuickTouchUtils::flush(&window);
+    QVERIFY(window.mouseGrabberItem() != mouseArea);
+    // That was a fake mouse event
+    QCOMPARE(root->property("lastEventSource").toInt(), int(Qt::MouseEventSynthesizedByQt));
+    QTest::touchEvent(&window, device).move(0, QPoint(69,69), &window);
+    QTest::touchEvent(&window, device).move(0, QPoint(69,69), &window);
+    QTest::touchEvent(&window, device).move(0, QPoint(47,47), &window);
+    QQuickTouchUtils::flush(&window);
+    QCOMPARE(window.mouseGrabberItem(), flickable);
+    QTest::touchEvent(&window, device).release(0, QPoint(47,47), &window);
+    QQuickTouchUtils::flush(&window);
+
+    // Flickable content should have moved
+    QTRY_VERIFY(flickable->contentX() > 1);
+    QVERIFY(flickable->contentY() > 1);
+
+
+    // Now tell the MouseArea to accept only synthesized events, and repeat the tests
+    root->setProperty("allowedSource", Qt::MouseEventSynthesizedByQt);
+    flickable->setContentX(0);
+    flickable->setContentY(0);
+
+
+    // MouseArea should ignore the press because it's interested in synthesized mouse events
+    QTest::mousePress(&window, Qt::LeftButton, 0, QPoint(80, 80));
+    QVERIFY(window.mouseGrabberItem() != mouseArea);
+    // That was a real mouse event
+    QVERIFY(root->property("lastEventSource").toInt() == Qt::MouseEventNotSynthesized);
+
+    // Flickable content should move
+    QTest::mouseMove(&window,QPoint(69,69));
+    QTest::mouseMove(&window,QPoint(58,58));
+    QTest::mouseMove(&window,QPoint(47,47));
+    QTRY_VERIFY(flickable->contentX() > 1);
+    QVERIFY(flickable->contentY() > 1);
+
+    QTest::mouseRelease(&window, Qt::LeftButton, 0, QPoint(47, 47));
+    flickable->setContentX(0);
+    flickable->setContentY(0);
+
+    // Now try touch events and confirm that MouseArea gets them, while Flickable doesn't
+
+    QTest::touchEvent(&window, device).press(0, QPoint(80, 80), &window);
+    QQuickTouchUtils::flush(&window);
+    QCOMPARE(window.mouseGrabberItem(), mouseArea);
+    QTest::touchEvent(&window, device).move(0, QPoint(69,69), &window);
+    QTest::touchEvent(&window, device).move(0, QPoint(69,69), &window);
+    QTest::touchEvent(&window, device).move(0, QPoint(47,47), &window);
+    QQuickTouchUtils::flush(&window);
+    QCOMPARE(window.mouseGrabberItem(), mouseArea);
+    QTest::touchEvent(&window, device).release(0, QPoint(47,47), &window);
+    QQuickTouchUtils::flush(&window);
+
+    // Flickable content should not have moved
+    QCOMPARE(flickable->contentX(), 0.);
+    QCOMPARE(flickable->contentY(), 0.);
+}
+
 QTEST_MAIN(tst_QQuickMouseArea)
 
 #include "tst_qquickmousearea.moc"
