@@ -4,7 +4,7 @@
 ** Copyright (C) 2011 Sergey Belyashov <Sergey.Belyashov@gmail.com>
 ** Copyright (C) 2012 Laszlo Papp <lpapp@kde.org>
 ** Copyright (C) 2012 Andre Hartmann <aha_1980@gmx.de>
-** Contact: http://www.qt-project.org/legal
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtSerialPort module of the Qt Toolkit.
 **
@@ -13,30 +13,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -44,31 +42,53 @@
 
 #include "qserialport.h"
 #include "qserialportinfo.h"
+#include "qserialportinfo_p.h"
 
-#ifdef Q_OS_WINCE
-#include "qserialport_wince_p.h"
-#elif defined (Q_OS_WIN)
-#include "qserialport_win_p.h"
-#elif defined (Q_OS_SYMBIAN)
-#include "qserialport_symbian_p.h"
-#elif defined (Q_OS_UNIX)
-#include "qserialport_unix_p.h"
-#else
-#error Unsupported OS
-#endif
-
-#ifndef SERIALPORT_BUFFERSIZE
-#  define SERIALPORT_BUFFERSIZE 16384
-#endif
+#include "qserialport_p.h"
 
 #include <QtCore/qdebug.h>
 
 QT_BEGIN_NAMESPACE
 
-QSerialPortPrivateData::QSerialPortPrivateData(QSerialPort *q)
+QSerialPortErrorInfo::QSerialPortErrorInfo(QSerialPort::SerialPortError newErrorCode,
+                                           const QString &newErrorString)
+    : errorCode(newErrorCode)
+    , errorString(newErrorString)
+{
+    if (errorString.isNull()) {
+        switch (errorCode) {
+        case QSerialPort::NoError:
+            errorString = QSerialPort::tr("No error");
+            break;
+        case QSerialPort::OpenError:
+            errorString = QSerialPort::tr("Device is already open");
+            break;
+        case QSerialPort::NotOpenError:
+            errorString = QSerialPort::tr("Device is not open");
+            break;
+        case QSerialPort::TimeoutError:
+            errorString = QSerialPort::tr("Operation timed out");
+            break;
+        case QSerialPort::ReadError:
+            errorString = QSerialPort::tr("Error reading from device");
+            break;
+        case QSerialPort::WriteError:
+            errorString = QSerialPort::tr("Error writing to device");
+            break;
+        case QSerialPort::ResourceError:
+            errorString = QSerialPort::tr("Device disappeared from the system");
+            break;
+        default:
+            // an empty string will be interpreted as "Unknown error"
+            // from the QIODevice::errorString()
+            break;
+        }
+    }
+}
+
+QSerialPortPrivate::QSerialPortPrivate()
     : readBufferMaxSize(0)
-    , readBuffer(SERIALPORT_BUFFERSIZE)
-    , writeBuffer(SERIALPORT_BUFFERSIZE)
+    , writeBuffer(InitialBufferSize)
     , error(QSerialPort::NoError)
     , inputBaudRate(9600)
     , outputBaudRate(9600)
@@ -76,20 +96,43 @@ QSerialPortPrivateData::QSerialPortPrivateData(QSerialPort *q)
     , parity(QSerialPort::NoParity)
     , stopBits(QSerialPort::OneStop)
     , flowControl(QSerialPort::NoFlowControl)
-    , policy(QSerialPort::IgnorePolicy)
-    , dataTerminalReady(false)
-    , requestToSend(false)
+#if QT_DEPRECATED_SINCE(5,3)
     , settingsRestoredOnClose(true)
-    , q_ptr(q)
+#endif
+    , isBreakEnabled(false)
+#if defined(Q_OS_WIN32)
+    , handle(INVALID_HANDLE_VALUE)
+    , readChunkBuffer(ReadChunkSize, 0)
+    , communicationStarted(false)
+    , writeStarted(false)
+    , readStarted(false)
+    , notifier(0)
+    , startAsyncWriteTimer(0)
+    , originalEventMask(0)
+    , triggeredEventMask(0)
+    , actualBytesToWrite(0)
+#elif defined(Q_OS_UNIX)
+    , descriptor(-1)
+    , readNotifier(0)
+    , writeNotifier(0)
+    , readPortNotifierCalled(false)
+    , readPortNotifierState(false)
+    , readPortNotifierStateSet(false)
+    , emittedReadyRead(false)
+    , emittedBytesWritten(false)
+    , pendingBytesWritten(0)
+    , writeSequenceStarted(false)
+#endif
 {
 }
 
-int QSerialPortPrivateData::timeoutValue(int msecs, int elapsed)
+void QSerialPortPrivate::setError(const QSerialPortErrorInfo &errorInfo)
 {
-    if (msecs == -1)
-        return msecs;
-    msecs -= elapsed;
-    return qMax(msecs, 0);
+    Q_Q(QSerialPort);
+
+    error = errorInfo.errorCode;
+    q->setErrorString(errorInfo.errorString);
+    emit q->error(error);
 }
 
 /*!
@@ -141,7 +184,7 @@ int QSerialPortPrivateData::timeoutValue(int msecs, int elapsed)
      int numRead = 0, numReadTotal = 0;
      char buffer[50];
 
-     forever {
+     for (;;) {
          numRead  = serial.read(buffer, 50);
 
          // Do whatever with the array
@@ -162,7 +205,7 @@ int QSerialPortPrivateData::timeoutValue(int msecs, int elapsed)
     used in non-GUI threads, to avoid freezing the user interface.
 
     For more details about these approaches, refer to the
-    \l {Examples}{example} applications.
+    \l {Qt Serial Port Examples}{example} applications.
 
     The QSerialPort class can also be used with QTextStream and QDataStream's
     stream operators (operator<<() and operator>>()). There is one issue to be
@@ -357,13 +400,16 @@ int QSerialPortPrivateData::timeoutValue(int msecs, int elapsed)
                                 QtSerialPort 5.2.
 
     \value ParityError          Parity error detected by the hardware while
-                                reading data.
+                                reading data. This value is obsolete. We strongly
+                                advise against using it in new code.
 
     \value FramingError         Framing error detected by the hardware while
-                                reading data.
+                                reading data. This value is obsolete. We strongly
+                                advise against using it in new code.
 
     \value BreakConditionError  Break condition detected by the hardware on
-                                the input line.
+                                the input line. This value is obsolete. We strongly
+                                advise against using it in new code.
 
     \value WriteError           An I/O error occurred while writing the data.
 
@@ -390,9 +436,10 @@ int QSerialPortPrivateData::timeoutValue(int msecs, int elapsed)
     Constructs a new serial port object with the given \a parent.
 */
 QSerialPort::QSerialPort(QObject *parent)
-    : QIODevice(parent)
-    , d_ptr(new QSerialPortPrivate(this))
-{}
+    : QIODevice(*new QSerialPortPrivate, parent)
+    , d_dummy(0)
+{
+}
 
 /*!
     Constructs a new serial port object with the given \a parent
@@ -401,8 +448,8 @@ QSerialPort::QSerialPort(QObject *parent)
     The name should have a specific format; see the setPort() method.
 */
 QSerialPort::QSerialPort(const QString &name, QObject *parent)
-    : QIODevice(parent)
-    , d_ptr(new QSerialPortPrivate(this))
+    : QIODevice(*new QSerialPortPrivate, parent)
+    , d_dummy(0)
 {
     setPortName(name);
 }
@@ -413,8 +460,8 @@ QSerialPort::QSerialPort(const QString &name, QObject *parent)
     \a serialPortInfo.
 */
 QSerialPort::QSerialPort(const QSerialPortInfo &serialPortInfo, QObject *parent)
-    : QIODevice(parent)
-    , d_ptr(new QSerialPortPrivate(this))
+    : QIODevice(*new QSerialPortPrivate, parent)
+    , d_dummy(0)
 {
     setPort(serialPortInfo);
 }
@@ -427,7 +474,6 @@ QSerialPort::~QSerialPort()
     /**/
     if (isOpen())
         close();
-    delete d_ptr;
 }
 
 /*!
@@ -441,7 +487,7 @@ QSerialPort::~QSerialPort()
 void QSerialPort::setPortName(const QString &name)
 {
     Q_D(QSerialPort);
-    d->systemLocation = QSerialPortPrivate::portNameToSystemLocation(name);
+    d->systemLocation = QSerialPortInfoPrivate::portNameToSystemLocation(name);
 }
 
 /*!
@@ -452,7 +498,7 @@ void QSerialPort::setPortName(const QString &name)
 void QSerialPort::setPort(const QSerialPortInfo &serialPortInfo)
 {
     Q_D(QSerialPort);
-    d->systemLocation = QSerialPortPrivate::portNameToSystemLocation(serialPortInfo.systemLocation());
+    d->systemLocation = serialPortInfo.systemLocation();
 }
 
 /*!
@@ -466,27 +512,16 @@ void QSerialPort::setPort(const QSerialPortInfo &serialPortInfo)
         \li Brief Description
     \row
         \li Windows
-        \li Removes the prefix "\\\\.\\" from the system location
+        \li Removes the prefix "\\\\.\\" or "//./" from the system location
            and returns the remainder of the string.
     \row
         \li Windows CE
         \li Removes the suffix ":" from the system location
            and returns the remainder of the string.
     \row
-        \li Symbian
-        \li Returns the system location as it is,
-           as it is equivalent to the port name.
-    \row
-        \li GNU/Linux
+        \li Unix, BSD
         \li Removes the prefix "/dev/" from the system location
            and returns the remainder of the string.
-    \row
-        \li Mac OSX
-        \li Removes the prefix "/dev/cu." and "/dev/tty." from the
-           system location and returns the remainder of the string.
-    \row
-        \li Other *nix
-        \li  The same as for GNU/Linux.
     \endtable
 
     \sa setPort(), QSerialPortInfo::portName()
@@ -494,7 +529,7 @@ void QSerialPort::setPort(const QSerialPortInfo &serialPortInfo)
 QString QSerialPort::portName() const
 {
     Q_D(const QSerialPort);
-    return QSerialPortPrivate::portNameFromSystemLocation(d->systemLocation);
+    return QSerialPortInfoPrivate::portNameFromSystemLocation(d->systemLocation);
 }
 
 /*!
@@ -518,14 +553,14 @@ bool QSerialPort::open(OpenMode mode)
     Q_D(QSerialPort);
 
     if (isOpen()) {
-        setError(QSerialPort::OpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::OpenError));
         return false;
     }
 
     // Define while not supported modes.
     static const OpenMode unsupportedModes = Append | Truncate | Text | Unbuffered;
     if ((mode & unsupportedModes) || mode == NotOpen) {
-        setError(QSerialPort::UnsupportedOperationError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, tr("Unsupported open mode")));
         return false;
     }
 
@@ -533,20 +568,16 @@ bool QSerialPort::open(OpenMode mode)
     if (!d->open(mode))
         return false;
 
-    QIODevice::open(mode);
-
     if (!d->setBaudRate()
         || !d->setDataBits(d->dataBits)
         || !d->setParity(d->parity)
         || !d->setStopBits(d->stopBits)
         || !d->setFlowControl(d->flowControl)) {
-        close();
+        d->close();
         return false;
     }
 
-    d->dataTerminalReady = isDataTerminalReady();
-    d->requestToSend = isRequestToSend();
-
+    QIODevice::open(mode);
     return true;
 }
 
@@ -562,14 +593,16 @@ void QSerialPort::close()
 {
     Q_D(QSerialPort);
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         return;
     }
 
-    QIODevice::close();
     d->close();
+    d->isBreakEnabled = false;
+    QIODevice::close();
 }
 
+#if QT_DEPRECATED_SINCE(5, 3)
 /*!
     \property QSerialPort::settingsRestoredOnClose
     \brief the flag which specifies to restore the previous settings when closing
@@ -595,9 +628,12 @@ bool QSerialPort::settingsRestoredOnClose() const
     Q_D(const QSerialPort);
     return d->settingsRestoredOnClose;
 }
+#endif // QT_DEPRECATED_SINCE(5,3)
 
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
     \fn void QSerialPort::settingsRestoredOnCloseChanged(bool restore)
+    \obsolete
 
     This signal is emitted after the flag which specifies to restore the
     previous settings while closing the serial port has been changed. The new
@@ -606,6 +642,7 @@ bool QSerialPort::settingsRestoredOnClose() const
 
     \sa QSerialPort::settingsRestoredOnClose
 */
+#endif // QT_DEPRECATED_SINCE(5, 5)
 
 /*!
     \property QSerialPort::baudRate
@@ -621,11 +658,10 @@ bool QSerialPort::settingsRestoredOnClose() const
     setting is done automatically in the \l{QSerialPort::open()} method right
     after that the opening of the port succeeds.
 
-    \warning Setting the AllDirections flag is only supported on
-    the Windows, Windows CE, and Symbian platforms.
+    \warning Setting the AllDirections flag is supported on all platforms.
+    Windows and Windows CE support only this mode.
 
-    \warning Returns equal baud rate in any direction on Windows, Windows CE, and
-    Symbian.
+    \warning Returns equal baud rate in any direction on Windows, Windows CE.
 
     The default value is Baud9600, i.e. 9600 bits per second.
 */
@@ -870,16 +906,15 @@ bool QSerialPort::setDataTerminalReady(bool set)
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
+    const bool dataTerminalReady = isDataTerminalReady();
     const bool retval = d->setDataTerminalReady(set);
-    if (retval && (d->dataTerminalReady != set)) {
-        d->dataTerminalReady = set;
+    if (retval && (dataTerminalReady != set))
         emit dataTerminalReadyChanged(set);
-    }
 
     return retval;
 }
@@ -918,17 +953,21 @@ bool QSerialPort::setRequestToSend(bool set)
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
-    const bool retval = d->setRequestToSend(set);
-    if (retval && (d->requestToSend != set)) {
-        d->requestToSend = set;
-        emit requestToSendChanged(set);
+    if (d->flowControl == QSerialPort::HardwareControl) {
+        d->setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError));
+        return false;
     }
 
+    const bool requestToSend = isRequestToSend();
+    const bool retval = d->setRequestToSend(set);
+    if (retval && (requestToSend != set))
+        emit requestToSendChanged(set);
+
     return retval;
 }
 
@@ -969,7 +1008,7 @@ QSerialPort::PinoutSignals QSerialPort::pinoutSignals()
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return QSerialPort::NoSignal;
     }
@@ -999,7 +1038,7 @@ bool QSerialPort::flush()
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
@@ -1021,13 +1060,13 @@ bool QSerialPort::clear(Directions directions)
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
     if (directions & Input)
-        d->readBuffer.clear();
+        d->buffer.clear();
     if (directions & Output)
         d->writeBuffer.clear();
     return d->clear(directions);
@@ -1057,10 +1096,10 @@ bool QSerialPort::clear(Directions directions)
  */
 bool QSerialPort::atEnd() const
 {
-    Q_D(const QSerialPort);
-    return QIODevice::atEnd() && (!isOpen() || (d->readBuffer.size() == 0));
+    return QIODevice::atEnd();
 }
 
+#if QT_DEPRECATED_SINCE(5, 2)
 /*!
     \property QSerialPort::dataErrorPolicy
     \brief the error policy for how the process receives characters in the case where
@@ -1082,26 +1121,27 @@ bool QSerialPort::setDataErrorPolicy(DataErrorPolicy policy)
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
-    const bool ret = d->policy == policy || d->setDataErrorPolicy(policy);
-    if (ret && (d->policy != policy)) {
-        d->policy = policy;
-        emit dataErrorPolicyChanged(d->policy);
+    if (policy != QSerialPort::IgnorePolicy) {
+        d->setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError,
+                    tr("The device supports only the ignoring policy")));
+        return false;
     }
 
-    return ret;
+    return true;
 }
 
 QSerialPort::DataErrorPolicy QSerialPort::dataErrorPolicy() const
 {
-    Q_D(const QSerialPort);
-    return d->policy;
+    return QSerialPort::IgnorePolicy;
 }
+#endif // QT_DEPRECATED_SINCE(5, 2)
 
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
     \fn void QSerialPort::dataErrorPolicyChanged(DataErrorPolicy policy)
     \obsolete
@@ -1113,6 +1153,7 @@ QSerialPort::DataErrorPolicy QSerialPort::dataErrorPolicy() const
 
     \sa QSerialPort::dataErrorPolicy
 */
+#endif // QT_DEPRECATED_SINCE(5, 5)
 
 /*!
     \property QSerialPort::error
@@ -1133,7 +1174,8 @@ QSerialPort::SerialPortError QSerialPort::error() const
 
 void QSerialPort::clearError()
 {
-    setError(QSerialPort::NoError);
+    Q_D(QSerialPort);
+    d->setError(QSerialPortErrorInfo(QSerialPort::NoError));
 }
 
 /*!
@@ -1205,8 +1247,7 @@ bool QSerialPort::isSequential() const
 */
 qint64 QSerialPort::bytesAvailable() const
 {
-    Q_D(const QSerialPort);
-    return d->readBuffer.size() + QIODevice::bytesAvailable();
+    return QIODevice::bytesAvailable();
 }
 
 /*!
@@ -1221,7 +1262,13 @@ qint64 QSerialPort::bytesAvailable() const
 qint64 QSerialPort::bytesToWrite() const
 {
     Q_D(const QSerialPort);
-    return d->writeBuffer.size() + QIODevice::bytesToWrite();
+    qint64 bytes = QIODevice::bytesToWrite();
+#ifdef Q_OS_WIN32
+    bytes += d->actualBytesToWrite;
+#else
+    bytes += d->writeBuffer.size();
+#endif
+    return bytes;
 }
 
 /*!
@@ -1234,9 +1281,7 @@ qint64 QSerialPort::bytesToWrite() const
 */
 bool QSerialPort::canReadLine() const
 {
-    Q_D(const QSerialPort);
-    const bool hasLine = (d->readBuffer.size() > 0) && d->readBuffer.canReadLine();
-    return hasLine || QIODevice::canReadLine();
+    return QIODevice::canReadLine();
 }
 
 /*!
@@ -1279,6 +1324,7 @@ bool QSerialPort::waitForBytesWritten(int msecs)
     return d->waitForBytesWritten(msecs);
 }
 
+#if QT_DEPRECATED_SINCE(5, 5)
 /*!
     Sends a continuous stream of zero bits during a specified period
     of time \a duration in msec if the terminal is using asynchronous
@@ -1300,45 +1346,86 @@ bool QSerialPort::sendBreak(int duration)
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
     return d->sendBreak(duration);
 }
+#endif // QT_DEPRECATED_SINCE(5, 5)
 
 /*!
-    Controls the signal break, depending on the flag \a set.
-    If successful, returns true; otherwise returns false.
+    \property QSerialPort::breakEnabled
+    \since 5.5
+    \brief the state of the transmission line in break
 
-    If \a set is true then enables the break transmission; otherwise disables.
+    Returns true on success, false otherwise.
+    If the flag is true then the transmission line is in break state;
+    otherwise is in non-break state.
 
-    \note The serial port has to be open before trying to set break enabled;
-    otherwise returns false and sets the NotOpenError error code.
-
-    \sa sendBreak()
+    \note The serial port has to be open before trying to set or get this
+    property; otherwise returns false and sets the NotOpenError error code.
+    This is a bit unusual as opposed to the regular Qt property settings of
+    a class. However, this is a special use case since the property is set
+    through the interaction with the kernel and hardware. Hence, the two
+    scenarios cannot be completely compared to each other.
 */
 bool QSerialPort::setBreakEnabled(bool set)
 {
     Q_D(QSerialPort);
 
     if (!isOpen()) {
-        setError(QSerialPort::NotOpenError);
+        d->setError(QSerialPortErrorInfo(QSerialPort::NotOpenError));
         qWarning("%s: device not open", Q_FUNC_INFO);
         return false;
     }
 
-    return d->setBreakEnabled(set);
+    if (d->setBreakEnabled(set)) {
+        if (d->isBreakEnabled != set) {
+            d->isBreakEnabled = set;
+            emit breakEnabledChanged(d->isBreakEnabled);
+        }
+        return true;
+    }
+    return false;
+}
+
+bool QSerialPort::isBreakEnabled() const
+{
+    Q_D(const QSerialPort);
+    return d->isBreakEnabled;
 }
 
 /*!
     \reimp
 */
+// This function does not really read anything, as we use QIODevicePrivate's
+// buffer. The buffer will be read inside of QIODevice before this
+// method will be called.
 qint64 QSerialPort::readData(char *data, qint64 maxSize)
 {
+    Q_UNUSED(data);
+    Q_UNUSED(maxSize);
+
+#if defined(Q_OS_WIN32)
+    // We need try to start async reading to read a remainder from a driver's queue
+    // in case we have a limited read buffer size. Because the read notification can
+    // be stalled since Windows do not re-triggered an EV_RXCHAR event if a driver's
+    // buffer has a remainder of data ready to read until a new data will be received.
     Q_D(QSerialPort);
-    return d->readBuffer.read(data, maxSize);
+    if (d->readBufferMaxSize || d->flowControl == QSerialPort::HardwareControl)
+        d->startAsyncRead();
+#elif defined(Q_OS_UNIX)
+    // We need try to re-trigger the read notification to read a remainder from a
+    // driver's queue in case we have a limited read buffer size.
+    Q_D(QSerialPort);
+    if (d->readBufferMaxSize && !d->isReadNotificationEnabled())
+        d->setReadNotificationEnabled(true);
+#endif
+
+    // return 0 indicating there may be more data in the future
+    return qint64(0);
 }
 
 /*!
@@ -1355,25 +1442,7 @@ qint64 QSerialPort::readLineData(char *data, qint64 maxSize)
 qint64 QSerialPort::writeData(const char *data, qint64 maxSize)
 {
     Q_D(QSerialPort);
-
-    ::memcpy(d->writeBuffer.reserve(maxSize), data, maxSize);
-    if (!d->writeBuffer.isEmpty())
-        d->startWriting();
-    return maxSize;
-}
-
-void QSerialPort::setError(QSerialPort::SerialPortError serialPortError, const QString &errorString)
-{
-    Q_D(QSerialPort);
-
-    d->error = serialPortError;
-
-    if (errorString.isNull())
-        setErrorString(qt_error_string(-1));
-    else
-        setErrorString(errorString);
-
-    emit error(serialPortError);
+    return d->writeData(data, maxSize);
 }
 
 #include "moc_qserialport.cpp"
