@@ -1,31 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the tools applications of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** As a special exception, The Qt Company gives you certain additional
-** rights. These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -190,6 +185,41 @@ void Generator::appendFullNames(Text& text, const NodeList& nodes, const Node* r
     }
 }
 
+/*!
+  Append the signature for the function named in \a node to
+  \a text, so that is is a link to the documentation for that
+  function.
+ */
+void Generator::appendSignature(Text& text, const Node* node)
+{
+    text << Atom(Atom::LinkNode, CodeMarker::stringForNode(node))
+         << Atom(Atom::FormattingLeft, ATOM_FORMATTING_LINK)
+         << Atom(Atom::String, node->signature(false, true))
+         << Atom(Atom::FormattingRight, ATOM_FORMATTING_LINK);
+}
+
+/*!
+  Generate a bullet list of function signatures. The function
+  nodes are in \a nodes. It uses the \a relative node and the
+  \a marker for the generation.
+ */
+void Generator::signatureList(const NodeList& nodes, const Node* relative, CodeMarker* marker)
+{
+    Text text;
+    int count = 0;
+    text << Atom(Atom::ListLeft, QString("bullet"));
+    NodeList::ConstIterator n = nodes.constBegin();
+    while (n != nodes.constEnd()) {
+        text << Atom(Atom::ListItemNumber, QString::number(++count));
+        text << Atom(Atom::ListItemLeft, QString("bullet"));
+        appendSignature(text, *n);
+        text << Atom(Atom::ListItemRight, QString("bullet"));
+        ++n;
+    }
+    text << Atom(Atom::ListRight, QString("bullet"));
+    generateText(text, relative, marker);
+}
+
 int Generator::appendSortedNames(Text& text, const ClassNode* cn, const QList<RelatedClass>& rc)
 {
     QList<RelatedClass>::ConstIterator r;
@@ -564,7 +594,7 @@ QString Generator::fullDocumentLocation(const Node *node, bool useSubdir)
     {
         const FunctionNode *fn = static_cast<const FunctionNode *>(node);
 
-        if (fn->metaness() == FunctionNode::Dtor)
+        if (fn->isDtor())
             anchorRef = "#dtor." + fn->name().mid(1);
 
         else if (fn->hasOneAssociatedProperty() && fn->doc().isEmpty())
@@ -752,8 +782,64 @@ void Generator::generateBody(const Node *node, CodeMarker *marker)
         }
     }
     if (node->doc().isEmpty()) {
-        if (!node->isWrapper() && !quiet && !node->isReimplemented()) { // ### might be unnecessary
-            node->location().warning(tr("No documentation for '%1'").arg(node->plainFullName()));
+        /*
+          Test for special function, like a destructor or copy constructor,
+          that has no documentation.
+        */
+        if (node->type() == Node::Function) {
+            const FunctionNode* func = static_cast<const FunctionNode*>(node);
+            if (func->isDtor()) {
+                Text text;
+                text << "Destroys the instance of ";
+                text << func->parent()->name() << ".";
+                if (func->isVirtual())
+                    text << " The destructor is virtual.";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (func->isCtor()) {
+                Text text;
+                text << "Default constructs an instance of ";
+                text << func->parent()->name() << ".";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (func->isCCtor()) {
+                Text text;
+                text << "Copy constructor.";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (func->isMCtor()) {
+                Text text;
+                text << "Move-copy constructor.";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (func->isCAssign()) {
+                Text text;
+                text << "Copy-assignment operator.";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (func->isMAssign()) {
+                Text text;
+                text << "Move-assignment operator.";
+                out() << "<p>";
+                generateText(text, node, marker);
+                out() << "</p>";
+            }
+            else if (!node->isWrapper() && !quiet && !node->isReimplemented()) {
+                node->location().warning(tr("No documentation for '%1'").arg(node->plainSignature()));
+            }
+        }
+        else if (!node->isWrapper() && !quiet && !node->isReimplemented()) {
+            node->location().warning(tr("No documentation for '%1'").arg(node->plainSignature()));
         }
     }
     else {
@@ -1336,24 +1422,83 @@ bool Generator::generateText(const Text& text,
     return result;
 }
 
+/*
+  The node is an aggregate, typically a class node, which has
+  a threadsafeness level. This function checks all the children
+  of the node to see if they are exceptions to the node's
+  threadsafeness. If there are any exceptions, the exceptions
+  are added to the appropriate set (reentrant, threadsafe, and
+  nonreentrant, and true is returned. If there are no exceptions,
+  the three node lists remain empty and false is returned.
+ */
+static bool hasExceptions(const Node* node,
+                          NodeList& reentrant,
+                          NodeList& threadsafe,
+                          NodeList& nonreentrant)
+{
+    bool result = false;
+    Node::ThreadSafeness ts = node->threadSafeness();
+    const Aggregate* a = static_cast<const Aggregate*>(node);
+    NodeList::ConstIterator c = a->childNodes().constBegin();
+    while (c != a->childNodes().constEnd()) {
+        if (!(*c)->isObsolete()){
+            switch ((*c)->threadSafeness()) {
+            case Node::Reentrant:
+                reentrant.append(*c);
+                if (ts == Node::ThreadSafe)
+                    result = true;
+                break;
+            case Node::ThreadSafe:
+                threadsafe.append(*c);
+                if (ts == Node::Reentrant)
+                    result = true;
+                break;
+            case Node::NonReentrant:
+                nonreentrant.append(*c);
+                result = true;
+                break;
+            default:
+                break;
+            }
+        }
+        ++c;
+    }
+    return result;
+}
+
+static void startNote(Text& text)
+{
+    text << Atom::ParaLeft
+         << Atom(Atom::FormattingLeft, ATOM_FORMATTING_BOLD)
+         << "Note:"
+         << Atom(Atom::FormattingRight, ATOM_FORMATTING_BOLD)
+         << " ";
+}
+
+/*!
+  Generates text that explains how threadsafe and/or reentrant
+  \a node is.
+ */
 void Generator::generateThreadSafeness(const Node *node, CodeMarker *marker)
 {
-    Text text;
-    Node::ThreadSafeness threadSafeness = node->threadSafeness();
+    Text text, rlink, tlink;
+    NodeList reentrant;
+    NodeList threadsafe;
+    NodeList nonreentrant;
+    Node::ThreadSafeness ts = node->threadSafeness();
+    bool exceptions = false;
 
-    Text rlink;
     rlink << Atom(Atom::Link,"reentrant")
           << Atom(Atom::FormattingLeft, ATOM_FORMATTING_LINK)
           << "reentrant"
           << Atom(Atom::FormattingRight, ATOM_FORMATTING_LINK);
 
-    Text tlink;
     tlink << Atom(Atom::Link,"thread-safe")
           << Atom(Atom::FormattingLeft, ATOM_FORMATTING_LINK)
           << "thread-safe"
           << Atom(Atom::FormattingRight, ATOM_FORMATTING_LINK);
 
-    switch (threadSafeness) {
+    switch (ts) {
     case Node::UnspecifiedSafeness:
         break;
     case Node::NonReentrant:
@@ -1370,110 +1515,76 @@ void Generator::generateThreadSafeness(const Node *node, CodeMarker *marker)
         break;
     case Node::Reentrant:
     case Node::ThreadSafe:
-        text << Atom::ParaLeft
-             << Atom(Atom::FormattingLeft,ATOM_FORMATTING_BOLD)
-             << "Note:"
-             << Atom(Atom::FormattingRight,ATOM_FORMATTING_BOLD)
-             << " ";
-
+        startNote(text);
         if (node->isAggregate()) {
-            const Aggregate* innerNode = static_cast<const Aggregate*>(node);
-            text << "All functions in this "
-                 << typeString(node)
-                 << " are ";
-            if (threadSafeness == Node::ThreadSafe)
+            exceptions = hasExceptions(node, reentrant, threadsafe, nonreentrant);
+            text << "All functions in this " << typeString(node) << " are ";
+            if (ts == Node::ThreadSafe)
                 text << tlink;
             else
                 text << rlink;
 
-            bool exceptions = false;
-            NodeList reentrant;
-            NodeList threadsafe;
-            NodeList nonreentrant;
-            NodeList::ConstIterator c = innerNode->childNodes().constBegin();
-            while (c != innerNode->childNodes().constEnd()) {
-
-                if ((*c)->status() != Node::Obsolete){
-                    switch ((*c)->threadSafeness()) {
-                    case Node::Reentrant:
-                        reentrant.append(*c);
-                        if (threadSafeness == Node::ThreadSafe)
-                            exceptions = true;
-                        break;
-                    case Node::ThreadSafe:
-                        threadsafe.append(*c);
-                        if (threadSafeness == Node::Reentrant)
-                            exceptions = true;
-                        break;
-                    case Node::NonReentrant:
-                        nonreentrant.append(*c);
-                        exceptions = true;
-                        break;
-                    default:
-                        break;
-                    }
-                }
-                ++c;
-            }
-            if (!exceptions)
+            if (!exceptions || (ts == Node::Reentrant && !threadsafe.isEmpty()))
                 text << ".";
-            else if (threadSafeness == Node::Reentrant) {
-                if (nonreentrant.isEmpty()) {
-                    if (!threadsafe.isEmpty()) {
-                        text << ", but ";
-                        appendFullNames(text,threadsafe,innerNode);
-                        singularPlural(text,threadsafe);
-                        text << " also " << tlink << ".";
-                    }
-                    else
-                        text << ".";
-                }
-                else {
-                    text << ", except for ";
-                    appendFullNames(text,nonreentrant,innerNode);
-                    text << ", which";
-                    singularPlural(text,nonreentrant);
-                    text << " nonreentrant.";
-                    if (!threadsafe.isEmpty()) {
-                        text << " ";
-                        appendFullNames(text,threadsafe,innerNode);
-                        singularPlural(text,threadsafe);
-                        text << " " << tlink << ".";
-                    }
-                }
-            }
-            else { // thread-safe
-                if (!nonreentrant.isEmpty() || !reentrant.isEmpty()) {
-                    text << ", except for ";
-                    if (!reentrant.isEmpty()) {
-                        appendFullNames(text,reentrant,innerNode);
-                        text << ", which";
-                        singularPlural(text,reentrant);
-                        text << " only " << rlink;
-                        if (!nonreentrant.isEmpty())
-                            text << ", and ";
-                    }
-                    if (!nonreentrant.isEmpty()) {
-                        appendFullNames(text,nonreentrant,innerNode);
-                        text << ", which";
-                        singularPlural(text,nonreentrant);
-                        text << " nonreentrant.";
-                    }
-                    text << ".";
-                }
-            }
+            else
+                text << " with the following exceptions:";
         }
         else {
             text << "This " << typeString(node) << " is ";
-            if (threadSafeness == Node::ThreadSafe)
+            if (ts == Node::ThreadSafe)
                 text << tlink;
             else
                 text << rlink;
-            text << ".";
         }
         text << Atom::ParaRight;
+        break;
+    default:
+        break;
     }
     generateText(text,node,marker);
+
+    if (exceptions) {
+        text.clear();
+        if (ts == Node::Reentrant) {
+            if (!nonreentrant.isEmpty()) {
+                startNote(text);
+                text << "These functions are not "
+                     << rlink
+                     << ":"
+                     << Atom::ParaRight;
+                signatureList(nonreentrant, node, marker);
+            }
+            if (!threadsafe.isEmpty()) {
+                text.clear();
+                startNote(text);
+                text << "These functions are also "
+                     << tlink
+                     << ":"
+                     << Atom::ParaRight;
+                generateText(text, node, marker);
+                signatureList(threadsafe, node, marker);
+            }
+        }
+        else { // thread-safe
+            if (!reentrant.isEmpty()) {
+                startNote(text);
+                text << "These functions are only "
+                     << rlink
+                     << ":"
+                     << Atom::ParaRight;
+                signatureList(reentrant, node, marker);
+            }
+            if (!nonreentrant.isEmpty()) {
+                text.clear();
+                startNote(text);
+                text << "These functions are not "
+                     << rlink
+                     << ":"
+                     << Atom::ParaRight;
+                signatureList(nonreentrant, node, marker);
+            }
+        }
+    }
 }
 
 /*!
@@ -1484,7 +1595,7 @@ void Generator::generateOverloadedSignal(const Node* node, CodeMarker* marker)
     if (node->type() != Node::Function)
         return;
     const FunctionNode *func = static_cast<const FunctionNode *>(node);
-    if (func->metaness() != FunctionNode::Signal)
+    if (!func->isSignal())
         return;
     if (node->parent()->overloads(node->name()).count() <= 1)
         return;
@@ -1532,7 +1643,7 @@ void Generator::generateOverloadedSignal(const Node* node, CodeMarker* marker)
          << Atom(Atom::FormattingLeft,ATOM_FORMATTING_BOLD)
          << "Note:"
          << Atom(Atom::FormattingRight,ATOM_FORMATTING_BOLD)
-         << "Signal "
+         << " Signal "
          << Atom(Atom::FormattingLeft,ATOM_FORMATTING_ITALIC)
          << node->name()
          << Atom(Atom::FormattingRight,ATOM_FORMATTING_ITALIC)
