@@ -1,31 +1,26 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the tools applications of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:GPL-EXCEPT$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** As a special exception, The Qt Company gives you certain additional
-** rights. These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -781,6 +776,30 @@ QmlTypeNode* QDocDatabase::findQmlType(const QString& qmid, const QString& name)
     return 0;
 }
 
+/*!
+  Looks up the QML basic type node identified by the Qml module id
+  \a qmid and QML basic type \a name and returns a pointer to the
+  QML basic type node. The key is \a qmid + "::" + \a name.
+
+  If the QML module id is empty, it looks up the QML basic type by
+  \a name only.
+ */
+Aggregate* QDocDatabase::findQmlBasicType(const QString& qmid, const QString& name)
+{
+    if (!qmid.isEmpty()) {
+        QString t = qmid + "::" + name;
+        Aggregate* a = forest_.lookupQmlBasicType(t);
+        if (a)
+            return a;
+    }
+
+    QStringList path(name);
+    Node* n = forest_.findNodeByNameAndType(path, Node::QmlBasicType);
+    if (n && n->isQmlBasicType())
+        return static_cast<Aggregate*>(n);
+    return 0;
+}
+
 /*!
   Looks up the QML type node identified by the Qml module id
   constructed from the strings in the \a import record and the
@@ -1025,10 +1044,8 @@ void QDocDatabase::findAllFunctions(Aggregate* node)
             }
             else if ((*c)->type() == Node::Function) {
                 const FunctionNode* func = static_cast<const FunctionNode*>(*c);
-                if ((func->status() > Node::Obsolete) &&
-                        !func->isInternal() &&
-                        (func->metaness() != FunctionNode::Ctor) &&
-                        (func->metaness() != FunctionNode::Dtor)) {
+                if ((func->status() > Node::Obsolete) && !func->isInternal() &&
+                    !func->isSomeCtor() && !func->isDtor()) {
                     funcIndex_[(*c)->name()].insert((*c)->parent()->fullDocumentName(), *c);
                 }
             }
@@ -1202,9 +1219,7 @@ void QDocDatabase::findAllSince(Aggregate* node)
             if ((*child)->type() == Node::Function) {
                 // Insert functions into the general since map.
                 FunctionNode *func = static_cast<FunctionNode *>(*child);
-                if ((func->status() > Node::Obsolete) &&
-                    (func->metaness() != FunctionNode::Ctor) &&
-                    (func->metaness() != FunctionNode::Dtor)) {
+                if ((func->status() > Node::Obsolete) && !func->isSomeCtor() && !func->isDtor()) {
                     nsmap.value().insert(func->name(),(*child));
                 }
             }
